/************************************************************************/
/*                                                                      */
/*      Print Edit - Generic WebExtension - Background Page             */
/*                                                                      */
/*      Javascript for Background Page                                  */
/*                                                                      */
/*      Last Edit - 02 Mar 2016                                         */
/*                                                                      */
/*      Copyright (C) 2013-2016 DW-dev                                  */
/*                                                                      */
/*      Distributed under the GNU General Public License version 2      */
/*      See LICENCE.txt file and http://www.gnu.org/licenses/           */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Refer to Google Chrome developer documentation:                     */
/*                                                                      */
/*  https://developer.chrome.com/extensions/overview                    */
/*  https://developer.chrome.com/extensions/content_scripts             */
/*  https://developer.chrome.com/extensions/messaging                   */
/*  https://developer.chrome.com/extensions/xhr                         */
/*  https://developer.chrome.com/extensions/contentSecurityPolicy       */
/*                                                                      */
/************************************************************************/

"use strict";

/************************************************************************/

/* Global variables */

var isFirefox = (typeof chrome.pageCapture == "undefined");
var popupTabId;
var blockedTabIds = [];

/************************************************************************/

/* Initialize on browser startup */

chrome.storage.local.get(null,
function(object)
{
    /* Load options */
    
    if (Object.keys(object).length == 0)  /* on installation */
    {
        chrome.storage.local.set(
        {
            /* Print Edit Options */
            
            "options-before-adjust": true,
            "options-hide-select": false,
            "options-undo-reinstate": true,
            "options-close-donotask": false,
            "options-unload-askonly": false,
            "options-textpieces-breaks": true,
            
            "options-purge-deleted": true,
            "options-purge-hidden": true,
            "options-purge-styles": false,
            
            "options-select-border": false,
            "options-select-area": true,
            "options-capture-snap": true,
            
            /* Format Properties - States */
            
            "format-left": false,
            "format-top": false,
            "format-width": false,
            "format-height": false,
            "format-margin-h": false,
            "format-margin-v": false,
            "format-padding-h": false,
            "format-padding-v": false,
            "format-border": false,
            "format-outline": false,
            "format-overflow": false,
            "format-float": false,
            "format-clear": false,
            
            "format-color": false,
            "format-background": false,
            "format-fontfamily": false,
            "format-fontstyle": false,
            "format-fontsize": false,
            "format-lineheight": false,
            "format-verticalalign": false,
            "format-textalign": false,
            "format-textdecoration": false,
            "format-texttransform": false,
            "format-liststyle": false,
            "format-tableborder": false,
            "format-pagebreak": false,
            
            /* Format Properties - Values */
            
            "format-left-value": "auto",
            "format-top-value": "auto",
            "format-width-value": "auto",
            "format-height-value": "auto",
            "format-margin-h-value": "0px",
            "format-margin-v-value": "0px",
            "format-padding-h-value": "0px",
            "format-padding-v-value": "0px",
            "format-border-value": "0px",
            "format-outline-value": "0px",
            "format-overflow-value": "visible",
            "format-float-value": "none",
            "format-clear-value": "none",
            
            "format-color-value": "black",
            "format-background-value": "white",
            "format-fontfamily-value": "Arial",
            "format-fontstyle-value": "normal",
            "format-fontsize-value": "10px",
            "format-lineheight-value": "1.0",
            "format-verticalalign-value": "baseline",
            "format-textalign-value": "left",
            "format-textdecoration-value": "none",
            "format-texttransform-value": "none",
            "format-liststyle-value": "none",
            "format-tableborder-value": "collapse",
            "format-pagebreak-value": "before",
            
            /* Format Controls */
            
            "format-subelements": false,
            "format-important": false,
            "format-restrict": false,
            "format-restrict-type": "div",
            
            /* Format Show Properties */
            
            "format-showprops": false,
            
            /* Text Insert Type */
            
            "text-insert-type": "ib",
            
            /* Help Show Pane */
            
            "help-showpane": "button"
        });
    }
    
    /* Add context menu items */
    
    chrome.contextMenus.create({ id: "print", title: "Print...", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "preview", title: "Print Preview", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "edit", title: "Print Edit", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "separator", type: "separator", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "options", title: "Options", contexts: [ "all" ], enabled: true });
    
    /* Add listeners */
    
    addListeners();
});

/************************************************************************/

/* Add listeners */

function addListeners()
{
    /* Web request listeners */
    
    chrome.webRequest.onBeforeRequest.addListener(
    function(details)
    {
        if (blockedTabIds.indexOf(details.tabId) == -1) return;
        
        /* Before HTTP request - redirect HTTP request */
        
        if (details.type == "main_frame")
        {
            /* If main-frame request (reload or navigation), redirect to about:blank and close tab */
            
            chrome.tabs.sendMessage(details.tabId,{ type: "removeBeforeUnload" });  /* avoid multiple unload prompts */
            
            chrome.tabs.remove(details.tabId);  /* terminate Print Edit  session */
        }
        else
        {
            /* If sub-frame or resource request, just redirect to about:blank */
        }
        
        return { redirectUrl: "about:blank" };
    }
    ,{ urls: ["<all_urls>"] },["blocking"]);
    
    chrome.webRequest.onBeforeSendHeaders.addListener(
    function(details)
    {
        if (blockedTabIds.indexOf(details.tabId) == -1) return;
        
        /* In case redirection failed - cancel HTTP request */
        
        if (details.type == "main_frame")
        {
            /* If main-frame request (reload or navigation), cancel request and close tab */
            
            chrome.tabs.sendMessage(details.tabId,{ type: "removeBeforeUnload" });  /* avoid multiple unload prompts */
            
            chrome.tabs.remove(details.tabId);  /* terminate Print Edit  session */
        }
        else
        {
            /* If sub-frame or resource request, just cancel request */
        }
        
        return { cancel: true };
    }
    ,{ urls: ["<all_urls>"] },["blocking"]);
    
    chrome.webRequest.onSendHeaders.addListener(
    function(details)
    {
        if (blockedTabIds.indexOf(details.tabId) == -1) return;
        
        /* In case cancel failed - should never happen - for any request, close tab */
        
        chrome.tabs.sendMessage(details.tabId,{ type: "removeBeforeUnload" });  /* avoid multiple unload prompts */
        
        chrome.tabs.remove(details.tabId);  /* terminate Print Edit  session */
        
        alertNotify("Print Edit has encountered an error: PE-WR-1\n\n" +
                    "Please e-mail support at: dw-dev@gmx.com");
    }
    ,{ urls: ["<all_urls>"] });
    
    /* Tab event listeners */
    
    chrome.tabs.onRemoved.addListener(
    function(tabId,removeInfo)
    {
        var i;
        
        i = blockedTabIds.indexOf(tabId);
        
        if (i >= 0)
        {
            blockedTabIds.splice(i,1);
        }
    });
    
    /* Browser action listener */
    
    chrome.browserAction.onClicked.addListener(
    function(tab)
    {
        initiatePrintEdit(tab);
    });
    
    /* Context menu listener */
    
    chrome.contextMenus.onClicked.addListener(
    function(info,tab)
    {
        if (info.menuItemId == "print") chrome.tabs.executeScript(tab.id,{ code: "window.print();" });
        else if (info.menuItemId == "preview")  chrome.tabs.executeScript(tab.id,{ code: "window.print();" });
        else if (info.menuItemId == "edit") initiatePrintEdit(tab);
        else if (info.menuItemId == "options") chrome.runtime.openOptionsPage();
    });
    
    /* Notification clicked listener */
    
    if (!isFirefox)  /* Firefox - workaround until chrome.notifications supports buttons */  /*???*/
    {
        chrome.notifications.onButtonClicked.addListener(
        function(notificationId,buttonIndex)
        {
            if (notificationId == "confirm-popup")
            {
                if (buttonIndex == 0)
                {
                    chrome.tabs.executeScript(popupTabId,{ file: "content.js" },
                    function()
                    {
                    });
                    
                    chrome.tabs.update(popupTabId,{ active: true });
                }
                
                chrome.notifications.clear("confirm-popup");
            }
        });
    }
    
    /* Message received listener */
    
    chrome.runtime.onMessage.addListener(
    function(message,sender,sendResponse)
    {
        var xhr = new Object();
        
        /* Messages from content script */
        
        switch (message.type)
        {
            case "buttonClicked":
                
                chrome.tabs.query({ lastFocusedWindow: true, active: true },
                function(tabs)
                {
                    if (message.mode == -2)  /* content script not loaded */
                    {
                        chrome.windows.getLastFocused({ },
                        function(focuswin)
                        {
                            if (focuswin.type == "normal")
                            {
                                chrome.tabs.create({ url: tabs[0].url, index: tabs[0].index+1 },
                                function(tab)
                                {
                                    window.setTimeout(function()  /* Firefox - callback function called before page loaded */  /*???*/
                                    {
                                        chrome.tabs.executeScript(tab.id,{ file: "content.js" },
                                        function()
                                        {
                                        });
                                    },300);
                                });
                            }
                            else if (focuswin.type == "popup")
                            {
                                popupTabId = tabs[0].id;
                                
                                if (isFirefox)  /* Firefox - workaround until chrome.notifications supports buttons */  /*???*/
                                {
                                    chrome.tabs.executeScript(tabs[0].id,
                                        { code: "chrome.runtime.sendMessage({" +
                                                "type: 'confirmPopup'," +
                                                "result: confirm('Print Edit will prevent further actions or navigation in this popup window.\\n\\n" +
                                                "Do you want to continue?') });" });
                                }
                                else
                                {
                                    confirmNotify("Print Edit will prevent further actions or navigation in this popup window.\n\n" +
                                                  "Do you want to continue?","popup","OK","Cancel");
                                }
                            }
                        });
                    }
                    else if (message.mode == 0)  /* initialized */
                    {
                        chrome.tabs.sendMessage(tabs[0].id,{ type: "startEditing" });
                        
                        chrome.browserAction.setTitle({ title: "Print Edit - Editing", tabId: tabs[0].id });
                        chrome.browserAction.setBadgeText({ text: "EDIT", tabId: tabs[0].id });
                        chrome.browserAction.setBadgeBackgroundColor({ color: "#0000FF", tabId: tabs[0].id });
                    }
                    else if (message.mode == 1)  /* suspended */
                    {
                        chrome.tabs.sendMessage(tabs[0].id,{ type: "resumeEditing" });
                        
                        chrome.browserAction.setTitle({ title: "Print Edit - Editing", tabId: tabs[0].id });
                        chrome.browserAction.setBadgeText({ text: "EDIT", tabId: tabs[0].id });
                        chrome.browserAction.setBadgeBackgroundColor({ color: "#0000FF", tabId: tabs[0].id });
                    }
                    else if (message.mode == 2)  /* editing */
                    {
                        chrome.tabs.sendMessage(tabs[0].id,{ type: "suspendEditing" });
                        
                        chrome.browserAction.setTitle({ title: "Print Edit - Suspended", tabId: tabs[0].id });
                        chrome.browserAction.setBadgeText({ text: "SUS", tabId: tabs[0].id });
                        chrome.browserAction.setBadgeBackgroundColor({ color: "#0000FF", tabId: tabs[0].id });
                    }
                });
                
                break;
                
            case "confirmPopup":  /* Firefox - workaround until chrome.notifications supports buttons */  /*???*/
                
                if (message.result)
                {
                    chrome.tabs.executeScript(popupTabId,{ file: "content.js" },
                    function()
                    {
                    });
                }
                
                break;
            
            case "activatedBadge":
                
                chrome.tabs.query({ lastFocusedWindow: true, active: true },
                function(tabs)
                {
                    chrome.browserAction.setTitle({ title: "Print Edit - Activated", tabId: tabs[0].id });
                    chrome.browserAction.setBadgeText({ text: "ACT", tabId: tabs[0].id });
                    chrome.browserAction.setBadgeBackgroundColor({ color: "#0000FF", tabId: tabs[0].id });
                });
                
                break;
                
            case "blockHTTP":
                
                chrome.tabs.query({ lastFocusedWindow: true, active: true },
                function(tabs)
                {
                    if (blockedTabIds.indexOf(tabs[0].id) == -1)
                    {
                        blockedTabIds.push(tabs[0].id);
                    }
                });
                
                break;
                
            case "editingBadge":
                
                chrome.tabs.query({ lastFocusedWindow: true, active: true },
                function(tabs)
                {
                    chrome.browserAction.setTitle({ title: "Print Edit - Editing", tabId: tabs[0].id });
                    chrome.browserAction.setBadgeText({ text: "EDIT", tabId: tabs[0].id });
                    chrome.browserAction.setBadgeBackgroundColor({ color: "#0000FF", tabId: tabs[0].id });
                });
                
                break;
                
            case "saveMHTML":  /* disabled for Firefox */
                
                chrome.tabs.query({ lastFocusedWindow: true, active: true },
                function(tabs)
                {
                    chrome.pageCapture.saveAsMHTML({ tabId: tabs[0].id },
                    function(mhtmlBlob)
                    {
                        var objectURL;
                        
                        objectURL = window.URL.createObjectURL(mhtmlBlob);
                        
                        /* Download MHTML blob as .mhtml file */
                        
                        chrome.downloads.onChanged.addListener(
                        function listener(downloadDelta)
                        {
                            if (downloadDelta.error && downloadDelta.error.current == "USER_CANCELED")
                            {
                                chrome.downloads.onChanged.removeListener(listener);
                                chrome.tabs.sendMessage(tabs[0].id,{ type: "saveDone", cancel: true });
                            }
                            else if (downloadDelta.filename && downloadDelta.filename.current.length > 0)
                            {
                                chrome.downloads.onChanged.removeListener(listener);
                                chrome.tabs.sendMessage(tabs[0].id,{ type: "saveDone", cancel: false });
                            }
                        });
                        
                        chrome.downloads.download({ url: objectURL, filename: tabs[0].title.replace(/(\\|\/|:|\*|\?|"|<|>|\|)/g," ") + ".mhtml", saveAs: true },
                        function(downloadItemId)
                        {
                        });
                    });
                });
                
                break;
                
            case "saveHTMLLoadImage":
                
                try
                {
                    xhr = new XMLHttpRequest();
                    
                    xhr.open("GET",message.location,true);
                    xhr.responseType = "arraybuffer";
                    xhr.timeout = 2000;
                    xhr._tabId = sender.tab.id;
                    xhr._imageIndex = message.index;
                    xhr.onload = onloadImage;
                    xhr.onerror = onerrorImage;
                    xhr.ontimeout = ontimeoutImage;
                    
                    xhr.send();  /* throws exception if url is invalid */
                }
                catch(e)
                {
                    chrome.tabs.sendMessage(sender.tab.id,{ type: "saveLoadFailure" });
                }
                
                function onloadImage()
                {
                    var i,binaryString,base64String,contentType;
                    var byteArray = new Uint8Array(typeof this.mozResponseArrayBuffer != "undefined" ? this.mozResponseArrayBuffer : this.response);  /* FF4 & FF5 */
                    
                    if (this.status == 200)
                    {
                        binaryString = "";
                        for (i = 0; i < byteArray.byteLength; i++) binaryString += String.fromCharCode(byteArray[i]);
                        base64String = btoa(binaryString);
                        
                        contentType = this.getResponseHeader("Content-Type").replace(/;.*/,"");
                        
                        chrome.tabs.sendMessage(this._tabId,{ type: "saveLoadSuccess", index: this._imageIndex, content: base64String, contenttype: contentType });
                    }
                    else chrome.tabs.sendMessage(this._tabId,{ type: "saveLoadFailure" });
                }
                
                function onerrorImage()
                {
                    chrome.tabs.sendMessage(this._tabId,{ type: "saveLoadFailure" });
                }
                
                function ontimeoutImage()
                {
                    chrome.tabs.sendMessage(this._tabId,{ type: "saveLoadFailure" });
                }
                
                break;
                
            // case "saveHTMLSaveFile":
                
                // chrome.tabs.query({ lastFocusedWindow: true, active: true },
                // function(tabs)
                // {
                    // var htmlBlob,objectURL;
                    
                    // /* Convert HTML Complete string to HTML Complete blob */
                    
                    // htmlBlob = new Blob( [ (new Uint8Array([ 0xEF,0xBB,0xBF ])), message.html ], { type : "text/html" });
                    
                    // objectURL = window.URL.createObjectURL(htmlBlob);
                    
                    // /* Download HTML Complete blob as .html file */
                    
                    // chrome.downloads.onChanged.addListener(
                    // function listener(downloadDelta)
                    // {
                        // if (downloadDelta.error && downloadDelta.error.current == "USER_CANCELED")
                        // {
                            // chrome.downloads.onChanged.removeListener(listener);
                            // chrome.tabs.sendMessage(tabs[0].id,{ type: "saveDone", cancel: true });
                        // }
                        // else if (downloadDelta.filename && downloadDelta.filename.current.length > 0)
                        // {
                            // chrome.downloads.onChanged.removeListener(listener);
                            // chrome.tabs.sendMessage(tabs[0].id,{ type: "saveDone", cancel: false });
                        // }
                    // });
                    
                    // chrome.downloads.download({ url: objectURL, filename: tabs[0].title.replace(/(\\|\/|:|\*|\?|"|<|>|\|)/g," ") + ".html", saveAs: true },
                    // function(downloadItemId)
                    // {
                    // });
                // });
                
                // break;
                
            case "closeTab":
                
                chrome.tabs.query({ lastFocusedWindow: true, active: true },
                function(tabs)
                {
                    chrome.tabs.remove(tabs[0].id);
                });
                
                break;
                
            case "openOptions":
                
                chrome.runtime.openOptionsPage();
                
                break;
        }
    });
}

/************************************************************************/

/* Initiate print edit function */

function initiatePrintEdit(tab)
{
    if (tab.url.substr(0,6) == "about:" || tab.url.substr(0,7) == "chrome:" || 
        tab.url.substr(0,17) == "chrome-extension:" || tab.url.substr(0,34) == "https://chrome.google.com/webstore")
    {
        alertNotify("Print Edit cannot be used with 'about:' or 'chrome:' pages.");
              
        return;
    }
    else if (tab.url.substr(0,5) == "file:")
    {
        chrome.extension.isAllowedFileSchemeAccess(
        function(isAllowedAccess)
        {
            if (isFirefox)  /* Firefox - workaround until chrome.tabs.create supports file: pages */  /*???*/
            {
                alertNotify("Print Edit cannot be used with 'file:' pages in Firefox.");
            }
            else
            {
                if (isAllowedAccess)
                {
                    chrome.tabs.executeScript(tab.id,{ code: "chrome.runtime.sendMessage({" +
                                                             "type: 'buttonClicked'," +
                                                             "mode: (typeof curMode == 'undefined') ? -2 : curMode });" });
                }
                else
                {
                    chrome.tabs.create({ url: 'chrome://extensions/?id=' + chrome.runtime.id });
                    
                    alertNotify("Print Edit has requested additional permissions to allow access to file URLs.\n\n" +
                                "Please tick the 'Allow access to file URLs' checkbox on the Extensions page.");
                }
            }
        });
    }
    else
    {
        chrome.tabs.executeScript(tab.id,{ code: "chrome.runtime.sendMessage({" +
                                                 "type: 'buttonClicked'," +
                                                 "mode: (typeof curMode == 'undefined') ? -2 : curMode });" });
    }
}

/************************************************************************/

/* Display alert and confirm notifications */

function alertNotify(message)
{
    chrome.notifications.create("alert",{ type: "basic", iconUrl: "icon32.png", title: "PRINT EDIT", message: "" + message });
}

function confirmNotify(message,id,title0,title1)
{
    chrome.notifications.create("confirm-" + id,{ type: "basic", iconUrl: "icon32.png", title: "PRINT EDIT", message: "" + message,
                                                  buttons: [ { title: title0 }, { title: title1 } ] });
}

/************************************************************************/

/* Display debug notification */

function debugNotify(message)
{
    chrome.notifications.create("debug",{ type: "basic", iconUrl: "icon32.png", title: "PRINT EDIT - DEBUG", message: "" + message });
}

function debugModal(message)
{
    chrome.tabs.query({ lastFocusedWindow: true, active: true },
    function(tabs)
    {
        chrome.tabs.executeScript(tabs[0].id,{ code: "alert('" + message + "');" });
    });
}

/************************************************************************/
