/************************************************************************/
/*                                                                      */
/*      Print Edit - Generic WebExtension - Content Pages               */
/*                                                                      */
/*      Javascript for Content Pages                                    */
/*                                                                      */
/*      Last Edit - 02 Mar 2016                                         */
/*                                                                      */
/*      Copyright (C) 2013-2016 DW-dev                                  */
/*                                                                      */
/*      Distributed under the GNU General Public License version 2      */
/*      See LICENCE.txt file and http://www.gnu.org/licenses/           */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/* Refer to Google Chrome developer documentation:                      */
/*                                                                      */
/*   https://developer.chrome.com/extensions/content_scripts            */
/*                                                                      */
/************************************************************************/

"use strict";

/************************************************************************/

/* Global variables */

var isFirefox = (typeof document.body.ownerDocument.defaultView.mozInnerScreenX != "undefined");

var beforeAdjust,hideSelect,undoReinstate,closeDoNotAsk,unloadAskOnly,textpiecesBreaks;
var purgeDeleted,purgeHidden,purgeStyles;    
var selectBorder,selectArea,captureSnap;    

var guiContainer = null;
var toolbarHeight = 31;

var curMode = -1;  /* -1 = loaded, 0 = initialized, 1 = suspended, 2 = editing */

var popupOpen = 0;  /* 0 = none, 1 = inspect panel, 2 = select menu, 3 = deleteexcept menu, 4 = format panel, 5 = text panel, 6 = save panel, 7 = help panel */

var downFlag = false;
var downKeys = 0;
var downPageX = 0;
var downPageY = 0;

var captureBox = null;
var captureWindow = null;
var captureWidth = 0;
var captureHeight = 0;
var captureLeft = 0;
var captureTop = 0;

var downPanel = null;
var downOffsetX = 0;
var downOffsetY =0;

var scrollIncX = 0;
var scrollIncY = 0;
var scrollScreenX = 0;
var scrollScreenY = 0;
var scrollTimeout = null;

var formatFirstCmd = false;

var textContainer = "";
var textPieceValue = "";
var textInsertValue = "";
var textInsertType = "";
var textFirstCmd = false;

var saveMaxDepth = 2;
var saveImageCount;
var saveBaseURI;

var saveImageLocation = new Array();
var saveImageContentType = new Array();
var saveImageContent = new Array();

var saveCmdNumber = 0;

var cmdNumber = 0;

var highlightStack = new Array();

var highlightElement = null;
var highlightBox = null;

var selectCount = new Array();
var selectElement = new Array();  /* two dimensional */
var selectBox = new Array();

var debugEnable = false;
var debugCount = 0;
   
var captureAll = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAABGdBTUEAALGPC/xhBQAAAAlwSFlzAAAOwgAADsIBFShKgAAAABh0RVh0U29mdHdhcmUAcGFpbnQubmV0IDQuMC4zjOaXUAAAACpJREFUKFONyKENAAAIA7D9tP9vA08qEDVNMv1hClOYwhSmMIUpTGFe0wVvY3eJZ3Uc2wAAAABJRU5ErkJggg==";
var captureGraphicsLeft = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsMAAA7DAcdvqGQAAAAYdEVYdFNvZnR3YXJlAHBhaW50Lm5ldCA0LjAuM4zml1AAAAAXSURBVChTY8AO/htjYqxgVCFtFP43BgBisTvFbKwfDwAAAABJRU5ErkJggg==";
var captureGraphicsRight = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsMAAA7DAcdvqGQAAAAYdEVYdFNvZnR3YXJlAHBhaW50Lm5ldCA0LjAuM4zml1AAAAAXSURBVChTY2Bg+G+MibGCUYX0U8jAAAAOTDvFe4Ki3gAAAABJRU5ErkJggg==";
var captureDeleteTop = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAABGdBTUEAALGPC/xhBQAAAAlwSFlzAAAOwgAADsIBFShKgAAAABh0RVh0U29mdHdhcmUAcGFpbnQubmV0IDQuMC4zjOaXUAAAABhJREFUKFNjYGD4b0wcxiqIDY8CygEDAwBpEBfp7rnlRgAAAABJRU5ErkJggg==";
var captureDeleteBottom = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAABGdBTUEAALGPC/xhBQAAAAlwSFlzAAAOwgAADsIBFShKgAAAABh0RVh0U29mdHdhcmUAcGFpbnQubmV0IDQuMC4zjOaXUAAAABZJREFUKFNjGAXUAv+NicNYBdHxf2MAxf8X6T4nSAgAAAAASUVORK5CYII=";

var styleAnimations = "body, body * { animation: none !important; -webkit-animation: none !important; transition-property: none !important; }";

var styleLineBreaks = "body span[printedit-break-wrapper] { margin-left: -10px !important; border-left: 10px solid transparent !important; }" +
                      "body span[printedit-break-wrapper] span { position: relative !important; pointer-events: none !important; }" +
                      "body span[printedit-break-wrapper] span::before { content: '' !important; display: inline-block !important; position: absolute !important;" +
                          "left: 0px !important; top: 10% !important; width: 9px !important; height: calc(90% - 1px) !important; margin-left: -10px !important;" +
                          "border-width: 0px 1px 1px 0px !important; border-style: solid !important; border-color: red !important; }" +
                      "body span[printedit-break-wrapper] span::after { content: '' !important; display: inline-block !important; position: absolute !important;" +
                          "left: -3px !important; top: calc(100% - 4px) !important; width: 1px !important; height: 1px !important; margin-left: -10px !important;" +
                          "background-clip: content-box !important; background-color: red !important;" +
                          "border-width: 3px !important; border-style: solid !important; border-color: transparent !important; border-right-color: red !important; }";

/************************************************************************/

/* Initialize on script load */

chrome.storage.local.get(null,
function(object)
{
    /* Load options */
    
    beforeAdjust = object["options-before-adjust"];
    hideSelect = object["options-hide-select"];
    undoReinstate = object["options-undo-reinstate"];
    closeDoNotAsk = object["options-close-donotask"];
    unloadAskOnly = object["options-unload-askonly"];
    textpiecesBreaks = object["options-textpieces-breaks"];
        
    purgeDeleted = object["options-purge-deleted"];
    purgeHidden = object["options-purge-hidden"];
    purgeStyles = object["options-purge-styles"];
        
    selectBorder = object["options-select-border"];
    selectArea = object["options-select-area"];
    captureSnap = object["options-capture-snap"];
    
    /* Load GUI */
    
    loadGUI();
});

/************************************************************************/

/* Load GUI */

function loadGUI()
{
    var i,count;
    var xhr = new Array();
    var urls = new Array("toolbar.html","inspectpanel.html","deleteexceptmenu.html","selectmenu.html",
                         "formatpanel.html","textpanel.html","savemenu.html","helppanel.html");

    /* Create GUI Container element */
    
    guiContainer = document.createElement("div");
    guiContainer.setAttribute("id","printedit-gui-container");
    document.documentElement.appendChild(guiContainer);
    
    /* Load HTML files into GUI Container element */
    
    count = 0;
    
    for (i = 0; i < urls.length; i++)
    {
        request(i);
    }
    
    function request(i)
    {
        xhr[i] = new XMLHttpRequest();
        xhr[i].open("GET",chrome.extension.getURL(urls[i]),true);
        xhr[i].onload = complete;
        xhr[i].send();
    
        function complete()
        {
            if (xhr[i].readyState == 4 && xhr[i].status == 200)
            {
                guiContainer.innerHTML += xhr[i].responseText;
                
                if (++count == urls.length)
                {
                    /* Remove comments loaded from HTML files */
                    
                    for (i = 0; i < guiContainer.childNodes.length; i++)
                        if (guiContainer.childNodes[i].nodeType == 8)  /* comment node */
                            guiContainer.removeChild(guiContainer.childNodes[i]);
                            
                    /* Initialize GUI */
                    
                    initGUI();
                    
                    /* Add listeners */
                    
                    addListeners();
                    
                    /* Wait for web page load before editing */
                    
                    waitBeforeEditing();
                }
            }
        }
    }
}

/************************************************************************/

/* Initialize GUI */

function initGUI()
{
    var select,selectwidth,deselect,deselectwidth,edit,editwidth,preview,previewwidth;
    var button,menu,panel,textarea;
    var buttonRect = new Object();
    
    /* Enable debug information display */
    
    if (debugEnable)
    {
        document.getElementById("printedit-logo").style.setProperty("display","none","important");
        document.getElementById("printedit-debug-1").style.setProperty("display","inline-block","important");
        document.getElementById("printedit-debug-2").style.setProperty("display","inline-block","important");
        document.getElementById("printedit-debug-3").style.setProperty("display","inline-block","important");
        document.getElementById("printedit-debug-4").style.setProperty("display","inline-block","important");
    }
    
    /* Position document body */
    
    document.body.style.setProperty("transform","translateY(31px)","important");
    document.body.style.setProperty("-webkit-transform","translateY(31px)","important");
    
    /* Adjust toolbar spacer */
    
    adjustToolbar();
    
    /* Equalize widths of select and deselect buttons */
    
    select = document.getElementById("printedit-select");
    buttonRect = select.getBoundingClientRect();
    selectwidth = buttonRect.width;

    deselect = document.getElementById("printedit-deselect");
    buttonRect = deselect.getBoundingClientRect();
    deselectwidth = buttonRect.width;
    
    if (selectwidth < deselectwidth) select.style.setProperty("width",(deselectwidth-10)+"px","important");
    else deselect.style.setProperty("width",(selectwidth-10)+"px","important");
    
    deselect.style.setProperty("display","none","important");
    
    /* Equalize widths of edit and preview buttons */
    
    edit = document.getElementById("printedit-edit");
    buttonRect = edit.getBoundingClientRect();
    editwidth = buttonRect.width;
    
    preview = document.getElementById("printedit-preview");
    buttonRect = preview.getBoundingClientRect();
    previewwidth = buttonRect.width;
    
    if (editwidth < previewwidth) edit.style.setProperty("width",(previewwidth-10)+"px","important");
    else preview.style.setProperty("width",(editwidth-10)+"px","important");
    
    edit.setAttribute("disabled","");
    preview.style.setProperty("display","none","important");
    
    /* Hide inspect panel */
    
    panel = document.getElementById("printedit-inspect-panel");
    panel.style.setProperty("display","none","important");
    
    /* Position and hide select menu */
    
    button = document.getElementById("printedit-select");
    menu = document.getElementById("printedit-select-menu");
    buttonRect = button.getBoundingClientRect();
    menu.style.setProperty("left",(buttonRect.left)+"px","important");
    menu.style.setProperty("display","none","important");
    
    /* Position and hide delete except menu */
    
    button = document.getElementById("printedit-deleteexcept");
    menu = document.getElementById("printedit-deleteexcept-menu");
    buttonRect = button.getBoundingClientRect();
    menu.style.setProperty("left",(buttonRect.left)+"px","important");
    menu.style.setProperty("display","none","important");
    
    /* Position and hide format panel */
    
    button = document.getElementById("printedit-format");
    panel = document.getElementById("printedit-format-panel");
    buttonRect = button.getBoundingClientRect();
    panel.style.setProperty("left",(buttonRect.left)+"px","important");
    panel.style.setProperty("display","none","important");
    
    /* Position and hide text panel */
    
    button = document.getElementById("printedit-text");
    panel = document.getElementById("printedit-text-panel");
    buttonRect = button.getBoundingClientRect();
    panel.style.setProperty("left",(buttonRect.left)+"px","important");
    panel.style.setProperty("display","none","important");
    
    /* Position and hide save menu */
    
    button = document.getElementById("printedit-save");
    menu = document.getElementById("printedit-save-menu");
    buttonRect = button.getBoundingClientRect();
    menu.style.setProperty("left",(buttonRect.left)+"px","important");
    menu.style.setProperty("display","none","important");
    
    /* Hide help panel */
    
    panel = document.getElementById("printedit-help-panel");
    panel.style.setProperty("display","none","important");
    
    /* Workaround to allow <textarea> in Text Panel to shrink below original width */  /*???*/
    
    textarea = document.getElementById("printedit-text-textarea");
    textarea.resizeStarted = false;
    textarea.addEventListener("mousedown",observeTextArea,true);
    
    function observeTextArea(event)
    {
        textarea.originalWidth = textarea.clientWidth;
        textarea.addEventListener("mousemove",resizeTextArea,true);
        window.addEventListener("mouseup",ignoreTextArea,true);
    }
    
    function resizeTextArea(event)
    {
        var width;
        
        if (!textarea.resizeStarted)
        {
            if (textarea.clientWidth != textarea.originalWidth)
            {
                textarea.resizeStarted = true;
                textarea.resizeStartX = event.clientX;
            }
        }
        else
        {
            width = textarea.originalWidth+(event.clientX-textarea.resizeStartX);
            if (width < textarea.originalWidth) textarea.style.setProperty("width",width+"px","important");
        }
    }
    
    function ignoreTextArea(event)
    {
        textarea.resizeStarted = false;
        textarea.removeEventListener("mousemove",resizeTextArea,true);
        window.removeEventListener("mouseup",ignoreTextArea,true);
    }
    
    /* Disable save as MHTML menu items for Firefox */
    
    if (isFirefox)
    {
        document.getElementById("printedit-save-mhtml").setAttribute("disabled","");
        document.getElementById("printedit-save-mhtmlpurge").setAttribute("disabled","");
    }
    
    /* Disable edit and undo buttons */
    
    disableEditButtons();
    disableUndoButtons();
    
    /* Disable other buttons */
    
    disableOtherButtons();
}

/************************************************************************/

/* Add listeners */

function addListeners()
{
    /* Toolbar button event listeners */
    
    document.getElementById("printedit-select").addEventListener("click",function() { cmdSelect(); },false);
    document.getElementById("printedit-deselect").addEventListener("click",function() { cmdDeselect(); },false);
    document.getElementById("printedit-hide").addEventListener("click",function() { cmdHide(); },false);
    document.getElementById("printedit-delete").addEventListener("click",function() { cmdDelete(); },false);
    document.getElementById("printedit-hideexcept").addEventListener("click",function() { cmdHideExcept(); },false);
    document.getElementById("printedit-deleteexcept").addEventListener("click",function() { cmdDeleteExcept(); },false);
    document.getElementById("printedit-format").addEventListener("click",function() { cmdFormat(); },false);
    document.getElementById("printedit-text").addEventListener("click",function() { cmdText(); },false);
    document.getElementById("printedit-undo").addEventListener("click",function() { cmdUndo(); },false);
    document.getElementById("printedit-undoall").addEventListener("click",function() { cmdUndoAll(); },false);
    document.getElementById("printedit-save").addEventListener("click",function() { cmdSave(); },false);
    document.getElementById("printedit-textpieces").addEventListener("click",function() { cmdTextPieces(); },false);
    document.getElementById("printedit-webstyle").addEventListener("click",function() { cmdWebStyle(); },false);
    document.getElementById("printedit-edit").addEventListener("click",function() { cmdEdit(); },false);
    document.getElementById("printedit-preview").addEventListener("click",function() { cmdPreview(); },false);
    document.getElementById("printedit-close").addEventListener("click",function() { cmdClose(); },false);
    document.getElementById("printedit-help").addEventListener("click",function() { cmdHelp(); },false);
    document.getElementById("printedit-options").addEventListener("click",function() { cmdOptions(); },false);
    
    /* Inspect panel event listener */
    
    document.getElementById("printedit-inspect-expand").addEventListener("click",function() { cmdExpandBox(); },false);
    document.getElementById("printedit-inspect-shrink").addEventListener("click",function() { cmdShrinkBox(); },false);
    document.getElementById("printedit-inspect-select").addEventListener("click",function() { cmdSelectElement(); },false);
    document.getElementById("printedit-inspect-close").addEventListener("click",function() { closePopup(); },false);
    
    /* Select menu event listeners */
    
    document.getElementById("printedit-select-all").addEventListener("click",function() { cmdSelectAll(); },false);
    document.getElementById("printedit-select-graphics").addEventListener("click",function() { cmdSelectGraphics(); },false);
    document.getElementById("printedit-select-previous").addEventListener("click",function() { cmdSelectPrevious(); },false);
    
    /* Delete except menu event listeners */
    
    document.getElementById("printedit-deleteexcept-restricted").addEventListener("click",function() { cmdDeleteExceptRestricted(); },false);
    document.getElementById("printedit-deleteexcept-withoutfloat").addEventListener("click",function() { cmdDeleteExceptWithoutFloat(); },false);
    document.getElementById("printedit-deleteexcept-unrestricted").addEventListener("click",function() { cmdDeleteExceptUnrestricted(); },false);
    
    /* Save menu event listeners */
    
    document.getElementById("printedit-save-mhtml").addEventListener("click",function() { cmdSaveStart(false,false); },false);
    document.getElementById("printedit-save-mhtmlpurge").addEventListener("click",function() { cmdSaveStart(false,true); },false);
    document.getElementById("printedit-save-html").addEventListener("click",function() { cmdSaveStart(true,false); },false);
    document.getElementById("printedit-save-htmlpurge").addEventListener("click",function() { cmdSaveStart(true,true); },false);
    
    /* Format panel event listeners */
    
    document.getElementById("printedit-format-close").addEventListener("click",function() { closePopup(); },false);
    
    document.getElementById("printedit-format-textstyle").addEventListener("click",function() { cmdFormatTextStyle(); },false);
    document.getElementById("printedit-format-decoration").addEventListener("click",function() { cmdFormatDecoration(); },false);
    document.getElementById("printedit-format-margins").addEventListener("click",function() { cmdFormatMargins(); },false);
    document.getElementById("printedit-format-dimensions").addEventListener("click",function() { cmdFormatDimensions(); },false);
    document.getElementById("printedit-format-blocklayout").addEventListener("click",function() { cmdFormatBlockLayout(); },false);
    document.getElementById("printedit-format-setall").addEventListener("click",function() { cmdFormatSetAll(true); },false);
    document.getElementById("printedit-format-clearall").addEventListener("click",function() { cmdFormatSetAll(false); },false);
    
    document.getElementById("printedit-format-left").addEventListener("click",function() { cmdFormatToggleState("left"); },false);
    document.getElementById("printedit-format-left-value").addEventListener("click",function() { cmdFormatResetValue("left","auto"); },false);
    document.getElementById("printedit-format-left-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("left","-200px,-150px,-100px,-50px,-30px,-20px,-10px,auto,0px,10px,20px,30px,50px,100px,150px,200px,250px,300px,400px,500px,600px,800px,1000px"); },false);
    document.getElementById("printedit-format-left-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("left","-200px,-150px,-100px,-50px,-30px,-20px,-10px,auto,0px,10px,20px,30px,50px,100px,150px,200px,250px,300px,400px,500px,600px,800px,1000px"); },false);
    
    document.getElementById("printedit-format-top").addEventListener("click",function() { cmdFormatToggleState("top"); },false);
    document.getElementById("printedit-format-top-value").addEventListener("click",function() { cmdFormatResetValue("top","auto"); },false);
    document.getElementById("printedit-format-top-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("top","-200px,-150px,-100px,-50px,-30px,-20px,-10px,auto,0px,10px,20px,30px,50px,100px,150px,200px,250px,300px,400px,500px,600px,800px,1000px"); },false);
    document.getElementById("printedit-format-top-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("top","-200px,-150px,-100px,-50px,-30px,-20px,-10px,auto,0px,10px,20px,30px,50px,100px,150px,200px,250px,300px,400px,500px,600px,800px,1000px"); },false);
    
    document.getElementById("printedit-format-width").addEventListener("click",function() { cmdFormatToggleState("width"); },false);
    document.getElementById("printedit-format-width-value").addEventListener("click",function() { cmdFormatResetValue("width","auto"); },false);
    document.getElementById("printedit-format-width-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("width","auto,10px,15px,20px,25px,30px,40px,50px,60px,80px,100px,150px,200px,250px,300px,400px,500px,600px,800px,1000px"); },false);
    document.getElementById("printedit-format-width-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("width","auto,10px,15px,20px,25px,30px,40px,50px,60px,80px,100px,150px,200px,250px,300px,400px,500px,600px,800px,1000px"); },false);
    
    document.getElementById("printedit-format-height").addEventListener("click",function() { cmdFormatToggleState("height"); },false);
    document.getElementById("printedit-format-height-value").addEventListener("click",function() { cmdFormatResetValue("height","auto"); },false);
    document.getElementById("printedit-format-height-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("height","auto,10px,15px,20px,25px,30px,40px,50px,60px,80px,100px,150px,200px,250px,300px,400px,500px,600px,800px,1000px"); },false);
    document.getElementById("printedit-format-height-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("height","auto,10px,15px,20px,25px,30px,40px,50px,60px,80px,100px,150px,200px,250px,300px,400px,500px,600px,800px,1000px"); },false);
    
    document.getElementById("printedit-format-margin-h").addEventListener("click",function() { cmdFormatToggleState("margin-h"); },false);
    document.getElementById("printedit-format-margin-h-value").addEventListener("click",function() { cmdFormatResetValue("margin-h","0px"); },false);
    document.getElementById("printedit-format-margin-h-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("margin-h","-200px,-150px,-100px,-50px,-30px,-20px,-10px,0px,2px,4px,6px,8px,10px,15px,20px,25px,30px,50px,100px,150px,200px"); },false);
    document.getElementById("printedit-format-margin-h-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("margin-h","-200px,-150px,-100px,-50px,-30px,-20px,-10px,0px,2px,4px,6px,8px,10px,15px,20px,25px,30px,50px,100px,150px,200px"); },false);
    
    document.getElementById("printedit-format-margin-v").addEventListener("click",function() { cmdFormatToggleState("margin-v"); },false);
    document.getElementById("printedit-format-margin-v-value").addEventListener("click",function() { cmdFormatResetValue("margin-v","0px"); },false);
    document.getElementById("printedit-format-margin-v-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("margin-v","-200px,-150px,-100px,-50px,-30px,-20px,-10px,0px,2px,4px,6px,8px,10px,15px,20px,25px,30px,50px,100px,150px,200px"); },false);
    document.getElementById("printedit-format-margin-v-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("margin-v","-200px,-150px,-100px,-50px,-30px,-20px,-10px,0px,2px,4px,6px,8px,10px,15px,20px,25px,30px,50px,100px,150px,200px"); },false);
    
    document.getElementById("printedit-format-padding-h").addEventListener("click",function() { cmdFormatToggleState("padding-h"); },false);
    document.getElementById("printedit-format-padding-h-value").addEventListener("click",function() { cmdFormatResetValue("padding-h","0px"); },false);
    document.getElementById("printedit-format-padding-h-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("padding-h","0px,2px,4px,6px,8px,10px,15px,20px,25px,30px,50px,100px,150px,200px"); },false);
    document.getElementById("printedit-format-padding-h-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("padding-h","0px,2px,4px,6px,8px,10px,15px,20px,25px,30px,50px,100px,150px,200px"); },false);
    
    document.getElementById("printedit-format-padding-v").addEventListener("click",function() { cmdFormatToggleState("padding-v"); },false);
    document.getElementById("printedit-format-padding-v-value").addEventListener("click",function() { cmdFormatResetValue("padding-v","0px"); },false);
    document.getElementById("printedit-format-padding-v-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("padding-v","0px,2px,4px,6px,8px,10px,15px,20px,25px,30px,50px,100px,150px,200px"); },false);
    document.getElementById("printedit-format-padding-v-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("padding-v","0px,2px,4px,6px,8px,10px,15px,20px,25px,30px,50px,100px,150px,200px"); },false);
    
    document.getElementById("printedit-format-border").addEventListener("click",function() { cmdFormatToggleState("border"); },false);
    document.getElementById("printedit-format-border-value").addEventListener("click",function() { cmdFormatResetValue("border","0px"); },false);
    document.getElementById("printedit-format-border-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("border","0px,1px,2px,3px,4px,5px,6px,7px,8px,9px,10px"); },false);
    document.getElementById("printedit-format-border-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("border","0px,1px,2px,3px,4px,5px,6px,7px,8px,9px,10px"); },false);
    
    document.getElementById("printedit-format-outline").addEventListener("click",function() { cmdFormatToggleState("outline"); },false);
    document.getElementById("printedit-format-outline-value").addEventListener("click",function() { cmdFormatResetValue("outline","0px"); },false);
    document.getElementById("printedit-format-outline-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("outline","0px,1px,2px,3px,4px,5px,6px,7px,8px,9px,10px"); },false);
    document.getElementById("printedit-format-outline-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("outline","0px,1px,2px,3px,4px,5px,6px,7px,8px,9px,10px"); },false);
    
    document.getElementById("printedit-format-overflow").addEventListener("click",function() { cmdFormatToggleState("overflow"); },false);
    document.getElementById("printedit-format-overflow-value").addEventListener("click",function() { cmdFormatResetValue("overflow","visible"); },false);
    document.getElementById("printedit-format-overflow-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("overflow","visible,hidden"); },false);
    document.getElementById("printedit-format-overflow-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("overflow","visible,hidden"); },false);
    
    document.getElementById("printedit-format-float").addEventListener("click",function() { cmdFormatToggleState("float"); },false);
    document.getElementById("printedit-format-float-value").addEventListener("click",function() { cmdFormatResetValue("float","none"); },false);
    document.getElementById("printedit-format-float-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("float","none,left,right"); },false);
    document.getElementById("printedit-format-float-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("float","none,left,right"); },false);
    
    document.getElementById("printedit-format-clear").addEventListener("click",function() { cmdFormatToggleState("clear"); },false);
    document.getElementById("printedit-format-clear-value").addEventListener("click",function() { cmdFormatResetValue("clear","none"); },false);
    document.getElementById("printedit-format-clear-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("clear","none,left,right,both"); },false);
    document.getElementById("printedit-format-clear-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("clear","none,left,right,both"); },false);
    
    document.getElementById("printedit-format-color").addEventListener("click",function() { cmdFormatToggleState("color"); },false);
    document.getElementById("printedit-format-color-value").addEventListener("click",function() { cmdFormatResetValue("color","black"); },false);
    document.getElementById("printedit-format-color-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("color","black,dark,onyx,iron,gray,silver,white,red,green,blue,yellow"); },false);
    document.getElementById("printedit-format-color-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("color","black,dark,onyx,iron,gray,silver,white,red,green,blue,yellow"); },false);
    
    document.getElementById("printedit-format-background").addEventListener("click",function() { cmdFormatToggleState("background"); },false);
    document.getElementById("printedit-format-background-value").addEventListener("click",function() { cmdFormatResetValue("background","white"); },false);
    document.getElementById("printedit-format-background-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("background","white,light,quartz,steel,silver,gray,black,red,green,blue,yellow"); },false);
    document.getElementById("printedit-format-background-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("background","white,light,quartz,steel,silver,gray,black,red,green,blue,yellow"); },false);
    
    document.getElementById("printedit-format-fontfamily").addEventListener("click",function() { cmdFormatToggleState("fontfamily"); },false);
    document.getElementById("printedit-format-fontfamily-value").addEventListener("click",function() { cmdFormatResetValue("fontfamily","Arial"); },false);
    document.getElementById("printedit-format-fontfamily-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("fontfamily","Arial,Times,Calibri,Courier"); },false);
    document.getElementById("printedit-format-fontfamily-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("fontfamily","Arial,Times,Calibri,Courier"); },false);
    
    document.getElementById("printedit-format-fontstyle").addEventListener("click",function() { cmdFormatToggleState("fontstyle"); },false);
    document.getElementById("printedit-format-fontstyle-value").addEventListener("click",function() { cmdFormatResetValue("fontstyle","normal"); },false);
    document.getElementById("printedit-format-fontstyle-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("fontstyle","normal,italic,bold"); },false);
    document.getElementById("printedit-format-fontstyle-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("fontstyle","normal,italic,bold"); },false);
    
    document.getElementById("printedit-format-fontsize").addEventListener("click",function() { cmdFormatToggleState("fontsize"); },false);
    document.getElementById("printedit-format-fontsize-value").addEventListener("click",function() { cmdFormatResetValue("fontsize","10px"); },false);
    document.getElementById("printedit-format-fontsize-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("fontsize","10px,11px,12px,13px,14px,15px,16px,17px,18px,19px,20px,22px,24px,26px,28px,30px"); },false);
    document.getElementById("printedit-format-fontsize-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("fontsize","10px,11px,12px,13px,14px,15px,16px,17px,18px,19px,20px,22px,24px,26px,28px,30px"); },false);
    
    document.getElementById("printedit-format-lineheight").addEventListener("click",function() { cmdFormatToggleState("lineheight"); },false);
    document.getElementById("printedit-format-lineheight-value").addEventListener("click",function() { cmdFormatResetValue("lineheight","1.0"); },false);
    document.getElementById("printedit-format-lineheight-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("lineheight","1.0,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9,2.0"); },false);
    document.getElementById("printedit-format-lineheight-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("lineheight","1.0,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9,2.0"); },false);
    
    document.getElementById("printedit-format-verticalalign").addEventListener("click",function() { cmdFormatToggleState("verticalalign"); },false);
    document.getElementById("printedit-format-verticalalign-value").addEventListener("click",function() { cmdFormatResetValue("verticalalign","baseline"); },false);
    document.getElementById("printedit-format-verticalalign-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("verticalalign","baseline,sub,super,bottom,middle,top"); },false);
    document.getElementById("printedit-format-verticalalign-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("verticalalign","baseline,sub,super,bottom,middle,top"); },false);
    
    document.getElementById("printedit-format-textalign").addEventListener("click",function() { cmdFormatToggleState("textalign"); },false);
    document.getElementById("printedit-format-textalign-value").addEventListener("click",function() { cmdFormatResetValue("textalign","left"); },false);
    document.getElementById("printedit-format-textalign-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("textalign","left,center,right"); },false);
    document.getElementById("printedit-format-textalign-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("textalign","left,center,right"); },false);
    
    document.getElementById("printedit-format-textdecoration").addEventListener("click",function() { cmdFormatToggleState("textdecoration"); },false);
    document.getElementById("printedit-format-textdecoration-value").addEventListener("click",function() { cmdFormatResetValue("textdecoration","none"); },false);
    document.getElementById("printedit-format-textdecoration-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("textdecoration","none,under,over,through"); },false);
    document.getElementById("printedit-format-textdecoration-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("textdecoration","none,under,over,through"); },false);
    
    document.getElementById("printedit-format-texttransform").addEventListener("click",function() { cmdFormatToggleState("texttransform"); },false);
    document.getElementById("printedit-format-texttransform-value").addEventListener("click",function() { cmdFormatResetValue("texttransform","none"); },false);
    document.getElementById("printedit-format-texttransform-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("texttransform","none,capital,lower,upper"); },false);
    document.getElementById("printedit-format-texttransform-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("texttransform","none,capital,lower,upper"); },false);
    
    document.getElementById("printedit-format-liststyle").addEventListener("click",function() { cmdFormatToggleState("liststyle"); },false);
    document.getElementById("printedit-format-liststyle-value").addEventListener("click",function() { cmdFormatResetValue("liststyle","none"); },false);
    document.getElementById("printedit-format-liststyle-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("liststyle","none,disc,square"); },false);
    document.getElementById("printedit-format-liststyle-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("liststyle","none,disc,square"); },false);
    
    document.getElementById("printedit-format-tableborder").addEventListener("click",function() { cmdFormatToggleState("tableborder"); },false);
    document.getElementById("printedit-format-tableborder-value").addEventListener("click",function() { cmdFormatResetValue("tableborder","collapse"); },false);
    document.getElementById("printedit-format-tableborder-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("tableborder","collapse,2px,4px,6px,8px,10px,15px,20px,25px,30px"); },false);
    document.getElementById("printedit-format-tableborder-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("tableborder","collapse,2px,4px,6px,8px,10px,15px,20px,25px,30px"); },false);
    
    document.getElementById("printedit-format-pagebreak").addEventListener("click",function() { cmdFormatToggleState("pagebreak"); },false);
    document.getElementById("printedit-format-pagebreak-value").addEventListener("click",function() { cmdFormatResetValue("pagebreak","before"); },false);
    document.getElementById("printedit-format-pagebreak-prevvalue").addEventListener("mousedown",function() { cmdFormatPrevValue("pagebreak","before,after,\u00ACinside"); },false);
    document.getElementById("printedit-format-pagebreak-nextvalue").addEventListener("mousedown",function() { cmdFormatNextValue("pagebreak","before,after,\u00ACinside"); },false);
    
    document.getElementById("printedit-format-restrict").addEventListener("click",function() { cmdFormatRestrictChange(); },false);
    
    document.getElementById("printedit-format-toggleprops").addEventListener("click",function() { cmdFormatToggleProps(); },false);
    document.getElementById("printedit-format-resetvalues").addEventListener("click",function() { cmdFormatResetValues(); },false);
    document.getElementById("printedit-format-apply").addEventListener("click",function() { cmdFormatApply(false); },false);
    document.getElementById("printedit-format-okay").addEventListener("click",function() { cmdFormatApply(true); },false);
    
    /* Text panel event listeners */
    
    document.getElementById("printedit-text-close").addEventListener("click",function() { closePopup(); },false);

    document.getElementById("printedit-text-textarea").addEventListener("input",function() { cmdTextInput(); },false);
    document.getElementById("printedit-text-insert-type").addEventListener("change",function() { cmdTextType(); },false);
    document.getElementById("printedit-text-apply").addEventListener("click",function() { cmdTextApply(false); },false);
    document.getElementById("printedit-text-okay").addEventListener("click",function() { cmdTextApply(true); },false);
    
    /* Add help panel event listeners */
    
    document.getElementById("printedit-help-radio-button").addEventListener("click",function() { cmdHelpShowPane("button"); },false);
    document.getElementById("printedit-help-radio-toolbar").addEventListener("click",function() { cmdHelpShowPane("toolbar"); },false);
    document.getElementById("printedit-help-radio-mouse").addEventListener("click",function() { cmdHelpShowPane("mouse"); },false);
    document.getElementById("printedit-help-radio-keyboard").addEventListener("click",function() { cmdHelpShowPane("keyboard"); },false);
    document.getElementById("printedit-help-close").addEventListener("click",function() { closePopup(); },false);
    
    /* Window keydown listener */
    
    window.addEventListener("keydown",onKeyDown,true);
    
    /* Storage changed listener */

    chrome.storage.onChanged.addListener(
    function(changes,areaName)
    {
        var key;
        
        /* Reload options */
        
        beforeAdjust = object["options-before-adjust"];
        hideSelect = object["options-hide-select"];
        undoReinstate = object["options-undo-reinstate"];
        closeDoNotAsk = object["options-close-donotask"];
        unloadAskOnly = object["options-unload-askonly"];
        textpiecesBreaks = object["options-textpieces-breaks"];
            
        purgeDeleted = object["options-purge-deleted"];
        purgeHidden = object["options-purge-hidden"];
        purgeStyles = object["options-purge-styles"];
            
        selectBorder = object["options-select-border"];
        selectArea = object["options-select-area"];
        captureSnap = object["options-capture-snap"];
       
        /* Re-draw highlight and select boxes */
        
        if ("options-select-border" in changes || "options-select-area" in changes)
        {
            showHighlightBox();
            showSelectBoxes();
        }
        
        /* Enable or disable line breaks */
        
        if ("options-textpieces-breaks" in changes)
        {
            if (document.getElementById("printedit-textpieces").hasAttribute("checked"))
            {
                disableTextPieces(document.body);
                enableTextPieces(document.body,textpiecesBreaks);
            }
        }
    });
            
    /* Message received listener */

    chrome.runtime.onMessage.addListener(
    function(message,sender,sendResponse)
    {
        /* Messages from background page */
        
        switch (message.type)
        {
            case "startEditing":
                
                startEditing();
                resumeEditing();
                
                break;
                
            case "resumeEditing":
                
                resumeEditing();
                
                break;
                
            case "suspendEditing":
                
                suspendEditing();
                
                break;
                
            case "saveLoadSuccess":
                
                cmdSaveLoadSuccess(message.index,message.contenttype,message.content);
                
                break;
                
            case "saveLoadFailure":
                
                cmdSaveLoadFailure();
                
                break;
                
            case "saveDone":
                
                cmdSaveDone(message.cancel);
                
                break;
                
            case "print":
                
                cmdPrint();
                
                break;
                
            case "removeBeforeUnload":
                
                window.onbeforeunload = null;
                
                break;
        }
    });
}

/************************************************************************/

/* Wait for web page load before editing */

function waitBeforeEditing()
{
    if (document.readyState == "complete")
    {
        window.setTimeout(
        function()
        {
            /* Enable edit button */
            
            document.getElementById("printedit-edit").removeAttribute("disabled","");
            
            /* Display activated badge */
            
            chrome.runtime.sendMessage({ type: "activatedBadge" });
            
            /* Set current mode to initialized */
            
            curMode = 0;
            
            /* Start editing - if web page adjustment not required before editing */
            
            if (!beforeAdjust) cmdEdit();
        },50);
    }
    else
    {
        window.addEventListener("load",
        function(event)
        {
            /* Enable edit button */
            
            document.getElementById("printedit-edit").removeAttribute("disabled","");
            
            /* Display activated badge */
            
            chrome.runtime.sendMessage({ type: "activatedBadge" });
            
            /* Set current mode to initialized */
            
            curMode = 0;
            
            /* Start editing - if web page adjustment not required before editing */
            
            if (!beforeAdjust) cmdEdit();
        },false);
    }
}

/************************************************************************/

/* Initiated from Print Edit toolbar Edit button */

function startEditing()
{
    var i,selection,retain,version;
    
    /* Remove window keydown listener */
    
    window.removeEventListener("keydown",onKeyDown,true);
    
    /* Load format panel settings from local storage */
    
    chrome.storage.local.get(null,
    function(object)
    {
        /* Format Properties - States */
        
        document.getElementById("printedit-format-left").checked = object["format-left"];
        document.getElementById("printedit-format-top").checked = object["format-top"];
        document.getElementById("printedit-format-width").checked = object["format-width"];
        document.getElementById("printedit-format-height").checked = object["format-height"];
        document.getElementById("printedit-format-margin-h").checked = object["format-margin-h"];
        document.getElementById("printedit-format-margin-v").checked = object["format-margin-v"];
        document.getElementById("printedit-format-padding-h").checked = object["format-padding-h"];
        document.getElementById("printedit-format-padding-v").checked = object["format-padding-v"];
        document.getElementById("printedit-format-border").checked = object["format-border"];
        document.getElementById("printedit-format-outline").checked = object["format-outline"];
        document.getElementById("printedit-format-overflow").checked = object["format-overflow"];
        document.getElementById("printedit-format-float").checked = object["format-float"];
        document.getElementById("printedit-format-clear").checked = object["format-clear"];
        
        document.getElementById("printedit-format-color").checked = object["format-color"];
        document.getElementById("printedit-format-background").checked = object["format-background"];
        document.getElementById("printedit-format-fontfamily").checked = object["format-fontfamily"];
        document.getElementById("printedit-format-fontstyle").checked = object["format-fontstyle"];
        document.getElementById("printedit-format-fontsize").checked = object["format-fontsize"];
        document.getElementById("printedit-format-lineheight").checked = object["format-lineheight"];
        document.getElementById("printedit-format-verticalalign").checked = object["format-verticalalign"];
        document.getElementById("printedit-format-textalign").checked = object["format-textalign"];
        document.getElementById("printedit-format-textdecoration").checked = object["format-textdecoration"];
        document.getElementById("printedit-format-texttransform").checked = object["format-texttransform"];
        document.getElementById("printedit-format-liststyle").checked = object["format-liststyle"];
        document.getElementById("printedit-format-tableborder").checked = object["format-tableborder"];
        document.getElementById("printedit-format-pagebreak").checked = object["format-pagebreak"];

        /* Format Properties - Values */

        document.getElementById("printedit-format-left-value").textContent = object["format-left-value"];
        document.getElementById("printedit-format-top-value").textContent = object["format-top-value"];
        document.getElementById("printedit-format-width-value").textContent = object["format-width-value"];
        document.getElementById("printedit-format-height-value").textContent = object["format-height-value"];
        document.getElementById("printedit-format-margin-h-value").textContent = object["format-margin-h-value"];
        document.getElementById("printedit-format-margin-v-value").textContent = object["format-margin-v-value"];
        document.getElementById("printedit-format-padding-h-value").textContent = object["format-padding-h-value"];
        document.getElementById("printedit-format-padding-v-value").textContent = object["format-padding-v-value"];
        document.getElementById("printedit-format-border-value").textContent = object["format-border-value"];
        document.getElementById("printedit-format-outline-value").textContent = object["format-outline-value"];
        document.getElementById("printedit-format-overflow-value").textContent = object["format-overflow-value"];
        document.getElementById("printedit-format-float-value").textContent = object["format-float-value"];
        document.getElementById("printedit-format-clear-value").textContent = object["format-clear-value"];
        
        document.getElementById("printedit-format-color-value").textContent = object["format-color-value"];
        document.getElementById("printedit-format-background-value").textContent = object["format-background-value"];
        document.getElementById("printedit-format-fontfamily-value").textContent = object["format-fontfamily-value"];
        document.getElementById("printedit-format-fontstyle-value").textContent = object["format-fontstyle-value"];
        document.getElementById("printedit-format-fontsize-value").textContent = object["format-fontsize-value"];
        document.getElementById("printedit-format-lineheight-value").textContent = object["format-lineheight-value"];
        document.getElementById("printedit-format-verticalalign-value").textContent = object["format-verticalalign-value"];
        document.getElementById("printedit-format-textalign-value").textContent = object["format-textalign-value"];
        document.getElementById("printedit-format-textdecoration-value").textContent = object["format-textdecoration-value"];
        document.getElementById("printedit-format-texttransform-value").textContent = object["format-texttransform-value"];
        document.getElementById("printedit-format-liststyle-value").textContent = object["format-liststyle-value"];
        document.getElementById("printedit-format-tableborder-value").textContent = object["format-tableborder-value"];
        document.getElementById("printedit-format-pagebreak-value").textContent = object["format-pagebreak-value"];
    
        /* Format Controls */
    
        document.getElementById("printedit-format-subelements").checked = object["format-subelements"];
        document.getElementById("printedit-format-important").checked = object["format-important"];
        document.getElementById("printedit-format-restrict").checked = object["format-restrict"];
        document.getElementById("printedit-format-restrict-type").value = object["format-restrict-type"];
        
        /* Load format show properties setting from local storage */
        
        if (object["format-showprops"]) cmdFormatToggleProps();
        
        /* Load text insert type setting from local storage */
        
        textInsertType = object["text-insert-type"];
        
        /* Load help show pane setting from local storage */
        
        document.getElementById("printedit-help-radio-" + object["help-showpane"]).checked = true;
        
        document.getElementById("printedit-help-" + object["help-showpane"]).setAttribute("printedit-showpane","");
    });
    
    /* Prevent browser leaving or reloading page (including meta refresh) */
    
    window.onbeforeunload =
    function confirmExit(event)
    {
        /* Can be reload, navigation, tab close or window close */
        
        if (cmdNumber == saveCmdNumber)
        {
            if (unloadAskOnly) return;
        }
        
        return "confirm exit";  /* any string shows default message */
    };
    
    /* Prevent browser page loads and refreshes (block all HTTP requests) */
    
    chrome.runtime.sendMessage({ type: "blockHTTP" });
    
    /* Clear any text selections */
    
    selection = window.getSelection();
    if (selection.rangeCount > 0) selection.removeAllRanges();
    
    /* Initialise for print edit mode */
    
    cmdNumber = 0;
    selectCount[0] = 0;
    selectElement[0] = new Array();
    
    saveCmdNumber = 0;
    
    /* Retain edits from previous session ? */
    
    retain = false;
    version = "";
    
    if (document.body.hasAttribute("printedit-previous"))
    {
        version = document.body.getAttribute("printedit-previous");
        if (version == "") version = "7.0";  /* first version to allow saving edited web page */
        
        if (confirm("PRINT EDIT\n\n" +
                    "Do you want to retain the edits from the previous session?")) retain = true;
    }
    
    /* Set previously edited attribute with compatibility version for re-editing */
    
    document.body.setAttribute("printedit-previous","16.1");
    
    /* Prepare elements for editing */
    
    prepareForEditing(document.body,retain,version);
    
    /* Display using print stylesheets for screen device */
    
    displayForDifferentDevice((document.getElementById("printedit-webstyle").checked)?"screen":"print");
    
    /* Show preview button */
    
    document.getElementById("printedit-edit").style.setProperty("display","none","important");
    document.getElementById("printedit-preview").style.removeProperty("display");
    
    /* Enable other buttons */
    
    enableOtherButtons();
    
    /* Display editing badge */
    
    chrome.runtime.sendMessage({ type: "editingBadge" });
    
    /* Set current mode to suspended */
    
    curMode = 1;
}

/************************************************************************/

/* Called after startEditing and after print or save */

function resumeEditing()
{
    var toolbar,selectIndex,element;
    
    /* Add [EDIT] prefix in tab title */
    
    document.title = "[EDIT] " + document.title;
    
    /* Insert print edit toolbar, menus and panels */
    
    if (guiContainer.parentNode == null)
    {
        document.body.style.setProperty("transform","translateY(31px)","important");
        document.body.style.setProperty("-webkit-transform","translateY(31px)","important");
        
        document.documentElement.appendChild(guiContainer);
    }
    
    /* Enable text pieces and line breaks */
    
    if (document.getElementById("printedit-textpieces").hasAttribute("checked"))
        enableTextPieces(document.body,textpiecesBreaks);
    
    /* Create highlight box and set default highlight element */
    
    highlightBox = createBox("highlightbox");
    highlightElement = document.body;
    
    /* Create and show select boxes */
    
    for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
    {
        element = selectElement[cmdNumber][selectIndex];
        selectBox[selectIndex] = createBox("selectbox");
        
        showBox(selectBox[selectIndex],element,0,0,0,0,0x0,0x0);
    }
    
    /* Enable edit and undo buttons */
    
    if (selectCount[cmdNumber] > 0) enableEditButtons();
    if (cmdNumber > 0) enableUndoButtons();
    
    /* Add window mouse listeners */
    
    window.addEventListener("mouseover",onMouseOver,true);
    window.addEventListener("mouseout",onMouseOut,true);
    window.addEventListener("mousedown",onMouseDown,true);
    window.addEventListener("mousemove",onMouseMove,true);
    window.addEventListener("mouseup",onMouseUp,true);
    
    window.addEventListener("click",onClick,true);
    window.addEventListener("dblclick",onDblClick,true);
    window.addEventListener("contextmenu",onContextMenu,true);
    
    /* Add window scroll listener */
    
    window.addEventListener("scroll",onScroll,true);
    
    /* Add window resize/zoom listener */
    
    window.addEventListener("resize",onResize,false);
    
    /* Add window keydown listener */
    
    window.addEventListener("keydown",onKeyDown,true);
    
    /* Set current mode to editing */
    
    curMode = 2;
}

/************************************************************************/

/* Called before save or print and from finishEditing */

function suspendEditing()
{
    var toolbar;
    
    /* Remove window keydown listener */
    
    window.removeEventListener("keydown",onKeyDown,true);
    
    /* Remove window resize/zoom listener */
    
    window.removeEventListener("resize",onResize,false);
    
    /* Remove window scroll listener */
    
    window.removeEventListener("scroll",onScroll,true);
    
    /* Remove window mouse listeners */
    
    window.removeEventListener("mouseover",onMouseOver,true);
    window.removeEventListener("mouseout",onMouseOut,true);
    window.removeEventListener("mousedown",onMouseDown,true);
    window.removeEventListener("mousemove",onMouseMove,true);
    window.removeEventListener("mouseup",onMouseUp,true);
    
    window.removeEventListener("click",onClick,true);
    window.removeEventListener("dblclick",onDblClick,true);
    window.removeEventListener("contextmenu",onContextMenu,true);
    
    /* Disable edit and undo buttons */
    
    disableEditButtons();
    disableUndoButtons();
    
    /* Destroy select boxes */
    
    destroySelectBoxes();
    
    /* Destroy highlight box and release highlight element */
    
    destroyBox(highlightBox);
    highlightBox = null;
    highlightElement = null;
    
    /* Disable text pieces and line breaks */
    
    if (document.getElementById("printedit-textpieces").hasAttribute("checked"))
        disableTextPieces(document.body);
    
    /* Remove print edit toolbar, menus and panels */
    
    document.body.style.removeProperty("transform");
    document.body.style.removeProperty("-webkit-transform");
    
    document.documentElement.removeChild(guiContainer);
    
    /* Remove [EDIT] prefix in tab title */
    
    document.title = document.title.substr(7);
    
    /* Set current mode to suspended */
    
    curMode = 1;
}

/************************************************************************/

/* Initiated from Print Edit toolbar Close button */

function finishEditing()
{
    /* Close without saving edited web page ? */
    
    if (cmdNumber != saveCmdNumber)
    {
        if (!closeDoNotAsk)
        {
            if (!confirm("PRINT EDIT\n\n" +
                         "Do you want to close this tab without saving the edited page?")) return;
        }
    }
    
    /* Disable other buttons */
    
    disableOtherButtons();
    
    /* Leave editing mode */
    
    if (curMode == 2) suspendEditing();
    
    /* Undo all commands */
    
    while (cmdNumber >= 0)
    {
        deselectElements();
        cmdNumber--;
    }
    
    cmdNumber = -1;
    
    undoCommands(document.body);
    
    cmdNumber = 0;
    
    saveCmdNumber = 0;
    
    /* Allow leaving or reloading page */
    
    window.onbeforeunload = null;
    
    /* Set current mode to initialized */
    
    curMode = 0;
}

/************************************************************************/

/* Toolbar button functions */

function enableEditButtons()
{
    document.getElementById("printedit-select").setAttribute("disabled","");
    document.getElementById("printedit-select").style.setProperty("display","none","important");
    document.getElementById("printedit-deselect").removeAttribute("disabled");
    document.getElementById("printedit-deselect").style.removeProperty("display");
    
    document.getElementById("printedit-hide").removeAttribute("disabled");
    document.getElementById("printedit-delete").removeAttribute("disabled");
    document.getElementById("printedit-hideexcept").removeAttribute("disabled");
    document.getElementById("printedit-deleteexcept").removeAttribute("disabled");
    document.getElementById("printedit-format").removeAttribute("disabled");
    if (selectCount[cmdNumber] == 1) document.getElementById("printedit-text").removeAttribute("disabled");
    else document.getElementById("printedit-text").setAttribute("disabled","");
}

function disableEditButtons()
{
    document.getElementById("printedit-select").removeAttribute("disabled");
    document.getElementById("printedit-select").style.removeProperty("display");
    document.getElementById("printedit-deselect").setAttribute("disabled","");
    document.getElementById("printedit-deselect").style.setProperty("display","none","important");
    
    document.getElementById("printedit-hide").setAttribute("disabled","");
    document.getElementById("printedit-delete").setAttribute("disabled","");
    document.getElementById("printedit-hideexcept").setAttribute("disabled","");
    document.getElementById("printedit-deleteexcept").setAttribute("disabled","");
    document.getElementById("printedit-format").setAttribute("disabled","");
    document.getElementById("printedit-text").setAttribute("disabled","");
}

function enableUndoButtons()
{
    document.getElementById("printedit-undo").removeAttribute("disabled");
    document.getElementById("printedit-undoall").removeAttribute("disabled");
}

function disableUndoButtons()
{
    document.getElementById("printedit-undo").setAttribute("disabled","");
    document.getElementById("printedit-undoall").setAttribute("disabled","");
}

function enableOtherButtons()
{
    document.getElementById("printedit-select").removeAttribute("disabled");
    document.getElementById("printedit-save").removeAttribute("disabled");
    document.getElementById("printedit-textpieces").removeAttribute("disabled");
    document.getElementById("printedit-webstyle").removeAttribute("disabled");
}

function disableOtherButtons()
{
    document.getElementById("printedit-select").setAttribute("disabled","");
    document.getElementById("printedit-save").setAttribute("disabled","");
    document.getElementById("printedit-textpieces").setAttribute("disabled","");
    document.getElementById("printedit-webstyle").setAttribute("disabled","");
}

/************************************************************************/

/* Toolbar and Popup utility functions */

function insideToolbar(element)
{
    while (element != null && element != document.getElementById("printedit-toolbar")) element = element.parentNode;
    
    return (element != null);
}

function insidePopup(element)
{
    while (element != null && element != document.getElementById("printedit-inspect-panel")
                           && element != document.getElementById("printedit-select-menu")
                           && element != document.getElementById("printedit-deleteexcept-menu")
                           && element != document.getElementById("printedit-format-panel")
                           && element != document.getElementById("printedit-save-menu")
                           && element != document.getElementById("printedit-text-panel")
                           && element != document.getElementById("printedit-help-panel")) element = element.parentNode;
                           
    return (element != null);
}

function adjustToolbar()
{
    if (document.body.scrollHeight > window.innerHeight) document.getElementById("printedit-spacer-1").setAttribute("printedit-scrollbar","");
    else document.getElementById("printedit-spacer-1").removeAttribute("printedit-scrollbar","");
}

function closePopup()
{
    document.getElementById("printedit-select").removeAttribute("checked");        
    document.getElementById("printedit-deleteexcept").removeAttribute("checked");        
    document.getElementById("printedit-format").removeAttribute("checked");        
    document.getElementById("printedit-text").removeAttribute("checked");        
    document.getElementById("printedit-save").removeAttribute("checked");        
    document.getElementById("printedit-help").removeAttribute("checked");        
    
    document.getElementById("printedit-inspect-panel").style.setProperty("display","none","important");
    document.getElementById("printedit-select-menu").style.setProperty("display","none","important");
    document.getElementById("printedit-deleteexcept-menu").style.setProperty("display","none","important");
    document.getElementById("printedit-format-panel").style.setProperty("display","none","important");
    document.getElementById("printedit-text-panel").style.setProperty("display","none","important");
    document.getElementById("printedit-save-menu").style.setProperty("display","none","important");
    document.getElementById("printedit-help-panel").style.setProperty("display","none","important");
    
    popupOpen = 0;
}

/************************************************************************/

/* Box object constructor */

function Box()
{
    this.container = null;
    this.rects = new Array();
}

/************************************************************************/

/* Create and display boxes functions */

function createBox(type)
{
    var box,zindex;
    
    if (type == "capturebox") zindex = "10000002";
    else if (type == "highlightbox") zindex = "10000001";
    else if (type == "selectbox") zindex = "10000000";
    
    box = new Box();
    
    box.container = document.createElement("div");
    box.container.setAttribute("class","printedit-" + type);
    box.container.style.setProperty("position","absolute","important");
    box.container.style.setProperty("width","0px","important");
    box.container.style.setProperty("height","0px","important");
    box.container.style.setProperty("left","0px","important");
    box.container.style.setProperty("top","0px","important");
    box.container.style.setProperty("background-color","transparent","important");
    box.container.style.setProperty("z-index",zindex,"important");
    box.container.style.setProperty("pointer-events","none","important");
    guiContainer.appendChild(box.container);
    
    return box;
}

function destroyBox(box)
{
    var i;
    
    for (i = 0; i < box.rects.length; i++) box.container.removeChild(box.rects[i]);
    box.rects.length = 0;

    guiContainer.removeChild(box.container);
    box.container = null;
}

function showBox(box,element,width,height,left,top,pattern,anchor)
{
    var i,zindex,borderColor,borderWidth,borderStyle,areaColor;
    var boundingRect = new Object();
    var rects = new Array();
    
    if (box.container.getAttribute("class") == "printedit-capturebox")
    {
        zindex = "10000002";
        borderWidth = 1;
        borderColor = "#0000FF";
    }
    else if (box.container.getAttribute("class") == "printedit-highlightbox")
    {
        zindex = "10000001";
        if (selectBorder) borderWidth = 2; else borderWidth = 1;
        borderStyle = "solid";
        borderColor = "#FF0000";
        areaColor = "rgba(0,0,0,0.0)";
    }
    else if (box.container.getAttribute("class") == "printedit-selectbox")
    {
        zindex = "10000000";
        if (selectBorder) borderWidth = 2; else borderWidth = 1;
        borderStyle = "dashed";
        borderColor = "#FF0000";
        if (selectArea) areaColor = "rgba(255,0,0,0.1)"; else areaColor = "rgba(0,0,0,0.0)";
    }
    
    for (i = 0; i < box.rects.length; i++) box.container.removeChild(box.rects[i]);
    box.rects.length = 0;
    
    if (box.container.getAttribute("class") == "printedit-capturebox")
    {
        box.container.style.setProperty("width",width + "px","important");
        box.container.style.setProperty("height",height + "px","important");
        box.container.style.setProperty("left",left + "px","important");
        box.container.style.setProperty("top",top + "px","important");
        
        box.rects[0] = document.createElement("div");
        
        box.rects[0].style.setProperty("position","absolute","important");
        box.rects[0].style.setProperty("width",width + "px","important");
        box.rects[0].style.setProperty("height",height + "px","important");
        box.rects[0].style.setProperty("left","0px","important");
        box.rects[0].style.setProperty("top","0px","important");
        box.rects[0].style.setProperty("z-index",zindex,"important");
        box.rects[0].style.setProperty("pointer-events","none","important");
        box.rects[0].style.setProperty("border",borderWidth + "px solid " + borderColor,"important");
        
        if (pattern == 0) box.rects[0].style.setProperty("background","url(" + captureAll + ") repeat","important");
        else if (pattern == 1)
        {
            if (anchor == 0 || anchor == 2) box.rects[0].style.setProperty("background","url(" + captureGraphicsRight + ") repeat left","important");
            else box.rects[0].style.setProperty("background","url(" + captureGraphicsLeft + ") repeat right","important");
        }
        else if (pattern >= 2) 
        {
            if (anchor == 0 || anchor == 1) box.rects[0].style.setProperty("background","url(" + captureDeleteBottom + ") repeat top","important");
            else box.rects[0].style.setProperty("background","url(" + captureDeleteTop + ") repeat bottom","important");
        }
        
        box.container.appendChild(box.rects[0]);
    }
    else
    {
        boundingRect = element.getBoundingClientRect();
        
        width = boundingRect.width;
        height = boundingRect.height;
        left = calculateFrameLeft(element,window)+boundingRect.left;
        top = calculateFrameTop(element,window)+boundingRect.top;
        
        box.container.style.setProperty("width",width + "px","important");
        box.container.style.setProperty("height",height + "px","important");
        box.container.style.setProperty("left",left + "px","important");
        box.container.style.setProperty("top",top + "px","important");
        
        rects = element.getClientRects();
        
        for (i = 0; i < rects.length; i++)
        {
            box.rects[i] = document.createElement("div");
            
            box.rects[i].style.setProperty("position","absolute","important");
            box.rects[i].style.setProperty("width",(rects[i].width-borderWidth*2) + "px","important");
            box.rects[i].style.setProperty("height",(rects[i].height-borderWidth*2) + "px","important");
            if (rects[i].width-borderWidth*2 >= 0) box.rects[i].style.setProperty("left",(rects[i].left-boundingRect.left) + "px","important");
            else box.rects[i].style.setProperty("left",(rects[i].left-boundingRect.left+(rects[i].width-borderWidth*2)/2) + "px","important");
            if (rects[i].height-borderWidth*2 >= 0) box.rects[i].style.setProperty("top",(rects[i].top-boundingRect.top) + "px","important");
            else box.rects[i].style.setProperty("top",(rects[i].top-boundingRect.top+(rects[i].height-borderWidth*2)/2) + "px","important");
            box.rects[i].style.setProperty("z-index",zindex,"important");
            box.rects[i].style.setProperty("pointer-events","none","important");
            box.rects[i].style.setProperty("border",borderWidth + "px " + borderStyle + " " + borderColor,"important");
            box.rects[i].style.setProperty("background-color",areaColor,"important");
            
            box.container.appendChild(box.rects[i]);
        }
    }
}

function hideBox(box)
{
    var i;
    
    for (i = 0; i < box.rects.length; i++) box.container.removeChild(box.rects[i]);
    box.rects.length = 0;
}

/************************************************************************/

/* Highlight management functions */

function showHighlightBox()
{
    if (highlightElement != null)
    {
        showBox(highlightBox,highlightElement,0,0,0,0,0x0,0x0);
        showBox(highlightBox,highlightElement,0,0,0,0,0x0,0x0);  /* occasionally the box sides are in the wrong position after the first call !!! */  /*???*/
    }
}

/************************************************************************/

/* Select management functions */

function showSelectBoxes()
{
    var selectIndex,element;
    
    for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
    {
        element = selectElement[cmdNumber][selectIndex];
        
        showBox(selectBox[selectIndex],element,0,0,0,0,0x0,0x0);
    }
}

function destroySelectBoxes()
{
    var selectIndex;
    
    for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
        destroyBox(selectBox[selectIndex]);
    
    selectBox.length = 0;
}

function deselectElements()
{
    selectCount[cmdNumber] = 0;
    selectElement[cmdNumber].length = 0;
}

function nextCommand()
{
    cmdNumber++;
    selectCount[cmdNumber] = 0;
    selectElement[cmdNumber] = new Array();
}

/************************************************************************/

/* Frame utility functions */

function calculateFrameLeft(element,baseWindow)
{
    var x;
    
    x = element.ownerDocument.defaultView.scrollX;
    
    while (element.ownerDocument.defaultView != baseWindow)
    {
        x -= element.ownerDocument.defaultView.scrollX;
        element = element.ownerDocument.defaultView.frameElement;
        
        if (element.localName == "frame")
        {
            if (isFirefox)
            {
                x += element.offsetLeft;
                return x;
            }
            else
            {
                while (element.parentElement)  /* frame or frameset */
                {
                    x += element.offsetLeft;
                    element = element.parentNode;
                }
            }
            
            return x;
        }
        
        while (element.offsetParent)  /* not body */
        {
            x += element.offsetLeft;
            element = element.offsetParent;
        }
    }
    
    return x;
}

function calculateFrameTop(element,baseWindow)
{
    var y;
    
    y = element.ownerDocument.defaultView.scrollY;
    
    if (baseWindow == window && element.ownerDocument.defaultView != window) y += toolbarHeight;
    
    while (element.ownerDocument.defaultView != baseWindow)
    {
        y -= element.ownerDocument.defaultView.scrollY;
        element = element.ownerDocument.defaultView.frameElement;
        
        if (element.localName == "frame")
        {
            if (isFirefox)
            {
                y += element.offsetTop;
                return y;
            }
            else
            {
                while (element.parentElement)  /* frame or frameset */
                {
                    y += element.offsetTop;
                    element = element.parentNode;
                }
            }
            
            return y;
        }
        
        while (element.offsetParent)  /* not body */
        {
            y += element.offsetTop;
            element = element.offsetParent;
        }
    }
    
    return y;
}

/************************************************************************/

/* Capture select and deselect recursive functions */

function selectCaptured(element)
{
    var i,scrollWidth,scrollHeight,left,right,top,bottom,childLeft,childTop,selectIndex;
    var boundingRect = new Object();
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                selectCaptured(element.contentDocument.body);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        /* Note: clientWidth, clientHeight, scrollWidth, scrollHeight on <body> or <html> are not always correct! */
        
        scrollWidth = captureWindow.document.body.scrollWidth;
        scrollHeight = captureWindow.document.body.scrollHeight;
        
        boundingRect = element.getBoundingClientRect();
        
        left = calculateFrameLeft(element,captureWindow)+boundingRect.left;
        right = left+boundingRect.width;
        top = calculateFrameTop(element,captureWindow)+boundingRect.top;
        bottom = top+boundingRect.height;
        
        if (left < 0) left = 0;
        if (right > scrollWidth-1) right = scrollWidth-1;
        if (top < 0 ) top = 0;
        if (bottom > scrollHeight-1) bottom = scrollHeight-1;
        
        if (left >= captureLeft && right <= captureLeft+captureWidth &&
            top >= captureTop && bottom <= captureTop+captureHeight)
        {
            /* Element within capture area */
            
            for (i = 0; i < element.children.length; i++)
            {
                if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                {
                    boundingRect = element.children[i].getBoundingClientRect();
                    
                    childLeft = calculateFrameLeft(element.children[i],captureWindow)+boundingRect.left;
                    childTop = calculateFrameTop(element.children[i],captureWindow)+boundingRect.top;
                    
                    if (childLeft <= left && childLeft+boundingRect.width >= right &&
                        childTop <= top && childTop+boundingRect.height >= bottom)
                    {
                        /* Child as large as element - don't select element */
                        /* Element would not be de-selectable with highlight box */
                        
                        break;
                    }
                }
            }
            
            if (i >= element.children.length)
            {
                /* Children are all smaller than element - select element */
                
                for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
                {
                    if (selectElement[cmdNumber][selectIndex] == element) return;
                }
                
                selectElement[cmdNumber][selectCount[cmdNumber]] = element;
                selectBox[selectCount[cmdNumber]] = createBox("selectbox");
                
                showBox(selectBox[selectCount[cmdNumber]],element,0,0,0,0,0x0,0x0);
                
                selectCount[cmdNumber]++;
                
                return;
            }
        }
        
        /* Element not within capture area or a child as large as element */
        
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                selectCaptured(element.children[i]);
    }
}

function selectCapturedGraphics(element)
{
    var i,scrollWidth,scrollHeight,left,right,top,bottom,selectIndex;
    var boundingRect = new Object();
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                selectCapturedGraphics(element.contentDocument.body);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        /* Note: clientWidth, clientHeight, scrollWidth, scrollHeight on <body> or <html> are not always correct! */
        
        scrollWidth = captureWindow.document.body.scrollWidth;
        scrollHeight = captureWindow.document.body.scrollHeight;
        
        boundingRect = element.getBoundingClientRect();
        
        left = calculateFrameLeft(element,captureWindow)+boundingRect.left;
        right = left+boundingRect.width;
        top = calculateFrameTop(element,captureWindow)+boundingRect.top;
        bottom = top+boundingRect.height;
        
        if (left < 0) left = 0;
        if (right > scrollWidth-1) right = scrollWidth-1;
        if (top < 0 ) top = 0;
        if (bottom > scrollHeight-1) bottom = scrollHeight-1;
        
        if (left >= captureLeft && right <= captureLeft+captureWidth &&
            top >= captureTop && bottom <= captureTop+captureHeight)
        {
            /* Element within capture area */
            
            if (element.localName == "img" || element.localName == "map" ||
                element.getAttribute("class") == "printedit-object" || element.getAttribute("class") == "printedit-embed")
            {
                /* Element is graphic - select element */
                
                for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
                {
                    if (selectElement[cmdNumber][selectIndex] == element) return;
                }
                
                selectElement[cmdNumber][selectCount[cmdNumber]] = element;
                selectBox[selectCount[cmdNumber]] = createBox("selectbox");
                
                showBox(selectBox[selectCount[cmdNumber]],element,0,0,0,0,0x0,0x0);
                
                selectCount[cmdNumber]++;
                
                return;
            }
        }
        
        /* Element not within capture area */
        
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                selectCapturedGraphics(element.children[i]);
    }
}

function deselectCaptured(element)
{
    var i,scrollWidth,scrollHeight,left,right,top,bottom,selectIndex;
    var boundingRect = new Object();
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                deselectCaptured(element.contentDocument.body);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        /* Note: clientWidth, clientHeight, scrollWidth, scrollHeight on <body> or <html> are not always correct! */
        
        scrollWidth = captureWindow.document.body.scrollWidth;
        scrollHeight = captureWindow.document.body.scrollHeight;
        
        boundingRect = element.getBoundingClientRect();
        
        left = calculateFrameLeft(element,captureWindow)+boundingRect.left;
        right = left+boundingRect.width;
        top = calculateFrameTop(element,captureWindow)+boundingRect.top;
        bottom = top+boundingRect.height;
        
        if (left < 0) left = 0;
        if (right > scrollWidth-1) right = scrollWidth-1;
        if (top < 0 ) top = 0;
        if (bottom > scrollHeight-1) bottom = scrollHeight-1;
        
        if (left >= captureLeft && right <= captureLeft+captureWidth &&
            top >= captureTop && bottom <= captureTop+captureHeight)
        {
            /* Element within capture area */
                
            for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
            {
                if (selectElement[cmdNumber][selectIndex] == element)
                {
                    destroyBox(selectBox[selectIndex]);
                    
                    for (i = selectIndex; i < selectCount[cmdNumber]-1; i++)
                    {
                        selectElement[cmdNumber][i] = selectElement[cmdNumber][i+1];
                        selectBox[i] = selectBox[i+1];
                    }
                    
                    selectCount[cmdNumber]--;
                    selectElement[cmdNumber].length--;
                    selectBox.length--;
                    selectIndex--;
                }
            }
        }
        
        /* Element not within capture area */
        
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                deselectCaptured(element.children[i]);
    }
}

/************************************************************************/

/* Prepare elements for editing function */

function prepareForEditing(element,retain,version)
{
    var i,style,width,height,div,frame,parent;
    var styleFields = new Array();
    var boundingRect = new Object();
    var textFields = new Array();
    
    if (element.hasAttribute("printedit-text-inserted"))
    {
        if (!retain)  /* remove any previous text inserts */
        {
            element.parentNode.removeChild(element);
            
            return true;  /* element removed */
        }
        
        element.removeAttribute("printedit-text-inserted");  /* remove inserted text undo information from previously edited document */
    }
    
    if (element.hasAttribute("printedit-text"))
    {
        if (!retain)  /* remove any previous text edits */
        {
            if (version == "7.0") textFields = element.getAttribute("printedit-text").split("\u0000");  /* 15.0 */
            else if (version == "15.1" || version == "16.1") textFields = element.getAttribute("printedit-text").split("\u204B");  /* >= 15.1 */
            
            element.textContent = textFields[textFields.length-1];
        }
        
        element.removeAttribute("printedit-text");  /* remove edited text undo information from previously edited document */
        
        if (version == "7.0" || version == "15.1") element.setAttribute("printedit-text-wrapper","");  /* 15.0 - 16.0 */
    }
    
    if (element.hasAttribute("printedit-style"))
    {
        if (!retain)  /* remove any previous style edits */
        {
            if (version == "7.0") styleFields = element.getAttribute("printedit-style").split("|");  /* 7.0 - 15.0 */  
            else if (version == "15.1" || version == "16.1") styleFields = element.getAttribute("printedit-style").split("\u204B");  /* >= 15.1 */
            
            element.style.cssText = styleFields[styleFields.length-1];
        }
        
        element.removeAttribute("printedit-style");  /* remove edited style undo information from previously edited document */
    }
    
    if (element.localName == "body")
    {
        /* Prevent animations and transitions on all elements */
        
        style = document.createElement("style");
        style.id = "printedit-style-animations";
        style.type = "text/css";
        style.textContent = styleAnimations;
        element.ownerDocument.head.appendChild(style);
    }
    else if (element.localName == "input" || element.localName == "button" || element.localName == "textarea" ||
             element.localName == "select" || element.localName == "optgroup" || element.localName == "option")
    {
        /* Disable <input>/<button>/<textarea>/<select>/<optgroup>/<option> elements by making read-only */
        
        element.setAttribute("readonly","readonly");
    }
    else if (element.localName == "object" || element.localName == "embed")
    {
        /* Disable <object> or <embed> element by replacing with <div> element */
        
        boundingRect = element.getBoundingClientRect();
        width = boundingRect.width;
        height = boundingRect.height;
        if (element.hasAttribute("width")) width = element.getAttribute("width");
        if (element.hasAttribute("height")) height = element.getAttribute("height");
        if (element.style.getPropertyValue("width") != null) { width = element.style.getPropertyValue("width"); width = width.substr(0,width.length-2); }
        if (element.style.getPropertyValue("height") != null) { height = element.style.getPropertyValue("height"); height = height.substr(0,height.length-2); }
        
        div = document.createElement("div");
        div.setAttribute("class","printedit-" + element.localName);
        div.style.setProperty("width",width+"px","important");
        div.style.setProperty("height",height+"px","important");
        div.style.setProperty("background","repeating-linear-gradient(0deg, #E0E0E0 0px, #E0E0E0 5px, white 5px, white 10px)","important");
        
        parent = element.parentNode;
        parent.insertBefore(div,element);
        parent.removeChild(element);
        
        return false;  /* element replaced */
    }
   
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                prepareForEditing(element.contentDocument.body,retain,version);
        }
        catch (e)  /* attempting cross-domain web page access */
        {
            /* Disable <iframe> or <frame> element by replacing with cloned element without src attribute */
            
            frame = element.cloneNode(true);
            frame.setAttribute("src","");
            frame.style.setProperty("background","repeating-linear-gradient(0deg, #E0E0E0 0px, #E0E0E0 5px, white 5px, white 10px)","important");
            
            parent = element.parentNode;
            parent.insertBefore(frame,element);
            parent.removeChild(element);
        }
    }
    else
    {
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                if (prepareForEditing(element.children[i],retain,version)) i--;
    }
    
    return false;  /* element not removed */
}

/************************************************************************/

/* Hide/Delete/Format recursive edit functions */

function hideExcept(element)
{
    var i,keepNode,selectIndex;
    
    keepNode = false;
    
    for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
    {
        if (selectElement[cmdNumber][selectIndex] == element) return true;
    }
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                if (hideExcept(element.contentDocument.body)) keepNode = true;
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                if (hideExcept(element.children[i])) keepNode = true;
    }
    
    if (!keepNode)
    {
        if (!element.hasAttribute("printedit-style")) element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText);
        else element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText + "\u204B" + element.getAttribute("printedit-style"));

        if (hideSelect)
        {
            element.style.setProperty("opacity","0","important");
        }
        else
        {
            element.style.setProperty("visibility","hidden","important");
        }
    }
    else
    {
        if (!element.hasAttribute("printedit-style")) element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText);
        else element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText + "\u204B" + element.getAttribute("printedit-style"));
        
        /* Reset non-inherited properties for higher level elements */
        
        element.style.setProperty("border","0","important");
        element.style.setProperty("outline","0","important");
        element.style.setProperty("overflow","visible","important");
        
        element.style.setProperty("background","white","important");
    }
    
    return keepNode;
}

function deleteExcept(element,restriction)
{
    var i,selectIndex,keepNode;
    
    keepNode = false;
    
    for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
    {
        if (selectElement[cmdNumber][selectIndex] == element)  /* selected element */
        {
            if (selectCount[cmdNumber] == 1)  /* remove margin if only one selected element */
            {
                if (!element.hasAttribute("printedit-style")) element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText);
                else element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText + "\u204B" + element.getAttribute("printedit-style"));
                
                element.style.setProperty("margin","0","important");
            }
            
            if (restriction >= 1)  /* without float layout*/
            {
                element.style.setProperty("float","none","important");
            }
            
            return true;
        }
    }
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                if (deleteExcept(element.contentDocument.body,restriction)) keepNode = true;
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                if (deleteExcept(element.children[i],restriction)) keepNode = true;
    }
    
    if (!keepNode)
    {
        if (!element.hasAttribute("printedit-style")) element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText);
        else element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText + "\u204B" + element.getAttribute("printedit-style"));
        
        element.style.setProperty("display","none","important");
    }
    else  /* higher level element */
    {
        if (!element.hasAttribute("printedit-style")) element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText);
        else element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText + "\u204B" + element.getAttribute("printedit-style"));
        
        /* Reset non-inherited properties for higher level elements */
        
        if (restriction == 2)  /* unrestricted layout */
        {
            element.style.setProperty("min-width","0","important");
            element.style.setProperty("min-height","0","important");
            element.style.setProperty("max-width","none","important");
            element.style.setProperty("max-height","none","important");
            
            element.style.setProperty("left","auto","important");
            element.style.setProperty("top","auto","important");
            element.style.setProperty("width","auto","important");
            element.style.setProperty("height","auto","important");
        }
        
        if (restriction >= 1)  /* without float layout*/
        {
            element.style.setProperty("float","none","important");
        }
        
        /* restricted layout */
        
        element.style.setProperty("margin","0","important");
        element.style.setProperty("padding","0","important");
        element.style.setProperty("border","0","important");
        element.style.setProperty("outline","0","important");
        
        element.style.setProperty("background","white","important");
        element.style.setProperty("overflow","visible","important");
    }
    
    return keepNode;
}

function formatElement(element)
{
    var i,important;
    
    if (element.hasAttribute("printedit-style") && +element.getAttribute("printedit-style").split("\u204B")[0] == cmdNumber) return;  /* must be sub-element of multiple selected elements */
    
    if (!document.getElementById("printedit-format-restrict").checked || document.getElementById("printedit-format-restrict-type").value == element.localName)
    {
        if (!element.hasAttribute("printedit-style")) element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText);
        else element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText + "\u204B" + element.getAttribute("printedit-style"));
        
        if (document.getElementById("printedit-format-important").checked) important = "important"; else important = "";
        
        /* Based on Eric Meyer's CSS Reset and Aardvark's features and some useful additions */
        
        if (document.getElementById("printedit-format-left").checked) element.style.setProperty("left",document.getElementById("printedit-format-left-value").textContent,important);
        if (document.getElementById("printedit-format-top").checked) element.style.setProperty("top",document.getElementById("printedit-format-top-value").textContent,important);
        if (document.getElementById("printedit-format-width").checked)
        {
            element.style.setProperty("width",document.getElementById("printedit-format-width-value").textContent,important);
            element.style.setProperty("min-width","0",important);
            element.style.setProperty("max-width","none",important);
        }
        if (document.getElementById("printedit-format-height").checked)
        {
            element.style.setProperty("height",document.getElementById("printedit-format-height-value").textContent,important);
            element.style.setProperty("min-height","0",important);
            element.style.setProperty("max-height","none",important);
        }
        if (document.getElementById("printedit-format-margin-h").checked)
        {
            element.style.setProperty("margin-left",document.getElementById("printedit-format-margin-h-value").textContent,important);
            element.style.setProperty("margin-right",document.getElementById("printedit-format-margin-h-value").textContent,important);
        }
        if (document.getElementById("printedit-format-margin-v").checked)
        {
            element.style.setProperty("margin-top",document.getElementById("printedit-format-margin-v-value").textContent,important);
            element.style.setProperty("margin-bottom",document.getElementById("printedit-format-margin-v-value").textContent,important);
        }
        if (document.getElementById("printedit-format-padding-h").checked)
        {
            element.style.setProperty("padding-left",document.getElementById("printedit-format-padding-h-value").textContent,important);
            element.style.setProperty("padding-right",document.getElementById("printedit-format-padding-h-value").textContent,important);
        }
        if (document.getElementById("printedit-format-padding-v").checked)
        {
            element.style.setProperty("padding-top",document.getElementById("printedit-format-padding-v-value").textContent,important);
            element.style.setProperty("padding-bottom",document.getElementById("printedit-format-padding-v-value").textContent,important);
        }
        if (document.getElementById("printedit-format-border").checked) element.style.setProperty("border",document.getElementById("printedit-format-border-value").textContent + " solid #606060",important);
        if (document.getElementById("printedit-format-outline").checked) element.style.setProperty("outline",document.getElementById("printedit-format-outline-value").textContent + " solid #606060",important);
        if (document.getElementById("printedit-format-overflow").checked) element.style.setProperty("overflow",document.getElementById("printedit-format-overflow-value").textContent,important);
        if (document.getElementById("printedit-format-float").checked) element.style.setProperty("float",document.getElementById("printedit-format-float-value").textContent,important);
        if (document.getElementById("printedit-format-clear").checked) element.style.setProperty("clear",document.getElementById("printedit-format-clear-value").textContent,important);
        
        if (document.getElementById("printedit-format-color").checked)
        {
            if (document.getElementById("printedit-format-color-value").textContent == "dark") element.style.setProperty("color","#202020",important);
            else if (document.getElementById("printedit-format-color-value").textContent == "onyx") element.style.setProperty("color","#404040",important);
            else if (document.getElementById("printedit-format-color-value").textContent == "iron") element.style.setProperty("color","#606060",important);
            else element.style.setProperty("color",document.getElementById("printedit-format-color-value").textContent,important);
        }
        if (document.getElementById("printedit-format-background").checked)
        {
            if (document.getElementById("printedit-format-background-value").textContent == "light") element.style.setProperty("background-color","#F0F0F0",important);
            else if (document.getElementById("printedit-format-background-value").textContent == "quartz") element.style.setProperty("background-color","#E0E0E0",important);
            else if (document.getElementById("printedit-format-background-value").textContent == "steel") element.style.setProperty("background-color","#D0D0D0",important);
            else if (document.getElementById("printedit-format-background-value").textContent == "red") element.style.setProperty("background-color","#FFE0E0",important);
            else if (document.getElementById("printedit-format-background-value").textContent == "green") element.style.setProperty("background-color","#E0FFE0",important);
            else if (document.getElementById("printedit-format-background-value").textContent == "blue") element.style.setProperty("background-color","#E0E0FF",important);
            else if (document.getElementById("printedit-format-background-value").textContent == "yellow") element.style.setProperty("background-color","#FFFFE0",important);
            else element.style.setProperty("background-color",document.getElementById("printedit-format-background-value").textContent,important);
        }
        if (document.getElementById("printedit-format-fontfamily").checked) element.style.setProperty("font-family",document.getElementById("printedit-format-fontfamily-value").textContent,important);
        if (document.getElementById("printedit-format-fontstyle").checked)
        {
            if (document.getElementById("printedit-format-fontstyle-value").textContent == "italic") element.style.setProperty("font-style","italic",important);
            else element.style.setProperty("font-style","normal",important);
            if (document.getElementById("printedit-format-fontstyle-value").textContent == "bold") element.style.setProperty("font-weight","bold",important);
            else element.style.setProperty("font-weight","normal",important);
        }
        if (document.getElementById("printedit-format-fontsize").checked) element.style.setProperty("font-size",document.getElementById("printedit-format-fontsize-value").textContent,important);
        if (document.getElementById("printedit-format-lineheight").checked) element.style.setProperty("line-height",document.getElementById("printedit-format-lineheight-value").textContent,important);
        if (document.getElementById("printedit-format-verticalalign").checked) element.style.setProperty("vertical-align",document.getElementById("printedit-format-verticalalign-value").textContent,important);
        if (document.getElementById("printedit-format-textalign").checked) element.style.setProperty("text-align",document.getElementById("printedit-format-textalign-value").textContent,important);
        if (document.getElementById("printedit-format-textdecoration").checked) element.style.setProperty("text-decoration","none",important);
        if (document.getElementById("printedit-format-texttransform").checked) element.style.setProperty("text-transform","none",important);
        if (document.getElementById("printedit-format-textdecoration").checked)
        {
            if (document.getElementById("printedit-format-textdecoration-value").textContent == "under") element.style.setProperty("text-decoration","underline",important);
            else if (document.getElementById("printedit-format-textdecoration-value").textContent == "over") element.style.setProperty("text-decoration","overline",important);
            else if (document.getElementById("printedit-format-textdecoration-value").textContent == "through") element.style.setProperty("text-decoration","line-through",important);
            else element.style.setProperty("text-decoration","none",important);
        }
        if (document.getElementById("printedit-format-texttransform").checked)
        {
            if (document.getElementById("printedit-format-texttransform-value").textContent == "capital") element.style.setProperty("text-transform","capitalize",important);
            else if (document.getElementById("printedit-format-texttransform-value").textContent == "lower") element.style.setProperty("text-transform","lowercase",important);
            else if (document.getElementById("printedit-format-texttransform-value").textContent == "upper") element.style.setProperty("text-transform","uppercase",important);
            else element.style.setProperty("text-transform","none",important);
        }
        if (document.getElementById("printedit-format-liststyle").checked) element.style.setProperty("list-style",document.getElementById("printedit-format-liststyle-value").textContent,important);
        if (document.getElementById("printedit-format-tableborder").checked)
        {
            if (document.getElementById("printedit-format-tableborder-value").textContent == "collapse") element.style.setProperty("border-collapse","collapse",important);
            else
            {
                element.style.setProperty("border-collapse","separate",important);
                element.style.setProperty("border-spacing",document.getElementById("printedit-format-tableborder-value").textContent,important);
            }
        }
        if (document.getElementById("printedit-format-pagebreak").checked)
        {
            if (document.getElementById("printedit-format-pagebreak-value").textContent == "before") element.style.setProperty("page-break-before","always",important);
            else if (document.getElementById("printedit-format-pagebreak-value").textContent == "after") element.style.setProperty("page-break-after","always",important);
            else if (document.getElementById("printedit-format-pagebreak-value").textContent == "\u00ACinside") element.style.setProperty("page-break-inside","avoid",important);
        }
    }
    
    if (!document.getElementById("printedit-format-subelements").checked) return;
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                formatElement(element.contentDocument.body);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                formatElement(element.children[i]);
    }
}

/************************************************************************/

/* Text Pieces construct function */

function constructTextPieces(container,textstr)
{
    var i,text,span,br;
    
    container.setAttribute("printedit-text-inserted",cmdNumber);
    
    while (container.childNodes.length > 0) container.removeChild(container.childNodes[0]);
    
    while ((i = textstr.indexOf("\n")) >= 0)
    {
        span = document.createElement("span");
        span.setAttribute("printedit-text-wrapper","");
        text = document.createTextNode(textstr.substr(0,i));
        span.appendChild(text);
        container.appendChild(span);
        
        textstr = textstr.substr(i+1);
        
        span = document.createElement("span");
        span.setAttribute("printedit-break-wrapper","");
        span.appendChild(document.createElement("span"));
        br = document.createElement("br");
        br.style.setProperty("line-height","1.8","");
        br.style.setProperty("vertical-align","top","");
        span.appendChild(br);
        container.appendChild(span);
    }
    
    span = document.createElement("span");
    span.setAttribute("printedit-text-wrapper","");
    text = document.createTextNode(textstr);
    span.appendChild(text);
    container.appendChild(span);
}

/************************************************************************/

/* Undo commands recursive function */

function undoCommands(element)
{
    var i;
    var styleFields = new Array();
    var textFields = new Array();
    
    if (element.hasAttribute("printedit-text-inserted"))
    {
        if (cmdNumber == +element.getAttribute("printedit-text-inserted") || cmdNumber == -1)
        {
            element.parentNode.removeChild(element);
            
            return true;  /* element removed */
        }
    }
    
    if (element.hasAttribute("printedit-text"))
    {
        textFields = element.getAttribute("printedit-text").split("\u204B");
        
        if (cmdNumber == +textFields[0])
        {
            element.textContent = textFields[1];
            
            if (textFields.length == 2) element.removeAttribute("printedit-text");
            else element.setAttribute("printedit-text",textFields.slice(2).join("\u204B"));
        }
        else if (cmdNumber == -1)
        {
            element.textContent = textFields[textFields.length-1];
            
            element.removeAttribute("printedit-text");
        }
    }
    
    if (element.hasAttribute("printedit-style"))
    {
        styleFields = element.getAttribute("printedit-style").split("\u204B");
        
        if (cmdNumber == +styleFields[0])
        {
            element.style.cssText = styleFields[1];
            
            if (styleFields.length == 2) element.removeAttribute("printedit-style");
            else element.setAttribute("printedit-style",styleFields.slice(2).join("\u204B"));
        }
        else if (cmdNumber == -1)
        {
            element.style.cssText = styleFields[styleFields.length-1];
            
            element.removeAttribute("printedit-style");
        }
    }
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                undoCommands(element.contentDocument.body);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                if (undoCommands(element.children[i])) i--;
    }
    
    return false;  /* element not removed */
}

/************************************************************************/

/* Purge elements function */

function purgeElements(element,hidden)
{
    var i,display,visibility,opacity,body,width,height,div;
    
    display = element.style.getPropertyValue("display");
    visibility = element.style.getPropertyValue("visibility");
    opacity = element.style.getPropertyValue("opacity");
    
    if (visibility == "hidden" || opacity == "0") hidden = true;
    
    if (purgeDeleted && display == "none")
    {
        if (element.localName == "body")
        {
            for (i = 0; i < element.children.length; i++)
                if (element.children[i] != null)  /* in case web page not fully loaded */
                    { element.removeChild(element.children[i]); i--; }
        }
        else element.parentNode.removeChild(element);
        
        return true;
    }
    else if (purgeHidden && hidden)
    {
        if (element.localName == "img")
        {
            width = window.getComputedStyle(element,null).getPropertyValue("width");
            height = window.getComputedStyle(element,null).getPropertyValue("height");
            
            div = document.createElement("img");
            div.style.setProperty("width",width,"important");
            div.style.setProperty("height",height,"important");
            div.style.setProperty("display","inline-block","important");
            element.parentNode.replaceChild(div,element);
        }
    }
    else if (purgeStyles && (element.localName == "style" || element.localName == "link")) { element.parentNode.removeChild(element); return true; }
    
    element.removeAttribute("printedit-text-inserted");  /* remove inserted text undo information */
    
    element.removeAttribute("printedit-text");  /* remove edited text undo information */
    
    element.removeAttribute("printedit-style");  /* remove edited style undo information */
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded */
                purgeElements(element.contentDocument.body,hidden);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded */
                if (purgeElements(element.children[i],hidden)) i--;
    }
    
    return false;
}

/************************************************************************/

/* Text Pieces enable and disable functions */

function enableTextPieces(element,linebreaks)
{
    var i,style,span,text,br;
    
    if (element.localName == "object" || element.localName == "embed") return;
    
    if (element.localName == "body" && linebreaks)
    {
        /* Show line breaks as visible back arrows */
        
        style = document.createElement("style");
        style.id = "printedit-style-linebreaks";
        style.type = "text/css";
        style.textContent = styleLineBreaks;
        element.ownerDocument.head.appendChild(style);
    }
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                enableTextPieces(element.contentDocument.body,linebreaks);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else if (element.hasAttribute("printedit-text-wrapper"));  /* wrapper for text piece */
    else if (element.hasAttribute("printedit-break-wrapper"));  /* wrapper for line break */
    else
    {
        for (i = 0; i < element.childNodes.length; i++)
            if (element.childNodes[i] != null)  /* in case web page not fully loaded before editing */
            {
                if (element.childNodes[i].nodeType == 3)  /* text node */
                {
                    if (element.childNodes[i].nodeValue.search(/[\x21-\x7E\x80-\xFF]/) != -1)
                    {
                        span = document.createElement("span");
                        span.setAttribute("printedit-text-wrapper","");
                        text = element.replaceChild(span,element.childNodes[i]);
                        span.appendChild(text);
                    }
                }
                else if (element.childNodes[i].nodeType == 1 && element.childNodes[i].localName == "br")  /* break element node */
                {
                    span = document.createElement("span");
                    span.setAttribute("printedit-break-wrapper","");
                    span.appendChild(document.createElement("span"));
                    br = element.replaceChild(span,element.childNodes[i]);
                    span.appendChild(br);
                }
                else if (element.childNodes[i].nodeType == 1)  /* other element node */
                {
                    enableTextPieces(element.childNodes[i],linebreaks);
                }
            }
    }
}

function disableTextPieces(element)
{
    var i,style,span,text,br;
    
    if (element.localName == "object" || element.localName == "embed") return;
    
    if (element.localName == "body")
    {
        /* Hide line breaks */
        
        style = element.ownerDocument.getElementById("printedit-style-linebreaks");
        
        if (style) element.ownerDocument.head.removeChild(style);
    }
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.body != null)  /* in case web page not fully loaded before editing */
                disableTextPieces(element.contentDocument.body);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
            {
                if (element.children[i].hasAttribute("printedit-text-wrapper"))  /* wrapper for text piece */
                {
                    if (!element.children[i].hasAttribute("printedit-text") && !element.children[i].hasAttribute("printedit-style"))  /* remove wrapper if no text edits made and no styling applied */
                    {
                        span = element.children[i];
                        text = span.removeChild(span.childNodes[0]);
                        element.replaceChild(text,element.children[i]);
                        i--;  /* text node is not in children */
                    }
                }
                else if (element.children[i].hasAttribute("printedit-break-wrapper"))  /* wrapper for line break */
                {
                    if (!element.children[i].hasAttribute("printedit-style"))  /* remove wrapper if no styling applied */
                    {
                        span = element.children[i];
                        span.removeChild(span.childNodes[0]);
                        br = span.removeChild(span.childNodes[0]);
                        element.replaceChild(br,element.children[i]);
                    }
                }
                else disableTextPieces(element.children[i]);
            }
    }
}

/************************************************************************/

/* Display for different device functions */

function displayForDifferentDevice(device)
{
    var i,j,k,contentDocument,styleSheet,cssRule;
    var contentDocuments = new Array();
    
    /* device == "" - reset print and screen stylesheets for print and screen devices */
    /* device == "print" - apply print stylesheets to both print and screen devices */
    /* device == "screen" - apply screen stylesheets to both print and screen devices */
    
    contentDocuments = getDocuments();
    
    for (i = 0; i < contentDocuments.length; i++)
    {
        contentDocument = contentDocuments[i];
       
        for (j = 0; j < contentDocument.styleSheets.length; j++)
        {
            styleSheet = contentDocument.styleSheets[j];
            
            if (isValidStyleSheet(styleSheet) && !isAlternateStyleSheet(styleSheet))
            {
                /* Change media types for media stylesheet */
                
                changeMediaTypes(styleSheet.media,device);
                
                /* Change media types for @media rules */
                
                try
                {
                    if (styleSheet.cssRules != null)
                    {
                        for (k = 0; k < styleSheet.cssRules.length; k++)
                        {
                            cssRule = styleSheet.cssRules[k];
                            
                            if (cssRule.cssText.substr(0,6) == "@media")
                            {
                                changeMediaTypes(cssRule.media,device);
                            }
                        }
                    }
                }
                catch (e) {}  /* attempting cross-domain stylesheet access in Firefox */
                
                /* Force stylesheet to be re-applied */
                
                styleSheet.disabled = true;
                styleSheet.disabled = false;
            }
        }
    }
}

function getDocuments()  /* this is not recursive */
{
    var i,frames,iframes;
    var documents = new Array();
    
    frames = document.getElementsByTagName("frame");
    iframes = document.getElementsByTagName("iframe");
    
    documents.push(document);
    
    for (i = 0; i < frames.length; i++)
    {
        try
        {
            if (frames[i].contentDocument.body != null)  /* in case web page not fully loaded before editing */
                documents.push(frames[i].contentDocument);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    
    for (i = 0; i < iframes.length; i++)
    {
        try
        {
            if (iframes[i].contentDocument.body != null)  /* in case web page not fully loaded before editing */
                documents.push(iframes[i].contentDocument);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    
    return documents;
}

function isValidStyleSheet(styleSheet)
{
    var styleSheetHref;
    
    styleSheetHref = styleSheet.href;
    
    /* If the style sheet href is set and this is not an internal style sheet */
    
    if (styleSheetHref && styleSheetHref.indexOf("about:PreferenceStyleSheet") != 0 && 
        styleSheetHref.indexOf("chrome://") != 0 && styleSheetHref.indexOf("data:text/css") != 0 &&
        styleSheetHref.indexOf("jar:file://") != 0 && styleSheetHref.indexOf("resource://") != 0) return true;
    
    /* If the style sheet href is not set */
    
    if (!styleSheetHref) return true;
    
    return false;
}

function isAlternateStyleSheet(styleSheet)
{
    var ownerNode;
    
    ownerNode = styleSheet.ownerNode;
    
    if (ownerNode)
    {
        if (ownerNode.nodeType == 7)  /* processing instruction node */
        {
            if (ownerNode.data.indexOf('alternate="yes"') != -1) return true;
        }
        else
        {
            if (ownerNode.hasAttribute("rel") && ownerNode.getAttribute("rel").toLowerCase() == "alternate stylesheet") return true;
        }
    }
    
    return false;
}

function hasMediaType(media,mediaType)
{
    var i,type;
    
    for (i = 0; i < media.length; i++)
    {
        type = media.item(i).toLowerCase();
        if (type == mediaType) return true;
    }
    
    return false;
}

function changeMediaTypes(media,device)
{
    /* Reset stylesheet (or @media rule) for device as normal */
    
    if (hasMediaType(media,"printedit-append-screen"))
    {
        media.deleteMedium("screen");
        media.deleteMedium("printedit-append-screen");
    }
    
    if (hasMediaType(media,"printedit-delete-screen"))
    {
        media.appendMedium("screen");
        media.deleteMedium("printedit-delete-screen");
    }
    
    if (hasMediaType(media,"printedit-append-print"))
    {
        media.deleteMedium("print");
        media.deleteMedium("printedit-append-print");
    }
    
    if (hasMediaType(media,"printedit-delete-print"))
    {
        media.appendMedium("print");
        media.deleteMedium("printedit-delete-print");
    }
    
    /* Use print stylesheet (or @media print rule) for screen device */
    
    if (device == "print")
    {
        if (hasMediaType(media,"print"))
        {
            if (!hasMediaType(media,"screen"))
            {
                media.appendMedium("screen");
                media.appendMedium("printedit-append-screen");
            }
        }
        else
        {
            if (hasMediaType(media,"screen"))
            {
                media.deleteMedium("screen");
                media.appendMedium("printedit-delete-screen");
            }
        }
    }
    
    /* Use screen stylesheet (or @media screen rule) for print device */
    
    if (device == "screen")
    {
        if (hasMediaType(media,"screen"))
        {
            if (!hasMediaType(media,"print"))
            {
                media.appendMedium("print");
                media.appendMedium("printedit-append-print");
            }
        }
        else
        {
            if (hasMediaType(media,"print"))
            {
                media.deleteMedium("print");
                media.appendMedium("printedit-delete-print");
            }
        }
    }
}

/************************************************************************/

/* Mouse event handlers */

function onMouseOver(event)
{
    var element;
    
    element = event.target;
    
    if (element.localName == "html") element = element.ownerDocument.body;
    
    if (element.parentNode.localName == "input") element = element.parentNode;  /* <input> element has <div> child which is not in children array */
    
    element.style.setProperty("cursor","default","important");
    if (element.localName == "a") element.style.setProperty("text-decoration","none","important");
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        element.contentWindow.addEventListener("mouseover",onMouseOver,true);
        element.contentWindow.addEventListener("mouseout",onMouseOut,true);
        element.contentWindow.addEventListener("mousedown",onMouseDown,true);
        element.contentWindow.addEventListener("mousemove",onMouseMove,true);
        element.contentWindow.addEventListener("mouseup",onMouseUp,true);
        
        element.contentWindow.addEventListener("click",onClick,true);
        element.contentWindow.addEventListener("dblclick",onDblClick,true);
        element.contentWindow.addEventListener("contextmenu",onContextMenu,true);
        
        element.contentWindow.addEventListener("scroll",onScroll,true);
    }
    
    if (!downFlag && !popupOpen && !insideToolbar(element))
    {
        highlightStack.length = 0;
        
        highlightElement = element;
        
        showHighlightBox();
    }
    
    event.preventDefault();
    event.stopPropagation();
}

function onMouseOut(event)
{
    /* necessary to handle cursor moving off content area */
    
    var element;
    
    element = event.target;
    
    if (element.localName == "html") return;  /* scroll bars */
    
    element.style.removeProperty("cursor");
    if (element.localName == "a") element.style.removeProperty("text-decoration");
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        element.contentWindow.removeEventListener("mouseover",onMouseOver,true);
        element.contentWindow.removeEventListener("mouseout",onMouseOut,true);
        element.contentWindow.removeEventListener("mousedown",onMouseDown,true);
        element.contentWindow.removeEventListener("mousemove",onMouseMove,true);
        element.contentWindow.removeEventListener("mouseup",onMouseUp,true);
        
        element.contentWindow.removeEventListener("click",onClick,true);
        element.contentWindow.removeEventListener("dblclick",onDblClick,true);
        element.contentWindow.removeEventListener("contextmenu",onContextMenu,true);
        
        element.contentWindow.removeEventListener("scroll",onScroll,true);
    }
    
    if (!downFlag && !popupOpen)
    { 
        highlightStack.length = 0;
        
        highlightElement = null;
        
        hideBox(highlightBox);
    }
    
    event.preventDefault();
    event.stopPropagation();
}

function onMouseDown(event)
{
    /* capture-box set start point */
    /* prevents mousedown actions (if there are any) */
    
    var i,element,snap,scrollWidth,scrollHeight;
    var boundingRect = new Object();
    
    element = event.target;
    
    if (element.localName == "html") return;  /* scroll bars */
    
    if (!popupOpen && !insideToolbar(element) && event.button == 0)  /* capture select */
    {
        downFlag = true;
        downKeys = Number(event.ctrlKey)*2+Number(event.shiftKey);
        
        captureWindow = element.ownerDocument.defaultView;
        
        if (captureSnap) snap = 20;
        else snap = 0;
        
        scrollWidth = captureWindow.document.body.scrollWidth;
        scrollHeight = captureWindow.document.body.scrollHeight;
        
        downPageX = event.clientX+captureWindow.scrollX;
        if (downPageX < snap) downPageX = 0;
        else if (downPageX > scrollWidth-snap) downPageX = scrollWidth;
        
        downPageY = event.clientY+captureWindow.scrollY;
        if (captureWindow != window && downPageY < snap) downPageY = 0;  /* iframe or frame */
        else if (downPageY < toolbarHeight+snap) downPageY = toolbarHeight;
        else if (downPageY > scrollHeight-snap) downPageY = scrollHeight;
    }
    else if (popupOpen && event.target.hasAttribute("id") && event.target.getAttribute("id").match(/printedit-.*-header/))  /* move panel */
    {
        downFlag = true;
        
        downPanel = event.target.parentNode;
        boundingRect = downPanel.getBoundingClientRect();
        downOffsetX = boundingRect.left-event.clientX;
        downOffsetY = boundingRect.top-event.clientY;
    }
    else if (popupOpen && insidePopup(element)) return;  /* allow default action and event propagation inside popup panels */
    else if (popupOpen && !insidePopup(element))  /* close open popup */
    {
        closePopup();
        
        if (!insideToolbar(event.target)) showHighlightBox();
    }
    
    event.preventDefault();
    event.stopPropagation();
}

function onMouseMove(event)
{
    /* capture-box drag and draw */
    
    var element,clientWidth,clientHeight,scrollWidth,scrollHeight,innerScreenX,innerScreenY;
    var pageX,pageY,width,height,left,top,anchor,frameLeft,frameTop,frameWidth,frameHeight,clientX,clientY,incX,incY,tbH;
    var boundingRect = new Object();
    
    element = event.target;
    
    if (!popupOpen && downFlag && event.button == 0)  /* capture select */
    {
        /* Note: clientWidth, clientHeight, scrollWidth, scrollHeight on <body> or <html> are not always correct! */
        
        clientWidth = captureWindow.innerWidth;
        scrollWidth = captureWindow.document.body.scrollWidth;
        
        clientHeight = captureWindow.innerHeight;
        scrollHeight = captureWindow.document.body.scrollHeight;
        
        if (scrollWidth > clientWidth) clientHeight = clientHeight-17;
        if (scrollHeight > clientHeight) clientWidth = clientWidth-17;
        
        if (isFirefox)
        {
            innerScreenX = captureWindow.mozInnerScreenX;
            innerScreenY = captureWindow.mozInnerScreenY;
            if (captureWindow != window) innerScreenY += toolbarHeight;  /* iframe or frame */
        }
        else
        {
            innerScreenX = window.screenLeft+(window.outerWidth-window.innerWidth)/2;
            if (captureWindow != window) innerScreenX += calculateFrameLeft(captureWindow.document.body,window)-window.scrollX;
            innerScreenY =  window.screenTop+(window.outerHeight-window.innerHeight)-(window.outerWidth-window.innerWidth)/2;
            if (captureWindow != window) innerScreenY += calculateFrameTop(captureWindow.document.body,window)-window.scrollY;
        }
        
        pageX = event.screenX-innerScreenX+captureWindow.scrollX;
        if (pageX < 0) pageX = 0;
        else if (pageX > scrollWidth) pageX = scrollWidth;
        if (pageX >= downPageX) { left = downPageX; width = pageX-downPageX+1; anchor = 0x0; }
        else { left = pageX; width = downPageX-pageX+1;  anchor = 0x1; }
        
        pageY = event.screenY-innerScreenY+captureWindow.scrollY;
        if (pageY < 0) pageY = 0;
        else if (pageY > scrollHeight) pageY = scrollHeight;
        if (pageY >= downPageY) { top = downPageY; height = pageY-downPageY+1; anchor += 0x0; }
        else { top = pageY; height = downPageY-pageY+1; anchor += 0x2; }
        
        if (captureBox == null && (width >= 6 || height >= 6)) captureBox = createBox("capturebox");
        
        if (captureBox != null)
        {
            captureWidth = width;
            captureHeight = height;
            captureLeft = left;
            captureTop = top;
            
            if (captureWindow != window)  /* iframe or frame scrolling */
            {
                if (isFirefox) frameLeft = captureWindow.mozInnerScreenX-window.mozInnerScreenX+window.scrollX;
                else frameLeft = calculateFrameLeft(captureWindow.document.body,window);
                
                frameWidth = captureWindow.innerWidth;
                if (scrollHeight > clientHeight) frameWidth = frameWidth-17;
                
                left = left-captureWindow.scrollX+frameLeft;
                if (left < frameLeft)
                {
                    if (left+width < frameLeft) { width = 0; left = frameLeft; }
                    else if (left+width > frameLeft+frameWidth) { width = frameWidth; left = frameLeft; }
                    else { width = left+width-frameLeft; left = frameLeft; }
                    
                }
                else if (left+width > frameLeft+frameWidth)
                {
                    if (left > frameLeft+frameWidth) { width = 0; left = frameLeft+frameWidth; }
                    else if (left < frameLeft) { width = frameWidth; left = frameLeft; }
                    else { width = frameLeft+frameWidth-left; }
                }
                
                if (isFirefox) frameTop = captureWindow.mozInnerScreenY-window.mozInnerScreenY+window.scrollY+toolbarHeight;
                else frameTop = calculateFrameTop(captureWindow.document.body,window);
                
                frameHeight = captureWindow.innerHeight;
                if (scrollWidth > clientWidth) frameHeight = frameHeight-17;
                
                top = top-captureWindow.scrollY+frameTop;
                if (top < frameTop)
                {
                    if (top+height < frameTop) { height = 0; top = frameTop; }
                    else if (top+height > frameTop+frameHeight) { height = frameHeight; top = frameTop; }
                    else { height = top+height-frameTop; top = frameTop; }
                }
                else if (top+height > frameTop+frameHeight)
                {
                    if (top > frameTop+frameHeight) { height = 0; top = frameTop+frameHeight; }
                    else if (top < frameTop) { height = frameHeight; top = frameTop; }
                    else { height = frameTop+frameHeight-top; }
                }
            }
            
            showBox(captureBox,null,width,height,left,top,downKeys,anchor);
            hideBox(highlightBox);
            
            clientX = event.screenX-innerScreenX;
            clientY = event.screenY-innerScreenY;
            
            incX = 0;
            if (clientX < 20)
            {
                incX = clientX-20;
                if (captureWindow.scrollX == 0) incX = 0;
            }
            else if (clientX > clientWidth-20)
            {
                incX = clientX-(clientWidth-20);
                if (captureWindow.scrollX >= scrollWidth-captureWindow.innerWidth) incX = 0;
            }
            
            tbH = (captureWindow == window) ? toolbarHeight : 0;
            incY = 0;
            if (clientY < 20+tbH)
            {
                incY = clientY-20-tbH;
                if (captureWindow.scrollY == tbH) incY = 0;
            }
            else if (clientY > clientHeight-20)
            {
                incY = clientY-(clientHeight-20);
                if (captureWindow.scrollY >= scrollHeight-captureWindow.innerHeight) incY = 0;
            }
            
            if (incX != 0 || incY != 0)
            {
                if (incX >= -10 && incX <= 10) incX = incX/2;
                else if (incX > 20) incX = 106;
                else if (incX > 10) incX = 6+(incX-10)*(incX-10);
                else if (incX < -20) incX = -106;
                else if (incX < -10) incX = -6-(incX+10)*(incX+10);
                
                if (incY >= -10 && incY <= 10) incY = incY/2;
                else if (incY > 20) incY = 106;
                else if (incY > 10) incY = 6+(incY-10)*(incY-10);
                else if (incY < -20) incY = -106;
                else if (incY < -10) incY = -6-(incY+10)*(incY+10);
                
                scrollIncX = incX;
                scrollIncY = incY;
                scrollScreenX = event.screenX;
                scrollScreenY = event.screenY;
                
                if (scrollTimeout == null)
                {
                    scrollTimeout = window.setTimeout(function() { onScrollTimeout(); },10);
                }
            }
            else
            {
                if (scrollTimeout != null)
                {
                    window.clearTimeout(scrollTimeout);
                    scrollTimeout = null;
                }
            }
        }
    }
    else if (popupOpen && downFlag)  /* move panel */
    {
        boundingRect = downPanel.getBoundingClientRect();
        
        clientX = event.clientX+downOffsetX;
        clientWidth = window.innerWidth;
        if (document.body.scrollHeight > window.innerHeight) clientWidth -= 17;
        if (clientX < 0) clientX = 0;
        else if (clientX+boundingRect.width > clientWidth) clientX = clientWidth-boundingRect.width;
        
        clientY = event.clientY+downOffsetY;
        clientHeight = window.innerHeight;
        if (document.body.scrollWidth > window.innerWidth) clientHeight -= 17;
        if (clientY < toolbarHeight) clientY = toolbarHeight;
        else if (clientY+26 > clientHeight) clientY = clientHeight-26;  /* panel header height == 26 */
        
        downPanel.style.setProperty("left",clientX+"px","important");
        downPanel.style.setProperty("top",clientY+"px","important");
    }
    else if (popupOpen && insidePopup(element)) return;  /* allow default action and event propagation inside popup panels */
    
    event.preventDefault();
    event.stopPropagation();
}

function onScrollTimeout()
{
    var event;
    
    scrollTimeout = null;
    
    captureWindow.scrollBy(scrollIncX,scrollIncY); 
   
    event = window.document.createEvent("MouseEvents");
    event.initMouseEvent("mousemove",true,true,window,1,scrollScreenX,scrollScreenY,0,0,false,false,false,false,0,null);
    window.dispatchEvent(event);
}

function onMouseUp(event)
{
    /* handles selection and de-selection of elements */
    /* prevents mouseup activation of <object> and <embed> links */
    
    var i,element,selectIndex;
    
    element = event.target;
    
    if (!popupOpen && downFlag && event.button == 0)  /* capture select */
    {        
        downFlag = false;
        
        if (captureBox != null)
        {
            /* capture-box selection or de-selection of multiple elements */
            
            if (downKeys == 0)  /* selection of all elements within capture area */
            {
                selectCaptured(captureWindow.document.body);
            }
            else if (downKeys == 1)  /* selection of all graphic elements within capture area */
            {
                selectCapturedGraphics(captureWindow.document.body);
            }
            else if (downKeys >= 2)  /* de-selection of all elements within capture area */
            {
                deselectCaptured(captureWindow.document.body);
            }
            
            hideBox(captureBox);
            
            destroyBox(captureBox);
            captureBox = null;
        }
        else if (highlightElement != null)
        {
            /* selection or de-selection of individual element */
            
            for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
            {
                if (selectElement[cmdNumber][selectIndex] == highlightElement) break;
            }
            
            if (selectIndex < selectCount[cmdNumber])  /* remove selection box */
            {
                destroyBox(selectBox[selectIndex]);
                hideBox(highlightBox);
                
                for (i = selectIndex; i < selectCount[cmdNumber]-1; i++)
                {
                    selectElement[cmdNumber][i] = selectElement[cmdNumber][i+1];
                    selectBox[i] = selectBox[i+1];
                }
                
                selectCount[cmdNumber]--;
                selectElement[cmdNumber].length--;
                selectBox.length--;
            }
            else  /* add selection box */
            {
                selectElement[cmdNumber][selectCount[cmdNumber]] = highlightElement;
                selectBox[selectCount[cmdNumber]] = createBox("selectbox");
                
                showBox(selectBox[selectCount[cmdNumber]],highlightElement,0,0,0,0,0x0,0x0);
                hideBox(highlightBox);
                
                selectCount[cmdNumber]++;
            }
        }
        
        captureWindow = null;
        
        if (selectCount[cmdNumber] > 0) enableEditButtons();
        else disableEditButtons();
    }
    else if (popupOpen && downFlag)  /* move panel */
    {
        downFlag = false;
    }
    else if (popupOpen && insidePopup(element)) return;  /* allow default action and event propagation inside popup panels */
    
    event.preventDefault();
    event.stopPropagation();
}

function onClick(event)
{
    /* prevents click activation of links */
    
    /* event not fired for right-clicks - see onContextMenu() */
    
    if (!insidePopup(event.target))
    {
        if (!insideToolbar(event.target))
        {
            event.preventDefault();
            event.stopPropagation();
        }
    }
}

function onDblClick(event)
{
    /* prevents double click activation of links */
    
    event.preventDefault();
    event.stopPropagation();
}

function onContextMenu(event)
{
    /* prevents normal context menu popup */
    
    var panel,width,height,left,top;
    var boundingRect = new Object();
    
    panel = document.getElementById("printedit-inspect-panel");
    
    if (popupOpen >= 2) closePopup();
    else if (!downFlag && highlightElement != null)
    {
        if (!insideToolbar(event.target))
        {
            cmdInspectInitPanel();
            
            panel.style.removeProperty("display");
            
            boundingRect = panel.getBoundingClientRect();
            
            width = boundingRect.width;
            height = boundingRect.height;
            
            left = calculateFrameLeft(event.target,window)+event.clientX-window.scrollX;
            top = calculateFrameTop(event.target,window)+event.clientY-window.scrollY;
            
            if (left+width > window.innerWidth && left-width >= 0) left -= width;
            if (top+height > window.innerHeight && top-height >= toolbarHeight) top -= height;
            
            panel.style.setProperty("left",left + "px","important");
            panel.style.setProperty("top",top + "px","important");
            
            popupOpen = 1;
            
            showHighlightBox();
        }
    }
    
    event.preventDefault();
    event.stopPropagation();
}

/************************************************************************/

/*  Scroll and Resize event handlers */

function onScroll(event)
{
    /* handles slider or mouse scroll and re-positions select boxes */
    
    showSelectBoxes();
}

function onResize(event)
{
    /* handles window resize and zoom in/out events and re-positions select boxes */
    
    window.setTimeout(function() { showSelectBoxes(); },300);
}

/************************************************************************/

/*  Key press event handler */

function onKeyDown(event)
{
    var modifiers,shortcut;
    var selection = new Object();
    
    modifiers = event.altKey*4+(event.ctrlKey||event.metaKey)*2+event.shiftKey;
    
    shortcut = true;
    
    if (!popupOpen)
    {
        if (event.keyCode == 0x1B /* ESC */ && modifiers == 0)
        {
            selection = window.getSelection();
            if (selection.rangeCount > 0) selection.removeAllRanges();
            else if (!document.getElementById("printedit-deselect").hasAttribute("disabled")) cmdDeselect();
            else shortcut = false;
        }
        
        else if (event.keyCode == 0x26 /* Up Arrow */ && modifiers == 0) cmdExpandBox();
        else if (event.keyCode == 0x28 /* Down Arrow */ && modifiers == 0) cmdShrinkBox();
        else if (event.keyCode == 0x0D /* Enter */ && modifiers == 0) cmdSelectElement();
        else if (event.keyCode == 0x2E /* Delete */ && modifiers == 0 && !document.getElementById("printedit-delete").hasAttribute("disabled")) cmdDelete();
        else if (event.keyCode == 0x2D /* Insert */ && modifiers == 0 && !document.getElementById("printedit-hide").hasAttribute("disabled")) cmdHide();
        else if (event.keyCode == 0x08 /* Backspace */ && modifiers == 0 && !document.getElementById("printedit-undo").hasAttribute("disabled")) cmdUndo();
        
        else if (event.keyCode == 0xDB /* [ */ && modifiers <= 1 && !document.getElementById("printedit-select").hasAttribute("disabled")) cmdSelectPrevious();
        else if (event.keyCode == 0xDD /* ] */ && modifiers <= 1 && !document.getElementById("printedit-deselect").hasAttribute("disabled")) cmdDeselect();
        
        else if (event.keyCode == "S".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-deselect").hasAttribute("disabled")) cmdDeselect();
        else if (event.keyCode == "S".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-select").hasAttribute("disabled")) cmdSelect();
        else if (event.keyCode == "H".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-hide").hasAttribute("disabled")) cmdHide();
        else if (event.keyCode == "D".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-delete").hasAttribute("disabled")) cmdDelete();
        else if (event.keyCode == "I".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-hideexcept").hasAttribute("disabled")) cmdHideExcept();
        else if (event.keyCode == "E".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-deleteexcept").hasAttribute("disabled")) cmdDeleteExcept();
        else if (event.keyCode == "F".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-format").hasAttribute("disabled")) cmdFormat();
        else if (event.keyCode == "T".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-text").hasAttribute("disabled")) cmdText();
        else if (event.keyCode == "U".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-undo").hasAttribute("disabled")) cmdUndo();
        else if (event.keyCode == "N".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-undoall").hasAttribute("disabled")) cmdUndoAll();
        else if (event.keyCode == "V".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-save").hasAttribute("disabled")) cmdSave();
        else if (event.keyCode == "X".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-textpieces").hasAttribute("disabled")) cmdTextPieces();
        else if (event.keyCode == "W".charCodeAt(0) && modifiers <= 1 && !document.getElementById("printedit-webstyle").hasAttribute("disabled")) cmdWebStyle();
        else if (event.keyCode == "E".charCodeAt(0) && modifiers <= 1 && curMode == 0) cmdEdit();
        else if (event.keyCode == "P".charCodeAt(0) && modifiers <= 1 && curMode == 2) cmdPreview();
        else if (event.keyCode == "C".charCodeAt(0) && modifiers <= 1) cmdClose();
        else if (event.keyCode == "L".charCodeAt(0) && modifiers <= 1) cmdHelp();
        else if (event.keyCode == "O".charCodeAt(0) && modifiers <= 1) cmdOptions();
        
        else if (event.keyCode == "A".charCodeAt(0) && modifiers == 2 && !document.getElementById("printedit-select").hasAttribute("disabled")) cmdSelectAll();
        else if (event.keyCode == "P".charCodeAt(0) && modifiers == 2 && curMode == 2) cmdPreview();
        else if (event.keyCode == "S".charCodeAt(0) && modifiers == 2 && !document.getElementById("printedit-save").hasAttribute("disabled")) cmdSaveStart(true,false);
        else if (event.keyCode == "Z".charCodeAt(0) && modifiers == 2 && !document.getElementById("printedit-undo").hasAttribute("disabled")) cmdUndo();
        
        else shortcut = false;
    }
    else
    {
        if (event.keyCode == 0x1B /* ESC */ && modifiers == 0)
        {
            closePopup();
        }
        else if (popupOpen == 1)  /* inspect */
        {
            if (event.keyCode == 0x26 /* Up Arrow */ && modifiers == 0) cmdExpandBox();
            else if (event.keyCode == 0x28 /* Down Arrow */ && modifiers == 0) cmdShrinkBox();
            else if (event.keyCode == 0x0D /* Enter */ && modifiers == 0) cmdSelectElement();
            else shortcut = false;
        }
        else if (popupOpen == 2)  /* select */
        {
            if (event.keyCode == "A".charCodeAt(0) && modifiers <= 1) cmdSelectAll();
            else if (event.keyCode == "G".charCodeAt(0) && modifiers <= 1) cmdSelectGraphics();
            else if (event.keyCode == "P".charCodeAt(0) && modifiers <= 1) cmdSelectPrevious();
            else shortcut = false;
        }
        else if (popupOpen == 3)  /* delete except */
        {
            if (event.keyCode == "R".charCodeAt(0) && modifiers <= 1) cmdDeleteExceptRestricted();
            else if (event.keyCode == "W".charCodeAt(0) && modifiers <= 1) cmdDeleteExceptWithoutFloat();
            else if (event.keyCode == "U".charCodeAt(0) && modifiers <= 1) cmdDeleteExceptUnrestricted();
            else shortcut = false;
        }
        else if (popupOpen == 6)  /* save */
        {
            if (event.keyCode == "M".charCodeAt(0) && modifiers <= 1) cmdSaveStart(false,false);
            else if (event.keyCode == "U".charCodeAt(0) && modifiers <= 1) cmdSaveStart(false,true);
            else if (event.keyCode == "H".charCodeAt(0) && modifiers <= 1) cmdSaveStart(true,false);
            else if (event.keyCode == "P".charCodeAt(0) && modifiers <= 1) cmdSaveStart(true,true);
            else shortcut = false;
        }
        else shortcut = false;
    }
    
    if (shortcut)
    {
        event.preventDefault();
        event.stopPropagation();
    }
}

/************************************************************************/

/* Inspect Panel */

function cmdInspectInitPanel()
{
    var i,field,top,bottom,left,right,overflowX,overflowY;
    var rgbString,fontFamily,fontStyle,fontWeight,fontSize,lineHeight;
    var textDecoration,textTransform,borderCollapse,borderSpacing,pageBreakBefore,pageBreakAfter,pageBreakInside;
    var rgbColors = new Array();
    
    /* Inspect header element */
    
    document.getElementById("printedit-inspect-element").textContent = "<" + highlightElement.localName + ">";
    
    /* Inspect element properties */
    
    field = document.getElementById("printedit-inspect-left");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("left");
    
    field = document.getElementById("printedit-inspect-top");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("top");
    
    field = document.getElementById("printedit-inspect-width");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("width");
    
    field = document.getElementById("printedit-inspect-height");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("height");
    
    field = document.getElementById("printedit-inspect-margin-h");
    left = window.getComputedStyle(highlightElement,null).getPropertyValue("margin-left");
    right = window.getComputedStyle(highlightElement,null).getPropertyValue("margin-right");
    field.textContent = Math.max(left.substr(0,left.length-2),right.substr(0,right.length-2))+"px";
    
    field = document.getElementById("printedit-inspect-margin-v");
    top = window.getComputedStyle(highlightElement,null).getPropertyValue("margin-top");
    bottom = window.getComputedStyle(highlightElement,null).getPropertyValue("margin-bottom");
    field.textContent = Math.max(top.substr(0,top.length-2),bottom.substr(0,bottom.length-2))+"px";
    
    field = document.getElementById("printedit-inspect-padding-h");
    left = window.getComputedStyle(highlightElement,null).getPropertyValue("padding-left");
    right = window.getComputedStyle(highlightElement,null).getPropertyValue("padding-right");
    field.textContent = Math.max(left.substr(0,left.length-2),right.substr(0,right.length-2))+"px";
    
    field = document.getElementById("printedit-inspect-padding-v");
    top = window.getComputedStyle(highlightElement,null).getPropertyValue("padding-top");
    bottom = window.getComputedStyle(highlightElement,null).getPropertyValue("padding-bottom");
    field.textContent = Math.max(top.substr(0,top.length-2),bottom.substr(0,bottom.length-2))+"px";
    
    field = document.getElementById("printedit-inspect-border");
    top = window.getComputedStyle(highlightElement,null).getPropertyValue("border-top-width");
    bottom = window.getComputedStyle(highlightElement,null).getPropertyValue("border-bottom-width");
    left = window.getComputedStyle(highlightElement,null).getPropertyValue("border-left-width");
    right = window.getComputedStyle(highlightElement,null).getPropertyValue("border-right-width");
    field.textContent = Math.max(top.substr(0,top.length-2),bottom.substr(0,bottom.length-2),left.substr(0,left.length-2),right.substr(0,right.length-2))+"px";
    
    field = document.getElementById("printedit-inspect-outline");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("outline-width");
    
    field = document.getElementById("printedit-inspect-overflow");
    overflowX = window.getComputedStyle(highlightElement,null).getPropertyValue("overflow-x");
    overflowY = window.getComputedStyle(highlightElement,null).getPropertyValue("overflow-y");
    if (overflowX == "hidden" || overflowY == "hidden") field.textContent = "hidden";
    else if (overflowX == "scroll" || overflowY == "scroll") field.textContent = "scroll";
    else if (overflowX == "auto" || overflowY == "auto") field.textContent = "auto";
    else field.textContent = "visible";
    
    field = document.getElementById("printedit-inspect-float");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("float");
    
    field = document.getElementById("printedit-inspect-clear");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("clear");
    
    field = document.getElementById("printedit-inspect-color");
    rgbString = window.getComputedStyle(highlightElement,null).getPropertyValue("color");
    if (rgbString.substr(0,4) == "rgb(")  /* rgb */
    {
        rgbColors = rgbString.match(/rgb\((\d+),\s(\d+),\s(\d+)\)/);
        if (rgbColors[1] == 0 && rgbColors[2] == 0 && rgbColors[3] == 0) field.textContent = "black";
        else if (rgbColors[1] == 255 && rgbColors[2] == 255 && rgbColors[3] == 255) field.textContent = "white";
        else { field.textContent = "#"; for (i = 1; i <= 3; i++) field.textContent += (rgbColors[i]<16?"0":"") + Number(rgbColors[i]).toString(16); }
    }
    else  /* rgba */
    {
        rgbColors = rgbString.match(/rgba\((\d+),\s(\d+),\s(\d+),\s([\d\.]+)\)/);
        if (rgbString == "transparent") field.textContent = "transparent";
        else if (rgbColors[1] == 0 && rgbColors[2] == 0 && rgbColors[3] == 0 && rgbColors[4] == 0 ) field.textContent = "transparent";
        else { field.textContent = "#"; for (i = 1; i <= 3; i++) field.textContent += (rgbColors[i]<16?"0":"") + Number(rgbColors[i]).toString(16); field.textContent += "/" + rgbColors[4]; }
    }
    
    field = document.getElementById("printedit-inspect-background");
    rgbString = window.getComputedStyle(highlightElement,null).getPropertyValue("background-color");
    if (rgbString.substr(0,4) == "rgb(")  /* rgb */
    {
        rgbColors = rgbString.match(/rgb\((\d+),\s(\d+),\s(\d+)\)/);
        if (rgbColors[1] == 0 && rgbColors[2] == 0 && rgbColors[3] == 0) field.textContent = "black";
        else if (rgbColors[1] == 255 && rgbColors[2] == 255 && rgbColors[3] == 255) field.textContent = "white";
        else { field.textContent = "#"; for (i = 1; i <= 3; i++) field.textContent += (rgbColors[i]<16?"0":"") + Number(rgbColors[i]).toString(16); }
    }
    else  /* rgba */
    {
        rgbColors = rgbString.match(/rgba\((\d+),\s(\d+),\s(\d+),\s([\d\.]+)\)/);
        if (rgbString == "transparent") field.textContent = "transparent";
        else if (rgbColors[1] == 0 && rgbColors[2] == 0 && rgbColors[3] == 0 && rgbColors[4] == 0 ) field.textContent = "transparent";
        else { field.textContent = "#"; for (i = 1; i <= 3; i++) field.textContent += (rgbColors[i]<16?"0":"") + Number(rgbColors[i]).toString(16); field.textContent += "/" + rgbColors[4]; }
    }
    
    field = document.getElementById("printedit-inspect-fontfamily");
    fontFamily = window.getComputedStyle(highlightElement,null).getPropertyValue("font-family");
    field.textContent = fontFamily.match(/^['"]?(\w[\s\w]*)['"]?/)[1];
    
    field = document.getElementById("printedit-inspect-fontstyle");
    fontStyle = window.getComputedStyle(highlightElement,null).getPropertyValue("font-style");
    fontWeight = window.getComputedStyle(highlightElement,null).getPropertyValue("font-weight");
    if (fontWeight == "bold" || fontWeight == "800" || fontWeight == "900") fontWeight = "bold"; else fontWeight = "normal";
    if (fontWeight == "normal") field.textContent = fontStyle;
    else if (fontStyle == "normal") field.textContent = fontWeight;
    else field.textContent = fontWeight + " " + fontStyle;
    
    field = document.getElementById("printedit-inspect-fontsize");
    fontSize = window.getComputedStyle(highlightElement,null).getPropertyValue("font-size");
    field.textContent = fontSize;
    
    field = document.getElementById("printedit-inspect-lineheight");
    lineHeight = window.getComputedStyle(highlightElement,null).getPropertyValue("line-height");
    if (lineHeight == "normal") field.textContent = "normal";
    else field.textContent = (lineHeight.substr(0,lineHeight.length-2)/fontSize.substr(0,fontSize.length-2)).toFixed(4);
    
    field = document.getElementById("printedit-inspect-verticalalign");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("vertical-align");
    
    field = document.getElementById("printedit-inspect-textalign");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("text-align");
    
    field = document.getElementById("printedit-inspect-textdecoration");
    textDecoration = window.getComputedStyle(highlightElement,null).getPropertyValue("text-decoration");
    field.textContent = "";
    if (textDecoration.indexOf("underline") >= 0) field.textContent += " under";
    if (textDecoration.indexOf("overline") >= 0) field.textContent += " over";
    if (textDecoration.indexOf("line-through") >= 0) field.textContent += " through";
    if (field.textContent == "") field.textContent = "none";
    
    field = document.getElementById("printedit-inspect-texttransform");
    textTransform = window.getComputedStyle(highlightElement,null).getPropertyValue("text-transform");
    if (textTransform.indexOf("capitalize") >= 0) field.textContent = "capital";
    else if (textTransform.indexOf("uppercase") >= 0) field.textContent = "upper";
    else if (textTransform.indexOf("lowercase") >= 0) field.textContent = "lower";
    else field.textContent = "none";
    
    field = document.getElementById("printedit-inspect-liststyle");
    field.textContent = window.getComputedStyle(highlightElement,null).getPropertyValue("list-style-type");
    
    field = document.getElementById("printedit-inspect-tableborder");
    borderCollapse = window.getComputedStyle(highlightElement,null).getPropertyValue("border-collapse");
    borderSpacing = window.getComputedStyle(highlightElement,null).getPropertyValue("border-spacing");
    borderSpacing = Math.max(borderSpacing.match(/^(\d+)px\s(\d+)px/)[1],borderSpacing.match(/^(\d+)px\s(\d+)px/)[2])+"px";
    if (borderCollapse == "collapse") field.textContent = "collapse";
    else field.textContent = borderSpacing;
    
    field = document.getElementById("printedit-inspect-pagebreak");
    pageBreakBefore = window.getComputedStyle(highlightElement,null).getPropertyValue("page-break-before");
    pageBreakAfter = window.getComputedStyle(highlightElement,null).getPropertyValue("page-break-after");
    pageBreakInside = window.getComputedStyle(highlightElement,null).getPropertyValue("page-break-inside");
    field.textContent = "";
    if (pageBreakBefore == "always") field.textContent += " before";  /* 'left' and 'right' reported as 'always' */
    else if (pageBreakBefore == "avoid") field.textContent += " \u00ACbefore";
    if (pageBreakAfter == "always") field.textContent += " after";  /* 'left' and 'right' reported as 'always' */
    else if (pageBreakAfter == "avoid") field.textContent += " \u00ACafter";
    if (pageBreakInside == "avoid") field.textContent += " \u00ACinside";
    if (field.textContent == "") field.textContent = "auto";
}

/************************************************************************/

/* Format Panel */

function cmdFormatInitPanel()
{
    var headerstr;
    
    /* Format header elements */
    
    headerstr = "<" + selectElement[cmdNumber][0].localName + ">";
    if (selectCount[cmdNumber] > 1) headerstr += " , <" + selectElement[cmdNumber][1].localName + ">";
    if (selectCount[cmdNumber] > 2) headerstr += " , <" + selectElement[cmdNumber][2].localName + ">";
    if (selectCount[cmdNumber] > 3) headerstr += " , ...";
    
    document.getElementById("printedit-format-element").textContent = headerstr;
    
    cmdFormatStateChange();
    
    cmdFormatValueChange();
    
    cmdFormatRestrictChange();
    
    /* First apply/okay command */
    
    formatFirstCmd = true;
}

function cmdFormatTextStyle()
{
    var button,state;
    
    cmdFormatStateChange();
    
    button = document.getElementById("printedit-format-textstyle");
    
    state = (button.getAttribute("printedit-color") != "green");
    
    cmdFormatSetState("color",state);
    cmdFormatSetState("background",state);
    cmdFormatSetState("fontfamily",state);
    cmdFormatSetState("fontsize",state);
    cmdFormatSetState("lineheight",state);
    cmdFormatSetState("textalign",state);
    
    cmdFormatStateChange();
}

function cmdFormatDecoration()
{
    var button,state;
    
    cmdFormatStateChange();
    
    button = document.getElementById("printedit-format-decoration");
    
    state = (button.getAttribute("printedit-color") != "green");
    
    cmdFormatSetState("border",state);
    cmdFormatSetState("outline",state);
    
    cmdFormatSetState("fontstyle",state);
    cmdFormatSetState("verticalalign",state);
    cmdFormatSetState("textdecoration",state);
    cmdFormatSetState("texttransform",state);
    cmdFormatSetState("liststyle",state);
    
    cmdFormatStateChange();
}

function cmdFormatMargins()
{
    var button,state;
    
    cmdFormatStateChange();
    
    button = document.getElementById("printedit-format-margins");
    
    state = (button.getAttribute("printedit-color") != "green");
    
    cmdFormatSetState("margin-h",state);
    cmdFormatSetState("margin-v",state);
    cmdFormatSetState("padding-h",state);
    cmdFormatSetState("padding-v",state);
    cmdFormatSetState("overflow",state);
    
    cmdFormatSetState("tableborder",state);
    
    cmdFormatStateChange();
}

function cmdFormatDimensions()
{
    var button,state;
    
    cmdFormatStateChange();
    
    button = document.getElementById("printedit-format-dimensions");
    
    state = (button.getAttribute("printedit-color") != "green");
    
    cmdFormatSetState("width",state);
    cmdFormatSetState("height",state);
    
    cmdFormatStateChange();
}

function cmdFormatBlockLayout()
{
    var button,state;
    
    cmdFormatStateChange();
    
    button = document.getElementById("printedit-format-blocklayout");
    
    state = (button.getAttribute("printedit-color") != "green");
    
    cmdFormatSetState("left",state);
    cmdFormatSetState("top",state);
    cmdFormatSetState("float",state);
    cmdFormatSetState("clear",state);
    
    cmdFormatSetState("pagebreak",state);
    
    cmdFormatStateChange();
}

function cmdFormatSetAll(state)
{
    /* Format Properties - States */
    
    cmdFormatSetState("left",state);
    cmdFormatSetState("top",state);
    cmdFormatSetState("width",state);
    cmdFormatSetState("height",state);
    cmdFormatSetState("margin-h",state);
    cmdFormatSetState("margin-v",state);
    cmdFormatSetState("padding-h",state);
    cmdFormatSetState("padding-v",state);
    cmdFormatSetState("border",state);
    cmdFormatSetState("outline",state);
    cmdFormatSetState("overflow",state);
    cmdFormatSetState("float",state);
    cmdFormatSetState("clear",state);
    
    cmdFormatSetState("color",state);
    cmdFormatSetState("background",state);
    cmdFormatSetState("fontfamily",state);
    cmdFormatSetState("fontstyle",state);
    cmdFormatSetState("fontsize",state);
    cmdFormatSetState("lineheight",state);
    cmdFormatSetState("verticalalign",state);
    cmdFormatSetState("textalign",state);
    cmdFormatSetState("textdecoration",state);
    cmdFormatSetState("texttransform",state);
    cmdFormatSetState("liststyle",state);
    cmdFormatSetState("tableborder",state);
    cmdFormatSetState("pagebreak",state);
    
    cmdFormatStateChange();
}

function cmdFormatToggleState(property)
{
    cmdFormatSetState(property,document.getElementById("printedit-format-" + property).checked);
    
    cmdFormatStateChange();
}

function cmdFormatSetState(property,state)
{
    document.getElementById("printedit-format-" + property).checked = state;
}

function cmdFormatStateChange()
{
    var button,count;
    
    /* Text Style button color*/
    
    count = 0;
    button = document.getElementById("printedit-format-textstyle");
    if (document.getElementById("printedit-format-color").checked) count++;
    if (document.getElementById("printedit-format-background").checked) count++;
    if (document.getElementById("printedit-format-fontfamily").checked) count++;
    if (document.getElementById("printedit-format-fontsize").checked) count++;
    if (document.getElementById("printedit-format-lineheight").checked) count++;
    if (document.getElementById("printedit-format-textalign").checked) count++;
    if (count == 6) button.setAttribute("printedit-color","green");
    else if (count > 0) button.setAttribute("printedit-color","amber");
    else button.removeAttribute("printedit-color");
    
    /* Decoration button color */
    
    count = 0;
    button = document.getElementById("printedit-format-decoration");
    if (document.getElementById("printedit-format-border").checked) count++;
    if (document.getElementById("printedit-format-outline").checked) count++;
    if (document.getElementById("printedit-format-fontstyle").checked) count++;
    if (document.getElementById("printedit-format-verticalalign").checked) count++;
    if (document.getElementById("printedit-format-textdecoration").checked) count++;
    if (document.getElementById("printedit-format-texttransform").checked) count++;
    if (document.getElementById("printedit-format-liststyle").checked) count++;
    if (count == 7) button.setAttribute("printedit-color","green");
    else if (count > 0) button.setAttribute("printedit-color","amber");
    else button.removeAttribute("printedit-color");
    
    /* Margins button color */
    
    count= 0;
    button = document.getElementById("printedit-format-margins");
    if (document.getElementById("printedit-format-margin-h").checked) count++;
    if (document.getElementById("printedit-format-margin-v").checked) count++;
    if (document.getElementById("printedit-format-padding-h").checked) count++;
    if (document.getElementById("printedit-format-padding-v").checked) count++;
    if (document.getElementById("printedit-format-tableborder").checked) count++;
    if (document.getElementById("printedit-format-overflow").checked) count++;
    if (count == 6) button.setAttribute("printedit-color","green");
    else if (count > 0) button.setAttribute("printedit-color","amber");
    else button.removeAttribute("printedit-color");
    
    /* Dimensions button color */
    
    count= 0;
    button = document.getElementById("printedit-format-dimensions");
    if (document.getElementById("printedit-format-width").checked) count++;
    if (document.getElementById("printedit-format-height").checked) count++;
    if (count == 2) button.setAttribute("printedit-color","green");
    else if (count > 0) button.setAttribute("printedit-color","amber");
    else button.removeAttribute("printedit-color");
    
    /* Block Layout button color */
    
    count= 0;
    button = document.getElementById("printedit-format-blocklayout");
    if (document.getElementById("printedit-format-left").checked) count++;
    if (document.getElementById("printedit-format-top").checked) count++;
    if (document.getElementById("printedit-format-float").checked) count++;
    if (document.getElementById("printedit-format-clear").checked) count++;
    if (document.getElementById("printedit-format-pagebreak").checked) count++;
    if (count == 5) button.setAttribute("printedit-color","green");
    else if (count > 0) button.setAttribute("printedit-color","amber");
    else button.removeAttribute("printedit-color");
}

function cmdFormatResetValue(property,value)
{
    
    cmdFormatSetValue(property,value);
    
    cmdFormatValueChange();
}

function cmdFormatPrevValue(property,valuestring)
{
    var i,label,interval,delay;
    var values = new Array();
    
    function repeat()
    {
        if (--delay < 0 && i > 0) cmdFormatSetValue(property,values[--i]);
    }
    
    function finish()
    {
        window.clearInterval(interval);
        window.removeEventListener("mouseup",finish,true);
        cmdFormatValueChange();
    }
    
    label = document.getElementById("printedit-format-" + property + "-value");
    values = valuestring.split(",");
    
    for (i = values.length-1; i >= 2; i--)
        if (label.textContent == values[i]) break;
        
    cmdFormatSetValue(property,values[--i]);
    
    interval = window.setInterval(repeat,100);
    delay = 5;
    
    window.addEventListener("mouseup",finish,true);
}

function cmdFormatNextValue(property,valuestring)
{
    var i,label,interval,delay;
    var values = new Array();
    
    function repeat()
    {
        if (--delay < 0 && i < values.length-1) cmdFormatSetValue(property,values[++i]);
    }
    
    function finish()
    {
        window.clearInterval(interval);
        window.removeEventListener("mouseup",finish,true);
        cmdFormatValueChange();
    }
    
    label = document.getElementById("printedit-format-" + property + "-value");
    values = valuestring.split(",");
    
    for (i = 0; i <= values.length-3; i++)
        if (label.textContent == values[i]) break;
    
    cmdFormatSetValue(property,values[++i]);

    interval = window.setInterval(repeat,100);
    delay = 5;
    
    window.addEventListener("mouseup",finish,true);
}
    
function cmdFormatSetValue(property,value)
{
    document.getElementById("printedit-format-" + property + "-value").textContent = value;
}

function cmdFormatValueChange()
{
    var changed;
    
    /* Reset Values button state */
    
    changed = document.getElementById("printedit-format-left-value").textContent != "auto" ||
              document.getElementById("printedit-format-top-value").textContent != "auto" ||
              document.getElementById("printedit-format-width-value").textContent != "auto" ||
              document.getElementById("printedit-format-height-value").textContent != "auto" ||
              document.getElementById("printedit-format-margin-h-value").textContent != "0px" ||
              document.getElementById("printedit-format-margin-v-value").textContent != "0px" ||
              document.getElementById("printedit-format-padding-h-value").textContent != "0px" ||
              document.getElementById("printedit-format-padding-v-value").textContent != "0px" ||
              document.getElementById("printedit-format-border-value").textContent != "0px" ||
              document.getElementById("printedit-format-outline-value").textContent != "0px" ||
              document.getElementById("printedit-format-overflow-value").textContent != "visible" ||
              document.getElementById("printedit-format-float-value").textContent != "none" ||
              document.getElementById("printedit-format-clear-value").textContent != "none" ||
              
              document.getElementById("printedit-format-color-value").textContent != "black" ||
              document.getElementById("printedit-format-background-value").textContent != "white" ||
              document.getElementById("printedit-format-fontfamily-value").textContent != "Arial" ||
              document.getElementById("printedit-format-fontstyle-value").textContent != "normal" ||
              document.getElementById("printedit-format-fontsize-value").textContent != "10px" ||
              document.getElementById("printedit-format-lineheight-value").textContent != "1.0" ||
              document.getElementById("printedit-format-verticalalign-value").textContent != "baseline" ||
              document.getElementById("printedit-format-textalign-value").textContent != "left" ||
              document.getElementById("printedit-format-textdecoration-value").textContent != "none" ||
              document.getElementById("printedit-format-texttransform-value").textContent != "none" ||
              document.getElementById("printedit-format-liststyle-value").textContent != "none" ||
              document.getElementById("printedit-format-tableborder-value").textContent != "collapse" ||
              document.getElementById("printedit-format-pagebreak-value").textContent != "before";
              
    if (changed) document.getElementById("printedit-format-resetvalues").removeAttribute("disabled");
    else document.getElementById("printedit-format-resetvalues").setAttribute("disabled","true");
}

function cmdFormatRestrictChange()
{
    document.getElementById("printedit-format-restrict-type").disabled = !document.getElementById("printedit-format-restrict").checked;
}

function cmdFormatResetValues()
{
    /* Format Properties - Values */
    
    document.getElementById("printedit-format-left-value").textContent = "auto";
    document.getElementById("printedit-format-top-value").textContent = "auto";
    document.getElementById("printedit-format-width-value").textContent = "auto";
    document.getElementById("printedit-format-height-value").textContent = "auto";
    document.getElementById("printedit-format-margin-h-value").textContent = "0px";
    document.getElementById("printedit-format-margin-v-value").textContent = "0px";
    document.getElementById("printedit-format-padding-h-value").textContent = "0px";
    document.getElementById("printedit-format-padding-v-value").textContent = "0px";
    document.getElementById("printedit-format-border-value").textContent = "0px";
    document.getElementById("printedit-format-outline-value").textContent = "0px";
    document.getElementById("printedit-format-overflow-value").textContent = "visible";
    document.getElementById("printedit-format-float-value").textContent = "none";
    document.getElementById("printedit-format-clear-value").textContent = "none";
    
    document.getElementById("printedit-format-color-value").textContent = "black";
    document.getElementById("printedit-format-background-value").textContent = "white";
    document.getElementById("printedit-format-fontfamily-value").textContent = "Arial";
    document.getElementById("printedit-format-fontstyle-value").textContent = "normal";
    document.getElementById("printedit-format-fontsize-value").textContent = "10px";
    document.getElementById("printedit-format-lineheight-value").textContent = "1.0";
    document.getElementById("printedit-format-verticalalign-value").textContent = "baseline";
    document.getElementById("printedit-format-textalign-value").textContent = "left";
    document.getElementById("printedit-format-textdecoration-value").textContent = "none";
    document.getElementById("printedit-format-texttransform-value").textContent = "none";
    document.getElementById("printedit-format-liststyle-value").textContent = "none";
    document.getElementById("printedit-format-tableborder-value").textContent = "collapse";
    document.getElementById("printedit-format-pagebreak-value").textContent = "before";
    
    document.getElementById("printedit-format-resetvalues").setAttribute("disabled","true");
}

function cmdFormatToggleProps()
{
    /* Update show properties attributes */
    
    if (document.getElementById("printedit-format-toggleprops").hasAttribute("printedit-showprops"))
    {
        document.getElementById("printedit-format-toggleprops").setAttribute("title","Show Properties");
        document.getElementById("printedit-format-toggleprops").removeAttribute("printedit-showprops");
        document.getElementById("printedit-format-properties").removeAttribute("printedit-showprops");
    }
    else
    {
        document.getElementById("printedit-format-toggleprops").setAttribute("title","Hide Properties");
        document.getElementById("printedit-format-toggleprops").setAttribute("printedit-showprops","");
        document.getElementById("printedit-format-properties").setAttribute("printedit-showprops","");
    }
    
    /* Save show properties setting to local storage */
    
    chrome.storage.local.set(
    {
        "format-showprops": document.getElementById("printedit-format-toggleprops").hasAttribute("printedit-showprops")
    });
}

function cmdFormatApply(okay)
{
    /* Save format panel settings to local storage */
    
    chrome.storage.local.set(
    {
        /* Format Properties - States */
        
        "format-left": document.getElementById("printedit-format-left").checked,
        "format-top": document.getElementById("printedit-format-top").checked,
        "format-width": document.getElementById("printedit-format-width").checked,
        "format-height": document.getElementById("printedit-format-height").checked,
        "format-margin-h": document.getElementById("printedit-format-margin-h").checked,
        "format-margin-v": document.getElementById("printedit-format-margin-v").checked,
        "format-padding-h": document.getElementById("printedit-format-padding-h").checked,
        "format-padding-v": document.getElementById("printedit-format-padding-v").checked,
        "format-border": document.getElementById("printedit-format-border").checked,
        "format-outline": document.getElementById("printedit-format-outline").checked,
        "format-overflow": document.getElementById("printedit-format-overflow").checked,
        "format-float": document.getElementById("printedit-format-float").checked,
        "format-clear": document.getElementById("printedit-format-clear").checked,
        
        "format-color": document.getElementById("printedit-format-color").checked,
        "format-background": document.getElementById("printedit-format-background").checked,
        "format-fontfamily": document.getElementById("printedit-format-fontfamily").checked,
        "format-fontstyle": document.getElementById("printedit-format-fontstyle").checked,
        "format-fontsize": document.getElementById("printedit-format-fontsize").checked,
        "format-lineheight": document.getElementById("printedit-format-lineheight").checked,
        "format-verticalalign": document.getElementById("printedit-format-verticalalign").checked,
        "format-textalign": document.getElementById("printedit-format-textalign").checked,
        "format-textdecoration": document.getElementById("printedit-format-textdecoration").checked,
        "format-texttransform": document.getElementById("printedit-format-texttransform").checked,
        "format-liststyle": document.getElementById("printedit-format-liststyle").checked,
        "format-tableborder": document.getElementById("printedit-format-tableborder").checked,
        "format-pagebreak": document.getElementById("printedit-format-pagebreak").checked,
        
        /* Format Properties - Values */
        
        "format-left-value": document.getElementById("printedit-format-left-value").textContent,
        "format-top-value": document.getElementById("printedit-format-top-value").textContent,
        "format-width-value": document.getElementById("printedit-format-width-value").textContent,
        "format-height-value": document.getElementById("printedit-format-height-value").textContent,
        "format-margin-h-value": document.getElementById("printedit-format-margin-h-value").textContent,
        "format-margin-v-value": document.getElementById("printedit-format-margin-v-value").textContent,
        "format-padding-h-value": document.getElementById("printedit-format-padding-h-value").textContent,
        "format-padding-v-value": document.getElementById("printedit-format-padding-v-value").textContent,
        "format-border-value": document.getElementById("printedit-format-border-value").textContent,
        "format-outline-value": document.getElementById("printedit-format-outline-value").textContent,
        "format-overflow-value": document.getElementById("printedit-format-overflow-value").textContent,
        "format-float-value": document.getElementById("printedit-format-float-value").textContent,
        "format-clear-value": document.getElementById("printedit-format-clear-value").textContent,
        
        "format-color-value": document.getElementById("printedit-format-color-value").textContent,
        "format-background-value": document.getElementById("printedit-format-background-value").textContent,
        "format-fontfamily-value": document.getElementById("printedit-format-fontfamily-value").textContent,
        "format-fontstyle-value": document.getElementById("printedit-format-fontstyle-value").textContent,
        "format-fontsize-value": document.getElementById("printedit-format-fontsize-value").textContent,
        "format-lineheight-value": document.getElementById("printedit-format-lineheight-value").textContent,
        "format-verticalalign-value": document.getElementById("printedit-format-verticalalign-value").textContent,
        "format-textalign-value": document.getElementById("printedit-format-textalign-value").textContent,
        "format-textdecoration-value": document.getElementById("printedit-format-textdecoration-value").textContent,
        "format-texttransform-value": document.getElementById("printedit-format-texttransform-value").textContent,
        "format-liststyle-value": document.getElementById("printedit-format-liststyle-value").textContent,
        "format-tableborder-value": document.getElementById("printedit-format-tableborder-value").textContent,
        "format-pagebreak-value": document.getElementById("printedit-format-pagebreak-value").textContent,
        
        /* Format Controls */
        
        "format-subelements": document.getElementById("printedit-format-subelements").checked,
        "format-important": document.getElementById("printedit-format-important").checked,
        "format-restrict": document.getElementById("printedit-format-restrict").checked,
        "format-restrict-type": document.getElementById("printedit-format-restrict-type").value
    });
    
    /* Perform format command */
    
    cmdFormatPerform(okay);
}

/************************************************************************/

/* Text Panel */

function cmdTextInitPanel(message)
{
    var element,wrapped,headerstr;
    
    element = selectElement[cmdNumber][0];
    wrapped = element.hasAttribute("printedit-text-wrapper");
    
    /* Text header element */
    
    if (wrapped) headerstr = "#text";
    else headerstr = "<" + element.localName + ">";
    
    document.getElementById("printedit-text-element").textContent = headerstr;
    
    /* Text type and content */
    
    if (wrapped)
    {
        document.getElementById("printedit-text-insert-type").removeAttribute("disabled");
        document.getElementById("printedit-text-insert-type").value = "tp";
        document.getElementById("printedit-text-insert-type-tp").removeAttribute("hidden");
        
        document.getElementById("printedit-text-textarea").value = element.textContent;
        
        textPieceValue = element.textContent;
        textInsertValue = "";
    }
    else
    {
        document.getElementById("printedit-text-insert-type").removeAttribute("disabled");
        document.getElementById("printedit-text-insert-type").value = textInsertType;
        document.getElementById("printedit-text-insert-type-tp").setAttribute("hidden",true);
        
        document.getElementById("printedit-text-textarea").value = "";
        
        textInsertValue = "";
    }
    
    /* Focus text area */
    
    document.getElementById("printedit-text-textarea").focus();
    
    /* First apply/okay command */
    
    textFirstCmd = true;
}

function cmdTextInput()
{
    var index,count;
    
    /* Text content */
    
    if (document.getElementById("printedit-text-insert-type").value == "tp")
    {
        index = document.getElementById("printedit-text-textarea").selectionStart;
        
        count = document.getElementById("printedit-text-textarea").value.match(/\n/g).length;  /* may have pasted text containing multiple new lines */
        
        document.getElementById("printedit-text-textarea").value = document.getElementById("printedit-text-textarea").value.replace(/\n/g,"");
        
        document.getElementById("printedit-text-textarea").selectionStart = index-count;
        document.getElementById("printedit-text-textarea").selectionEnd = index-count;
        
        textPieceValue = document.getElementById("printedit-text-textarea").value;
    }
    else textInsertValue = document.getElementById("printedit-text-textarea").value;
}

function cmdTextType()
{
    /* Text content */
    
    if (document.getElementById("printedit-text-insert-type").value == "tp")
    {
        document.getElementById("printedit-text-textarea").value = textPieceValue;
    }
    else document.getElementById("printedit-text-textarea").value = textInsertValue;
    
    /* Re-focus text area */
    
    document.getElementById("printedit-text-textarea").focus();
}

function cmdTextApply(okay)
{
    /* Save insert type setting to local storage */
    
    if (document.getElementById("printedit-text-insert-type").value != "tp")
    {
        textInsertType = document.getElementById("printedit-text-insert-type").value;
        
        chrome.storage.local.set(
        {
            "text-insert-type": textInsertType
        });
    }
    
    /* Perform text command */
    
    cmdTextPerform(okay);
}

/************************************************************************/

/* Help Panel */

function cmdHelpShowPane(pane)
{
    /* Update show pane attributes */
    
    document.getElementById("printedit-help-button").removeAttribute("printedit-showpane");
    document.getElementById("printedit-help-toolbar").removeAttribute("printedit-showpane");
    document.getElementById("printedit-help-mouse").removeAttribute("printedit-showpane");
    document.getElementById("printedit-help-keyboard").removeAttribute("printedit-showpane");
    
    document.getElementById("printedit-help-" + pane).setAttribute("printedit-showpane","");
    
    /* Save show pane setting to local storage */
    
    chrome.storage.local.set(
    {
        "help-showpane": pane
    });
}

/************************************************************************/

/* Selection navigation keys/buttons */

function cmdExpandBox()
{
    var element;
    var boundingRect = new Object();
    
    if (!downFlag && popupOpen <= 1 && highlightElement != null)
    {
        for (element = highlightElement.parentNode; element.localName != "html"; element = element.parentNode)
        {
            boundingRect = element.getBoundingClientRect();
            if (boundingRect.width > 0 || boundingRect.height > 0) break;
        }
        
        if (element.localName == "html") return;
        
        highlightStack.push(highlightElement);
        
        highlightElement = element;
        
        showHighlightBox();
        
        cmdInspectInitPanel();
    }
}

function cmdShrinkBox()
{
    if (!downFlag && popupOpen <= 1 && highlightElement != null)
    {
        if (highlightStack.length == 0) return;
        
        highlightElement = highlightStack.pop();
        
        showHighlightBox();
        
        cmdInspectInitPanel();
    }
}

function cmdSelectElement()
{
    var i,selectIndex;
    
    if (!downFlag && popupOpen <= 1 && highlightElement != null)
    {
        for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
        {
            if (selectElement[cmdNumber][selectIndex] == highlightElement) break;
        }
        
        if (selectIndex < selectCount[cmdNumber])  /* remove selection box */
        {
            destroyBox(selectBox[selectIndex]);
            hideBox(highlightBox);
            
            for (i = selectIndex; i < selectCount[cmdNumber]-1; i++)
            {
                selectElement[cmdNumber][i] = selectElement[cmdNumber][i+1];
                selectBox[i] = selectBox[i+1];
            }
            
            selectCount[cmdNumber]--;
            selectElement[cmdNumber].length--;
            selectBox.length--;
        }
        else  /* add selection box */
        {
            selectElement[cmdNumber][selectCount[cmdNumber]] = highlightElement;
            selectBox[selectCount[cmdNumber]] = createBox("selectbox");
            
            showBox(selectBox[selectCount[cmdNumber]],highlightElement,0,0,0,0,0x0,0x0);
            hideBox(highlightBox);
            
            selectCount[cmdNumber]++;
        }
        
        if (selectCount[cmdNumber] > 0) enableEditButtons();
        else disableEditButtons();
    }
}

/************************************************************************/

/* Print Edit toolbar button commands */

function cmdDeselect()
{
    if (document.getElementById("printedit-deselect").hasAttribute("disabled")) return;
    
    destroySelectBoxes();
    deselectElements();
    
    disableEditButtons();
}

function cmdSelect()
{
    var selectIndex,element;
    
    if (document.getElementById("printedit-select").hasAttribute("disabled")) return;
    
    document.getElementById("printedit-select-previous").setAttribute("disabled","");
    
    if (cmdNumber > 0)
    {
        for (selectIndex = 0; selectIndex < selectCount[cmdNumber-1]; selectIndex++)
        {
            element = selectElement[cmdNumber-1][selectIndex];
            if (element.style.getPropertyValue("display") != "none" &&
                element.style.getPropertyValue("visibility") != "hidden")
                    document.getElementById("printedit-select-previous").removeAttribute("disabled");
        }
    }
    
    document.getElementById("printedit-select").setAttribute("checked","");
    
    document.getElementById("printedit-select-menu").style.removeProperty("display");
    
    popupOpen = 2;
}

function cmdSelectAll()
{
    document.getElementById("printedit-select").removeAttribute("checked");        
    
    document.getElementById("printedit-select-menu").style.setProperty("display","none","important");
    
    popupOpen = 0;
    
    selectElement[cmdNumber][0] = document.body;
    selectBox[0] = createBox("selectbox");
    
    showBox(selectBox[0],document.body,0,0,0,0,0x0,0x0);
    hideBox(highlightBox);
    
    selectCount[cmdNumber]++;
    
    enableEditButtons();
}

function cmdSelectGraphics()
{
    document.getElementById("printedit-select").removeAttribute("checked");        
    
    document.getElementById("printedit-select-menu").style.setProperty("display","none","important");
    
    popupOpen = 0;
    
    captureWindow = window;
    
    captureWidth = document.body.scrollWidth;
    captureHeight = document.body.scrollHeight;
    captureLeft = 0;
    captureTop = 0;
    
    selectCapturedGraphics(document.body);
    
    captureWindow = null;
    
    if (selectCount[cmdNumber] > 0) enableEditButtons();
}

function cmdSelectPrevious()
{
    var selectIndex,element;
    
    document.getElementById("printedit-select").removeAttribute("checked");        
    
    document.getElementById("printedit-select-menu").style.setProperty("display","none","important");
    
    popupOpen = 0;
    
    if (document.getElementById("printedit-select-previous").hasAttribute("disabled")) return;

    for (selectIndex = 0; selectIndex < selectCount[cmdNumber-1]; selectIndex++)
    {
        element = selectElement[cmdNumber-1][selectIndex];
        
        if (element.style.getPropertyValue("display") == "none" ||
            element.style.getPropertyValue("visibility") == "hidden") continue;
        
        selectElement[cmdNumber][selectCount[cmdNumber]] = element;
        selectBox[selectCount[cmdNumber]] = createBox("selectbox");
        
        showBox(selectBox[selectCount[cmdNumber]],element,0,0,0,0,0x0,0x0);
        
        selectCount[cmdNumber]++;
    }
    
    if (selectCount[cmdNumber] > 0) enableEditButtons();
}

function cmdHide()
{
    var selectIndex,element;
    
    if (document.getElementById("printedit-hide").hasAttribute("disabled")) return;
    
    for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
    {
        element = selectElement[cmdNumber][selectIndex];
        
        if (!element.hasAttribute("printedit-style")) element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText);
        else element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText + "\u204B" + element.getAttribute("printedit-style"));
        
        if (hideSelect)
        {
            element.style.setProperty("opacity","0","important");
        }
        else
        {
            element.style.setProperty("visibility","hidden","important");
        }
    }
    
    destroySelectBoxes();
    nextCommand();
    
    disableEditButtons();
    enableUndoButtons();
    
    adjustToolbar();
}

function cmdDelete()
{
    var selectIndex,element;
    
    if (document.getElementById("printedit-delete").hasAttribute("disabled")) return;
    
    for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
    {
        element = selectElement[cmdNumber][selectIndex];
        
        if (!element.hasAttribute("printedit-style")) element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText);
        else element.setAttribute("printedit-style",cmdNumber + "\u204B" + element.style.cssText + "\u204B" + element.getAttribute("printedit-style"));
        
        element.style.setProperty("display","none","important");
    }
    
    destroySelectBoxes();
    nextCommand();
    
    disableEditButtons();
    enableUndoButtons();
    
    adjustToolbar();
}

function cmdHideExcept()
{
    if (document.getElementById("printedit-hideexcept").hasAttribute("disabled")) return;
    
    hideExcept(document.body);
    
    destroySelectBoxes();
    nextCommand();
    
    disableEditButtons();
    enableUndoButtons();
    
    adjustToolbar();
}

function cmdDeleteExcept()
{
    if (document.getElementById("printedit-deleteexcept").hasAttribute("disabled")) return;
    
    document.getElementById("printedit-deleteexcept").setAttribute("checked","");
    
    document.getElementById("printedit-deleteexcept-menu").style.removeProperty("display");
    
    popupOpen = 3;
}

function cmdDeleteExceptRestricted()
{
    document.getElementById("printedit-deleteexcept").removeAttribute("checked");        
    
    document.getElementById("printedit-deleteexcept-menu").style.setProperty("display","none","important");
    
    popupOpen = 0;
    
    deleteExcept(document.body,0);
    
    destroySelectBoxes();
    nextCommand();
    
    disableEditButtons();
    enableUndoButtons();
    
    adjustToolbar();
    
}

function cmdDeleteExceptWithoutFloat()
{
    document.getElementById("printedit-deleteexcept").removeAttribute("checked");        
    
    document.getElementById("printedit-deleteexcept-menu").style.setProperty("display","none","important");
    
    popupOpen = 0;
    
    deleteExcept(document.body,1);
    
    destroySelectBoxes();
    nextCommand();
    
    disableEditButtons();
    enableUndoButtons();
    
    adjustToolbar();
}

function cmdDeleteExceptUnrestricted()
{
    document.getElementById("printedit-deleteexcept").removeAttribute("checked");        
    
    document.getElementById("printedit-deleteexcept-menu").style.setProperty("display","none","important");
    
    popupOpen = 0;
    
    deleteExcept(document.body,2);
    
    destroySelectBoxes();
    nextCommand();
    
    disableEditButtons();
    enableUndoButtons();
    
    adjustToolbar();
}

function cmdFormat()
{
    if (document.getElementById("printedit-format").hasAttribute("disabled")) return;
    
    document.getElementById("printedit-format").setAttribute("checked","");
    
    document.getElementById("printedit-format-panel").style.removeProperty("display");
    
    popupOpen = 4;
    
    cmdFormatInitPanel();
}

function cmdFormatPerform(okay)
{
    var selectIndex,element;
    
    if (okay)
    {
        document.getElementById("printedit-format").removeAttribute("checked");        
        
        document.getElementById("printedit-format-panel").style.setProperty("display","none","important");
        
        popupOpen = 0;
    }
    
    if (!formatFirstCmd)
    {
        deselectElements();
        cmdNumber--;
    }
    
    /* Perform format command */
    
    if (!formatFirstCmd) undoCommands(document.body);
    
    if (document.getElementById("printedit-format-textstyle").hasAttribute("printedit-color") ||
        document.getElementById("printedit-format-decoration").hasAttribute("printedit-color") ||
        document.getElementById("printedit-format-margins").hasAttribute("printedit-color") ||
        document.getElementById("printedit-format-dimensions").hasAttribute("printedit-color") ||
        document.getElementById("printedit-format-blocklayout").hasAttribute("printedit-color"))
    {
        for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
        {
            element = selectElement[cmdNumber][selectIndex];
            
            formatElement(element);
        }
    }
    
    formatFirstCmd = false;
    
    if (!okay)
    {
        nextCommand();
        
        for (selectIndex = 0; selectIndex < selectCount[cmdNumber-1]; selectIndex++)
        {
            element = selectElement[cmdNumber-1][selectIndex];
            selectElement[cmdNumber][selectIndex] = element;
            
            showBox(selectBox[selectIndex],element,0,0,0,0,0x0,0x0);
            
            selectCount[cmdNumber]++;
        }
    }
    else
    {
        destroySelectBoxes();
        nextCommand();
        
        disableEditButtons();
    }
    
    enableUndoButtons();
    
    adjustToolbar();
}

function cmdText()
{
    if (document.getElementById("printedit-text").hasAttribute("disabled")) return;
    
    document.getElementById("printedit-text").setAttribute("checked","");
    
    document.getElementById("printedit-text-panel").style.removeProperty("display");
    
    popupOpen = 5;
    
    cmdTextInitPanel();
}

function cmdTextPerform(okay)
{
    var element,type,textstr,width;
    
    if (okay)
    {
        document.getElementById("printedit-text").removeAttribute("checked");        
        
        document.getElementById("printedit-text-panel").style.setProperty("display","none","important");
        
        popupOpen = 0;
    }
    else document.getElementById("printedit-text-textarea").focus();
    
    if (!textFirstCmd)
    {
        deselectElements();
        cmdNumber--;
    }
    
    /* Perform text command */
    
    element = selectElement[cmdNumber][0];
    
    type = document.getElementById("printedit-text-insert-type").value;
    textstr = document.getElementById("printedit-text-textarea").value;
    width = document.getElementById("printedit-text-textarea").clientWidth-4;
    
    if (textFirstCmd)
    {
        if (type == "tp")  /* Text Piece */
        {
            textContainer = element;
            
            if (!element.hasAttribute("printedit-text")) element.setAttribute("printedit-text",cmdNumber + "\u204B" + element.textContent);
            else element.setAttribute("printedit-text",cmdNumber + "\u204B" + element.textContent + "\u204B" + element.getAttribute("printedit-text"));
            
            element.textContent = textstr;
        }
        else if (type == "ib" || type == "ia")  /* Inline Before or Inline After */
        {
            textContainer = document.createElement("span");
            constructTextPieces(textContainer,textstr);
            
            if (type == "ib") element.parentNode.insertBefore(textContainer,element);
            else element.parentNode.insertBefore(textContainer,element.nextSibling);
        }
        else if (type == "bb" || type == "ba" )  /* Block Before or Block After */
        {
            textContainer = document.createElement("div");
            constructTextPieces(textContainer,textstr);
            
            if (type == "bb") element.parentNode.insertBefore(textContainer,element);
            else element.parentNode.insertBefore(textContainer,element.nextSibling);
        }
        else if (type == "sb" || type == "sa")  /* Sticky Before or Sticky After */
        {
            textContainer = document.createElement("div");
            constructTextPieces(textContainer,textstr);
            
            if (type == "sb") element.parentNode.insertBefore(textContainer,element);
            else element.parentNode.insertBefore(textContainer,element.nextSibling);
            
            textContainer.style.setProperty("width",width+"px","");
            
            textContainer.style.setProperty("margin","4px","");
            textContainer.style.setProperty("padding","8px 10px","");
            textContainer.style.setProperty("border","1px solid #A0A0A0","");
            textContainer.style.setProperty("font-family","Arial,Helvetica","");
            textContainer.style.setProperty("font-size","13px","");
            textContainer.style.setProperty("font-style","normal","");
            textContainer.style.setProperty("font-weight","normal","");
            textContainer.style.setProperty("line-height","16px","");
            textContainer.style.setProperty("word-wrap","break-word","");
            textContainer.style.setProperty("color","#606060","");
            textContainer.style.setProperty("background-color","#FFFFE0","");
        }
    }
    else
    {
        if (type == "tp")  /* Text Piece */
        {
            element.textContent = textstr;
        }
        else
        {
            constructTextPieces(textContainer,textstr);
            
            if (type == "sb" || type == "sa")  /* Sticky Before or Sticky After */
                textContainer.style.setProperty("width",width+"px","");
        }
    }
    
    textFirstCmd = false;
    
    document.getElementById("printedit-text-insert-type").setAttribute("disabled","true");
    
    if (!okay)
    {
        nextCommand();
        
        selectElement[cmdNumber][0] = textContainer;
        
        showBox(selectBox[0],textContainer,0,0,0,0,0x0,0x0);
        
        selectCount[cmdNumber]++;
    }
    else
    {
        destroySelectBoxes();
        nextCommand();
        
        disableEditButtons();
    }
    
    enableUndoButtons();
    
    adjustToolbar();
}

function cmdUndo()
{
    var selectIndex,element;
    
    if (document.getElementById("printedit-undo").hasAttribute("disabled")) return;
    
    destroySelectBoxes();
    deselectElements();
    cmdNumber--;
    
    if (cmdNumber < saveCmdNumber) saveCmdNumber = -1;
    
    undoCommands(document.body);
    
    if (undoReinstate)
    {
        for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
        {
            element = selectElement[cmdNumber][selectIndex];
            selectBox[selectIndex] = createBox("selectbox");
            
            showBox(selectBox[selectIndex],element,0,0,0,0,0x0,0x0);
        }
        
        window.setTimeout(function()  /* in case select element positions change after slow re-layout */
        {
            for (selectIndex = 0; selectIndex < selectCount[cmdNumber]; selectIndex++)
            {
                element = selectElement[cmdNumber][selectIndex];
                
                showBox(selectBox[selectIndex],element,0,0,0,0,0x0,0x0);
            }
        },300);
        
        enableEditButtons();
    }
    else 
    {
        deselectElements();
        
        disableEditButtons();
    }
    
    if (cmdNumber == 0) disableUndoButtons();
    
    adjustToolbar();
}

function cmdUndoAll()
{
    if (document.getElementById("printedit-undoall").hasAttribute("disabled")) return;
    
    destroySelectBoxes();
    
    while (cmdNumber >= 0)
    {
        deselectElements();
        cmdNumber--;
    }
    
    cmdNumber = -1;
    
    undoCommands(document.body);
    
    cmdNumber = 0;
    
    if (cmdNumber < saveCmdNumber) saveCmdNumber = -1;
    
    disableEditButtons();
    disableUndoButtons();
    
    adjustToolbar();
}

function cmdSave()
{
    if (document.getElementById("printedit-save").hasAttribute("disabled")) return;
    
    document.getElementById("printedit-save").setAttribute("checked","");
    
    document.getElementById("printedit-save-menu").style.removeProperty("display");
    
    popupOpen = 6;
}

function cmdSaveStart(asHTML,withPurge)
{
    var i,head,body;
    
    document.getElementById("printedit-save").removeAttribute("checked");        
    
    document.getElementById("printedit-save-menu").style.setProperty("display","none","important");
    
    popupOpen = 0;
    
    if (!asHTML && !withPurge && document.getElementById("printedit-save-mhtml").hasAttribute("disabled")) return;
    if (!asHTML && withPurge && document.getElementById("printedit-save-mhtmlpurge").hasAttribute("disabled")) return;
    
    if (withPurge)
    {
        if (!confirm("PRINT EDIT\n\n" +
                     "Purge may change the page layout and may change element formatting.\n\n" +
                     "Purge cannot be undone. Previous commands cannot be undone.\n\n" +
                     "Do you want to purge the elements in this page?")) return;
        
        /* Purge stylesheets from head */
        
        head = document.head;
        
        for (i = 0; i < head.childNodes.length; i++)
        {
            if (purgeStyles && (head.childNodes[i].localName == "style" || head.childNodes[i].localName == "link"))
            {
                head.removeChild(head.childNodes[i]);
                i--;
            }
        }
        
        /* Purge deleted elements, hidden images, stylesheets and command undo information from body */
        
        body = document.body;
        
        purgeElements(body,false);
        
        /* Destroy select boxes and forget elements selected in previous commands */
        
        destroySelectBoxes();
        
        while (cmdNumber >= 0)
        {
            deselectElements();
            cmdNumber--;
        }
        
        cmdNumber = 0;
        
        saveCmdNumber = -1;
        
        disableEditButtons();
        disableUndoButtons();
        
        adjustToolbar();
    }
        
    /* Save edited web page as MHTML or HTML file */
    
    suspendEditing();
    
    if (asHTML) window.setTimeout(function() { cmdSaveHTML(); },100);
    else chrome.runtime.sendMessage({ type: "saveMHTML" });
}

function cmdSaveHTML()
{
    /* Save as HTML Complete */
    
    cmdSaveFindImages(window,document.documentElement);
    
    cmdSaveLoadImages();
}

function cmdSaveFindImages(frame,element)
{
    var i,baseuri,csstext,style,regex,location;
    var matches = new Array();
    
    if (element == guiContainer) return;
    
    baseuri = frame.document.baseURI;
    
    /* Find external images referenced in <img> elements */
    
    if (element.localName == "img")
    {
        if (element.src.substr(0,5) != "data:")  /* exclude existing data uri - re-editing will not change image */
        {
            if (baseuri != null)
            {
                location = cmdSaveResolveURL(element.src,baseuri);
                
                if (location != null)
                {
                    for (i = 0; i < saveImageLocation.length; i++) if (saveImageLocation[i] == location) break;
                    
                    if (i == saveImageLocation.length)
                    {
                        saveImageLocation[i] = location;
                        saveImageContentType[i] = "";
                        saveImageContent[i] = "";
                    }
                }
            }
        }
    }
    
    /* Find external images referenced in CSS rules */
    
    csstext = "";
    
    if (frame.getComputedStyle(element) != null)
    {
        style = frame.getComputedStyle(element);
        csstext += style.getPropertyValue("background-image") + " ";
        csstext += style.getPropertyValue("list-style-image") + " ";
        
        style = frame.getComputedStyle(element,"before");
        csstext += style.getPropertyValue("background-image") + " ";
        csstext += style.getPropertyValue("list-style-image") + " ";
        csstext += style.getPropertyValue("content") + " ";
        
        style = frame.getComputedStyle(element,"after");
        csstext += style.getPropertyValue("background-image") + " ";
        csstext += style.getPropertyValue("list-style-image") + " ";
        csstext += style.getPropertyValue("content") + " ";
    }
    
    regex = /url\(\s*(?:'|")?([^'")]*)(?:'|")?\s*\)/g;
    
    while ((matches = regex.exec(csstext)) != null)  /* image referenced in style sheet */
    {
        if (matches[1].indexOf("data:") == -1)  /* exclude existing data uri - re-editing will not change image */
        {
            if (baseuri != null)
            {
                location = cmdSaveResolveURL(matches[1],baseuri);
                
                if (location != null)
                {
                    for (i = 0; i < saveImageLocation.length; i++) if (saveImageLocation[i] == location) break;
                    
                    if (i == saveImageLocation.length)
                    {
                        saveImageLocation[i] = location;
                        saveImageContentType[i] = "";
                        saveImageContent[i] = "";
                    }
                }
            }
        }
    }
    
    if (element.localName == "iframe" || element.localName == "frame")
    {
        try
        {
            if (element.contentDocument.documentElement != null)  /* in case web page not fully loaded before editing */
                cmdSaveFindImages(element.contentWindow,element.contentDocument.documentElement);
        }
        catch (e) {}  /* attempting cross-domain web page access */
    }
    else
    {
        for (i = 0; i < element.children.length; i++)
            if (element.children[i] != null)  /* in case web page not fully loaded before editing */
                cmdSaveFindImages(frame,element.children[i]);
    }
}

function cmdSaveLoadImages()
{
    var i;
    
    /* Load external images referenced in <img> elements and CSS */
    
    for (i = 0; i < saveImageLocation.length; i++)
    {
        saveImageContentType[i] = "image/png";  /* default if load fails */
        saveImageContent[i] = "";  /* default if load fails */
        
        chrome.runtime.sendMessage({ type: "saveHTMLLoadImage", index: i, location: saveImageLocation[i] });
    }
    
    saveImageCount = saveImageLocation.length;
    
    if (saveImageCount <= 0) cmdSaveGenerateHTML();
}

function cmdSaveLoadSuccess(index,contenttype,content)
{
    saveImageContentType[index] = contenttype;
    saveImageContent[index] = content;
    
    if (--saveImageCount <= 0) cmdSaveGenerateHTML();
}

function cmdSaveLoadFailure()
{
    if (--saveImageCount <= 0) cmdSaveGenerateHTML();
}

function cmdSaveGenerateHTML()
{
    var htmlString,htmlBlob,objectURL,link;

    /* Generate HTML Complete */
    
    htmlString = cmdSaveExtractHTML(document.documentElement,0);
    
    /* Save to file using HTML5 download attribute */
    
    htmlBlob = new Blob( [ (new Uint8Array([ 0xEF,0xBB,0xBF ])), htmlString ], { type : "text/html" });
    
    objectURL = window.URL.createObjectURL(htmlBlob);
    
    link = document.createElement("a");
    link.download = document.title + ".html";
    link.href = objectURL;
    
    document.body.appendChild(link);
    
    link.click();
    
    window.setTimeout(
    function()
    {
        document.body.removeChild(link);
        
        window.URL.revokeObjectURL(objectURL);
        
        cmdSaveDone(true);
    },100);

    // chrome.runtime.sendMessage({ type: "saveHTMLSaveFile", html: htmlString });
}

function cmdSaveExtractHTML(element,depth)
{
    var i,j,htmlstr,doctype,startTag,endTag,csstext,regex,datauri,htmltext,text;
    var voidElements = new Array("area","base","br","col","embed","frame","hr","img","input","keygen","link","meta","param","source","track","wbr");
    
    /* Extract HTML from DOM and replace external resources with Data URIs */
    
    htmlstr = "";
    
    if (element.localName == "html")
    {
        doctype = element.ownerDocument.doctype;
        
        if (doctype != null)
        {
            htmlstr += '<!DOCTYPE ' + doctype.name + (doctype.publicId ? ' PUBLIC "' + doctype.publicId + '"' : '') +
                       (doctype.systemId && !doctype.publicId ? ' SYSTEM' : '') + (doctype.systemId ? ' "' + doctype.systemId + '"' : '') + '>';
        }
    }
    
    startTag = "<" + element.localName;
    for (i = 0; i < element.attributes.length; i++)
    {
        startTag += " " + element.attributes[i].name;
        if (element.attributes[i].value != "")
        {
            startTag += "=\"";
            startTag += element.attributes[i].value.replace(/"/g,"&quot;");
            startTag += "\"";
        }
    }
    startTag += ">";
    
    if (voidElements.indexOf(element.localName) >= 0) endTag = "";
    else endTag = "</" + element.localName + ">";
    
    if (element == guiContainer) return;
    
    if (element.hasAttribute("style"))  /* any tag with style attribute */
    {
        csstext = element.getAttribute("style");
        
        saveBaseURI = element.ownerDocument.baseURI;
        
        regex = /url\(\s*(?:'|")?(?!data:)([^'")]*)(?:'|")?\s*\)/g;  /* exclude existing data uri's - re-editing will not change images */
        
        csstext = csstext.replace(regex,cmdSaveReplaceCSSURL);
        
        startTag = startTag.split("style=\"" + element.getAttribute("style").replace(/"/g,"&quot;") + "\"").join("style=\"" + csstext.replace(/"/g,"&quot;") + "\"");
    }
    
    if (element.localName == "script") ;  /* ignore script */
    else if (element.localName == "style")  /* internal style sheet */
    {
        csstext = element.textContent;
        
        saveBaseURI = element.ownerDocument.baseURI;
        
        regex = /url\(\s*(?:'|")?(?!data:)([^'")]*)(?:'|")?\s*\)/g;  /* exclude existing data uri's - re-editing will not change images */
        
        csstext = csstext.replace(regex,cmdSaveReplaceCSSURL);
        
        htmlstr += startTag + csstext + endTag;
    }
    else if ((element.localName == "link" && element.hasAttribute("rel") && element.getAttribute("rel") == "stylesheet"))  /* external style sheet */
    {
        if (element.href != null)
        {
            if (element.href.substr(0,5) != "data:")  /* exclude existing data uri - re-editing will not change style sheet */
            {
                /* should be able to use element.sheet - but this is null in print preview static DOM */
                
                csstext = "";
                
                for (i = 0; i < element.ownerDocument.styleSheets.length; i++)
                    if (element.href == element.ownerDocument.styleSheets[i].href) break;
                
                if (i < element.ownerDocument.styleSheets.length)
                {
                    try
                    {
                        for (j = 0; j < element.ownerDocument.styleSheets[i].cssRules.length; j++)
                        {
                            csstext += element.ownerDocument.styleSheets[i].cssRules[j].cssText;
                        }
                    }
                    catch (e) {}  /* attempting cross-domain web page access */
                }
                
                saveBaseURI = element.href;
                
                regex = /url\(\s*(?:'|")?(?!data:)([^'")]*)(?:'|")?\s*\)/g;  /* exclude existing data uri's - re-editing will not change images */
                
                csstext = csstext.replace(regex,cmdSaveReplaceCSSURL);
                
                csstext = "data:text/css;charset=utf-8," + encodeURIComponent(csstext);
                
                startTag = startTag.replace(/href="[^"]*"/,"href=\"" + csstext + "\"");
                
                htmlstr += startTag;
            }
            else htmlstr += startTag;
        }
    }
    else if (element.localName == "img")  /* external image */
    {
        if (element.src != null)
        {
            if (element.src.substr(0,5) != "data:")  /* exclude existing data uri - re-editing will not change image */
            {
                saveBaseURI = element.ownerDocument.baseURI;
                
                datauri = cmdSaveReplaceURL(element.src);
          
                startTag = startTag.replace(/src="[^"]*"/,"src=\"" + datauri + "\"");
                
                htmlstr += startTag;
            }
            else htmlstr += startTag;
        }
    }
    else if (element.localName == "iframe" || element.localName == "frame")  /* inline frame or frameset frame */
    {
        if (element.src != null)
        {
            /* allow existing data uri - re-editing may change frame html */
            
            try
            {
                if (element.contentDocument.documentElement != null)  /* in case web page not fully loaded before editing */
                {
                    if (depth < saveMaxDepth)
                    {
                        htmltext = cmdSaveExtractHTML(element.contentDocument.documentElement,depth+1);
                        
                        htmltext = "data:text/html;charset=utf-8," + encodeURIComponent(htmltext);
                        
                        startTag = startTag.replace(/src="[^"]*"/,"src=\"" + htmltext + "\"");
                        
                        htmlstr += startTag + endTag;
                    }
                }
            }
            catch (e) {}  /* attempting cross-domain web page access */
        }
    }
    else  /* any other tag */
    {
        htmlstr += startTag;
        
        for (i = 0; i < element.childNodes.length; i++)
        {
            if (element.childNodes[i] != null)  /* in case web page not fully loaded before editing */
            {
                if (element.childNodes[i].nodeType == 1) htmlstr += cmdSaveExtractHTML(element.childNodes[i],depth);
                else if (element.childNodes[i].nodeType == 3)
                {
                    text = element.childNodes[i].textContent;
                    text = text.replace(/[\u00A0-\uFFFF]/g,function(charstr) { return "&#" + charstr.charCodeAt(0) + ";"; } );
                    text = text.replace(/</g,"&lt;");
                    text = text.replace(/>/g,"&gt;");
                    htmlstr += text;
                }
            }
        }
            
        htmlstr += endTag;
    }
    
    return htmlstr;
}

function cmdSaveReplaceCSSURL(match,p1,offset,string)  /* saveBaseURI must already be set */
{
    return ("url(" + cmdSaveReplaceURL(p1) + ")");
}

function cmdSaveReplaceURL(url)  /* saveBaseURI must already be set */
{
    var i,location;
    
    if (saveBaseURI != null)
    {
        location = cmdSaveResolveURL(url,saveBaseURI);
        
        if (location != null)
        {
            for (i = 0; i < saveImageLocation.length; i++)
                if (location == saveImageLocation[i])
                    return ("data:" + saveImageContentType[i] + ";base64," + saveImageContent[i]);
            
            return ("data:image/png;base64,");
        }
        else return ("data:image/png;base64,");
    }
    else return ("data:image/png;base64,");
}

function cmdSaveResolveURL(url,baseuri)
{
    var old_base,old_href,base,a;
    
    old_base = document.getElementsByTagName("base")[0];
    if (old_base && old_base.href) old_href = old_base.href;
    
    if (old_base) base = old_base;
    else base = document.head.appendChild(document.createElement("base"));
    
    a = document.createElement("a");
    
    base.href = baseuri;
    a.href = url;
    url = a.href;  /* browser resolves url */
    
    if (old_base) old_base.href = old_href;
    else document.head.removeChild(base);
    
    return url;
}

function cmdSaveDone(cancel)
{
    if (!cancel) saveCmdNumber = cmdNumber;
    
    resumeEditing();
}

function cmdTextPieces()
{
    var checked;
    
    if (document.getElementById("printedit-textpieces").hasAttribute("disabled")) return;
    
    checked = document.getElementById("printedit-textpieces").hasAttribute("checked");
    
    if (checked) document.getElementById("printedit-textpieces").removeAttribute("checked");
    else document.getElementById("printedit-textpieces").setAttribute("checked","");
    
    checked = !checked;
    
    if (checked) enableTextPieces(document.body,textpiecesBreaks);
    else disableTextPieces(document.body);
}

function cmdWebStyle()
{
    var checked,selectIndex,element;
    
    if (document.getElementById("printedit-webstyle").hasAttribute("disabled")) return;
    
    checked = document.getElementById("printedit-webstyle").hasAttribute("checked");
    
    if (checked) document.getElementById("printedit-webstyle").removeAttribute("checked");
    else document.getElementById("printedit-webstyle").setAttribute("checked","");
    
    checked = !checked;
    
    displayForDifferentDevice((checked)?"screen":"print");
    
    showSelectBoxes();
    
    adjustToolbar();
}

function cmdEdit()
{
    startEditing();
    resumeEditing();
}

function cmdPreview()
{
    suspendEditing();
    
    window.print();  /* blocks execution of all content scripts */
    
    resumeEditing();
}

function cmdClose()
{
    if (curMode == 0) chrome.runtime.sendMessage({ type: "closeTab" });  /* close before editing */
    else
    {
        finishEditing();
        
        if (curMode == 0) chrome.runtime.sendMessage({ type: "closeTab" });  /* editing exited */
    }
}

function cmdHelp()
{
    var button,panel;
    var buttonRect = new Object();
    var panelRect = new Object();
    
    document.getElementById("printedit-help").setAttribute("checked","");
    
    document.getElementById("printedit-help-panel").style.removeProperty("display");
    
    popupOpen = 6;
    
    /* Position help panel */
    
    button = document.getElementById("printedit-help");
    panel = document.getElementById("printedit-help-panel");
    buttonRect = button.getBoundingClientRect();
    panelRect = panel.getBoundingClientRect();
    panel.style.setProperty("left",(buttonRect.right-panelRect.width)+"px","important");
}

function cmdOptions()
{
    chrome.runtime.sendMessage({ type: "openOptions" });
}

/************************************************************************/

/* Debug function */

function debugMessage(module,information)
{
    var debug1,debug2,debug3,debug4;

    debugCount++;
    
    debug1 = document.getElementById("printedit-debug-1");
    debug2 = document.getElementById("printedit-debug-2");
    debug3 = document.getElementById("printedit-debug-3");
    debug4 = document.getElementById("printedit-debug-4");
    
    debug4.textContent = debug3.textContent;
    debug3.textContent = debug2.textContent;
    debug2.textContent = debug1.textContent;
    debug1.textContent = debugCount + " - " + module + " - " + information;
}

/************************************************************************/
